/*
 *   IBM Linux Device Font Library
 *   Copyright (c) International Business Machines Corp., 2002
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Author: David L. Wagner
 *   October 18, 2002
 *   $Id: XmlFile.cpp,v 1.2 2002/12/06 00:16:52 jzhuo Exp $
 */

#include "XmlFile.hpp"

#include <iostream>
#include <string>

#include "MyDOMErrorHandler.hpp"
#include "DeviceFontException.hpp"

#include <xercesc/dom/DOM.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/sax/HandlerBase.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/PlatformUtils.hpp>

using namespace std;
using namespace DevFont;

/** The static initializer. */
bool XmlFile::initialize ()
{
  try
  {
    XMLPlatformUtils::Initialize();
  }
  catch(const XMLException& toCatch)
  {
    char* message = XMLString::transcode(toCatch.getMessage());
    cout << "Error during initialization! :\n"
         << string(message) << "\n";
    delete [] message;
  }
  return true;
}

/** Run the static initializer. */
bool XmlFile::initialized = XmlFile::initialize();

/**
 * The constructor parses the given XML file and remembers the root element
 * of the document.  We keep track of the parser as well, since the document
 * goes away when the parser does.
 *
 * @param path  the path to the XML file
 */
XmlFile::XmlFile (const char *path):
  parser(0), rootElement(0)
{
#ifdef DEBUG
cout << "XmlFile::XmlFile(" << path << ") called" << endl;
#endif
  //
  // Store file name
  // 
  fileName = new char[1+strlen(path)];
  strcpy(fileName, path); 

  //
  // Instantiate the DOM parser.
  // 

  static const XMLCh gLS[] = { chLatin_L, chLatin_S, chNull };
  //DOMImplementation *impl = DOMImplementationRegistry::getDOMImplementation(gLS);
  //parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
  DOMImplementationLS *impl = static_cast<DOMImplementationLS *>
    (DOMImplementationRegistry::getDOMImplementation(gLS));
  parser = impl->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);

  //
  // Set checking flags; for simplicity and efficiency, we're just going to
  // assume that all UPDF files are going to be valid.
  //
  bool doNamespaces       = false;
  bool doSchema           = false;
  bool schemaFullChecking = false;
  bool doList             = false;
  bool errorOccurred      = false;

  parser->setFeature(XMLUni::fgDOMNamespaces, doNamespaces);
  parser->setFeature(XMLUni::fgXercesSchema, doSchema);
  parser->setFeature(XMLUni::fgXercesSchemaFullChecking, schemaFullChecking);

  parser->setFeature(XMLUni::fgDOMValidateIfSchema, doList);
  parser->setFeature(XMLUni::fgDOMValidation, errorOccurred);

  // enable datatype normalization - default is off
  parser->setFeature(XMLUni::fgDOMDatatypeNormalization, true);

  MyDOMErrorHandler* errHandler = new MyDOMErrorHandler();
  parser->setErrorHandler(errHandler);

   try
     {
      //
      // Now parse the UPDF document
      //
      DOMDocument *fontDOMDoc = parser->parseURI(path);
         if (fontDOMDoc == NULL)
	{
	  throw DeviceFontException("Device UPDF file not found!");
	}
#ifdef DEBUG
         cout << "    ...fontDOMDoc = " << fontDOMDoc << endl;
#endif
         rootElement = fontDOMDoc->getDocumentElement();
     }
   catch (const XMLException& toCatch) 
     {
       //char* message = XMLString::transcode(toCatch.getMessage());
       throw DeviceFontException(XMLString::transcode(toCatch.getMessage()));
       //delete [] message;
     }
   catch (const DOMException& toCatch)
     {
       //char* message = XMLString::transcode(toCatch.msg);
       throw DeviceFontException(XMLString::transcode(toCatch.msg));
       //delete [] message;
     }
   //catch (...)
     // {
   // throw DeviceFontException("Exception message is: Unexpected Excpetion!");
   // }
   
   //
   // Delete the parser (and the DOMDocument) and the error handler.
   //
   delete errHandler;
}


/**
 * The destructor deletes the parser (and therefore the whole DOM tree).
 */
XmlFile::~XmlFile ()
{
  //
  // When the parser is deleted, so is the whole DOM tree (which is why we
  // have to keep the parser around for the entire lifetime of this object)
  //
  delete parser;
  parser = 0;
  rootElement = 0;

  //
  // Delete the file name
  //
  if (fileName)
    {
      delete fileName;
      fileName = 0;
    }
}


/**
 * Find an Element with the given name in the tree start a the given Node.  If
 * the given node is the one we are looking for, then just return that.
 * Otherwise it returns the (first) descendent that it finds with that name
 * (therefore, this will be mostly useful only if there is exactly one Element
 * with that name in the whole tree, or if we checking to see if the given node
 * is exactly the one we want).
 *
 * @param n     starting node to search
 * @param name  name of the node to retrieve
 */
DOMElement* XmlFile::findElement (const DOMNode *n, const char *name)
{
  //
  // If the node is null, we are done.
  //
  if (n) {
    //
    // First check to see if the node we are starting at is the node we want.
    //
    if (n->getNodeType() == DOMNode::ELEMENT_NODE) {
      const char* nodeName = XMLString::transcode(n->getNodeName());
      //
      // Check the name
      //
      if(strcmp(nodeName, name) == 0) {
        //
        // Found it; clean up our temporary, and return the object
        //
        delete nodeName;
        return (DOMElement*)n;
      }
      //
      // Didn't find it; still clean up the temporary
      //
      delete nodeName;

      //
      // OK, so now let's check to see if there is a child of this node that is
      // the one we want.
      //
      for (DOMNode *child = n->getFirstChild(); child != 0; child=child->getNextSibling())
        {
          //
          // Check to see if the child is the node
          //
          DOMElement* tmp = findElement (child, name);
          if(tmp) {
            return tmp;
          }
        }
    }
  }
  return NULL;
}


/**
 * Find an Element with the given name starting from the root of the document.
 * If we were being fancy, we should really use XPATH expressions to find a
 * particular node, but here we are relying on the fact that in UPDF files
 * there are no Elements with the same name in different contexts.
 * Otherwise, it would be likely that one would get the wrong node.
 *
 * @param name  name of the node to retrieve
 */
DOMElement* XmlFile::findElement (const char *name) const
{
  return findElement (rootElement, name);
}


namespace DevFont
{
/** Print out the XmlFile name for debugging. */
ostream & DevFont::operator<< (ostream & s, const XmlFile * file)
{
   s << "XmlFile: " << file->fileName;
   return s;
}
}


