/*
 *   IBM Linux Device Font Library
 *   Copyright (c) International Business Machines Corp., 2002
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Author: Julie Zhuo
 *   Dec. 24, 2002
 *   $Id: UPDFFontParser.cpp,v 1.1 2002/12/24 22:04:55 jzhuo Exp $
 */

#include "UPDFFontParser.hpp"
#include "DeviceFont.hpp"
#include "DeviceFontException.hpp"

#include <iostream>
#include <string>
#include <stdio.h>
#include <sys/stat.h>

#include "CommandSequence.hpp"
#include "Event.hpp"
#include "MyDOMErrorHandler.hpp"
#include "Panose.hpp"
#include "Parameter.hpp"
#include "XmlFile.hpp"

#include <xercesc/dom/DOM.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/XMLUni.hpp>
#include <xercesc/util/XMLUTF8Transcoder.hpp>

using namespace std;
using namespace DevFont;

/**
 * Construct a new parser object that will read in the indicated file.  It
 * will then construct a list of DeviceFont objects and a list of command
 * sequence strings.  The path must point to the top-level UPDF Device
 * Configuration file; this object will then read the Device Description file
 * to get the font information and the Command Sequence file.
 *
 * Since the XmlFile objects hold on to the DOM tree for their entire lifetime,
 * we try to delete the objects as soon as we possibly can for memory
 * efficiency.
 *
 * @param dcPath  path to the Device Configuration file
 */
UPDFFontParser::UPDFFontParser (const char *dcPath)
{
  //
  // First verify that the configuration file exists
  //
  struct stat dcFileStat;
  if ( -1 == stat (dcPath, &dcFileStat))
    throw DeviceFontException ("Device UPDF configuration file not found!");
  //
  // Next create the transcoder, which we will need to get the attributes
  //
  transcoder = new XMLUTF8Transcoder (XMLUni::fgUTF8EncodingString, BUFSIZE);
  //
  // Open the Device Configuration file
  //
  XmlFile *dcFile = new XmlFile (dcPath);

  const DOMElement *dcElement = dcFile->findElement ("DeviceConfiguration");
  //
  // Get the names of the files containing the Device Description and the
  // Command Sequences
  //
  const char *ddName = getAttribute (dcElement, "Unit_Description_Reference");
  const char *csName = getAttribute (dcElement, "CommandSequences_Reference");
  delete dcFile;
  //
  // Get the paths to the files using the Device Configuration file path
  // as the base.
  //
  const char *ddPath = getPath (dcPath, ddName);
  const char *csPath = getPath (dcPath, csName);

  delete [] ddName;
  delete [] csName;
  //
  // Run the common initialization
  //
  init (ddPath, csPath);
  //
  // And clean up
  //
  delete ddPath;
  delete csPath;
}

/**
 * Constructor that takes paths to the Device Description and Command
 * Sequence files.
 *
 * @param ddPath  path to the Device Description file
 * @param csPath  path to the Device Description file
 */
UPDFFontParser::UPDFFontParser (const char *ddPath, const char *csPath)
{
    //
    // Next create the transcoder, which we will need to get the attributes
    //
    transcoder = new XMLUTF8Transcoder (XMLUni::fgUTF8EncodingString, BUFSIZE);
    //
    // Run the common initialization
    //
    init (ddPath, csPath);
}

/**
 * Common initialize function shared by both constructors.
 */
void UPDFFontParser::init (const char *ddPath, const char *csPath)
{
    //
    // First verify that the files exist
    //
    struct stat ddFileStat;
    if ( -1 == stat (ddPath, &ddFileStat))
        throw DeviceFontException ("Device UPDF description file not found!");

    struct stat csFileStat;
    if ( -1 == stat (csPath, &csFileStat))
        throw DeviceFontException ("Device UPDF command sequence file not found!");
    //
    // Parse the Device Configuration file
    //
    XmlFile *ddFile = new XmlFile(ddPath);
    const DOMElement *ddRootElement = ddFile->getRootElement ();
    myFonts = findFonts (ddRootElement);
    myEvents = findEvents (ddRootElement);
    delete ddFile;
    //
    // Parse the Command Sequences file
    //
    XmlFile *csFile = new XmlFile(csPath);
    const DOMElement *csRootElement = csFile->getRootElement ();
    myCmdSeqs = findCmdSeqs (csRootElement);
    myParameters = findParameters (csRootElement);
    delete csFile;
}
    
/**
 * The destructor must get rid of all the objects this class owns.
 */
UPDFFontParser::~UPDFFontParser()
{
  //
  // Delete all the fonts
  //
  if (myFonts != NULL)
    {
      const DeviceFont *current = myFonts;
      while (current != NULL)
        {
          const DeviceFont *temp = current->getNext();
          delete current;
          current = temp;
        }
    }

  //
  // Delete all the command sequences
  //
  if (myCmdSeqs != NULL)
    {
      const CommandSequence *current = myCmdSeqs;
      while (current != NULL)
        {
          const CommandSequence *temp = current->getNext();
          delete current;
          current = temp;
        }
    }

  //
  // Delete all the events
  //
  if (myEvents != NULL)
    {
      const Event *current = myEvents;
      while (current != NULL)
        {
          const Event *temp = current->getNext();
          delete current;
          current = temp;
        }
    }

  //
  // Delete the transcoder
  //
  delete transcoder;
  //
  // Set these to null for good measure
  //
  myFonts = NULL;
  myCmdSeqs = NULL;
  myEvents = NULL;
  transcoder = NULL;
}


/**
 * Find all the DeviceFont elements and return a linked list of DeviceFont
 * objects with all their information.  Or it will return NULL if there is
 * no font information, but this may be a bit pathological.
 */
DeviceFont *UPDFFontParser::findFonts (const DOMNode *root) const
{
#ifdef DEBUG
cout << "UPDFFontParser::findFonts(" << root << ") called" << endl;
#endif
  //
  // First get the DeviceFontList node
  //
  DOMElement *devicefontlist = XmlFile::findElement (root, "DeviceFontList");
  //
  // Set the head of the list to null
  //
  DeviceFont *head = NULL;
  DeviceFont *current = NULL;
  //
  // Now create a DeviceFont object from each of the children of this node
  //
  for (DOMNode *child = devicefontlist->getFirstChild(); child != 0; child=child->getNextSibling())
    {
      //
      // If the child is a "DeviceFont" element, findElement() will just return
      // that element; otherwise it returns NULL.
      //
      DOMElement* tmp = XmlFile::findElement (child, "DeviceFont");
      if(tmp) {
        //
        // Convert this Element into a DeviceFont object.  We throw away the
        // const'ness, since we need to modify its next pointer (which still
        // preserves the logical constness of the object).
        //
        DeviceFont *fn = const_cast<DeviceFont *>(getDeviceFont (tmp));
        //
        // If the list head is null, then set the current node to be the list
        // head; otherwise, the setNext() function adds the node to the tail.
        //
        if (head == NULL)
          head = fn;
        else
          current->setNext (fn);
        //
        // Set the pointer for the next time around
        //
        current = fn;
      }
    }
  return head;
}


/**
 * Find all the CommandSequence elements and return a linked list of 
 * CommandSequence objects with all their information.  It will return null
 * if there is no command sequence information.
 */
CommandSequence *UPDFFontParser::findCmdSeqs (const DOMNode *root) const
{
  //
  // First get the CommandSequences node
  //
  DOMElement *commandsequences = XmlFile::findElement (root, "CommandSequences");
  if (commandsequences == NULL)
    return NULL;
  //
  // Set the head of the list to null
  //
  CommandSequence *head = NULL;
  CommandSequence *current = NULL;
  //
  // Now create a  object from each of the children of this node
  //
  for (DOMNode *child = commandsequences->getFirstChild(); child != 0; child=child->getNextSibling())
    {
      //
      // If the child is a "CommandSequence" element, findElement() will just
      // return that element; otherwise it returns NULL.
      //
      DOMElement* tmp = XmlFile::findElement (child, "CommandSequence");
      if(tmp) {
        //
        // Convert this Element into a CommandSequence object.  We throw away
        // the const'ness, since we need to modify its next pointer (which
        // still preserves the logical constness of the object).
        //
        CommandSequence *csn = 
          const_cast<CommandSequence *>(getCommandSequence (tmp));
        //
        // If the list head is null, then set the current node to be the list
        // head; otherwise, the setNext() function adds the node to the tail.
        //
        if (head == NULL)
          head = csn;
        else
          current->setNext (csn);
        //
        // And set the pointer for the next time around
        //
        current = csn;
      }
    }
  return head;
}


/**
 * Find all the Event elements and return a linked list of 
 * Event objects with all their information.  It will return null
 * if there is no command sequence information.
 */
Event *UPDFFontParser::findEvents (const DOMNode *root) const
{
  //
  // First get the Events node
  //
  DOMElement *events = XmlFile::findElement (root, "Events");
  if (events == NULL)
    return NULL;
  //
  // Set the head of the list to null
  //
  Event *head = NULL;
  Event *current = NULL;
  //
  // Now create a  object from each of the children of this node
  //
  for (DOMNode *child = events->getFirstChild(); child != 0; child=child->getNextSibling())
    {
      //
      // If the child is a "Event" element, findElement() will just
      // return that element; otherwise it returns NULL.
      //
      DOMElement* tmp = XmlFile::findElement (child, "Event");
      if(tmp) {
        //
        // Convert this Element into a Event object.  We throw away the
        // const'ness, since we need to modify its next pointer (which still
        // preserves the logical constness of the object).
        //
        Event *ev = const_cast<Event *>(getEvent (tmp));
        //
        // If the list head is null, then set the current node to be the list
        // head; otherwise, the setNext() function adds the node to the tail.
        //
        if (head == NULL)
          head = ev;
        else
          current->setNext (ev);
        //
        // And set the pointer for the next time around
        //
        current = ev;
      }
    }
  return head;
}


/**
 * Find all the Parameter elements and return a linked list of 
 * Parameter objects with all their information.  It will return null
 * if there is no command sequence information.
 */
Parameter *UPDFFontParser::findParameters (const DOMNode *root) const
{
    //
    // First get the Parameters node
    //
    DOMElement *parm = XmlFile::findElement (root, "Parameters");
    if (parm == NULL)
        return NULL;
    //
    // Set the head of the list to null
    //
    Parameter *head = NULL;
    Parameter *current = NULL;
    //
    // Now create a  object from each of the children of this node
    //
    for (DOMNode *child = parm->getFirstChild(); child != 0; child=child->getNextSibling())
    {
        //
        // If the child is a "Parameter" element, findElement() will just
        // return that element; otherwise it returns NULL.
        //
        DOMElement* tmp = XmlFile::findElement (child, "Parameter");
        if(tmp) {
            //
            // Convert this Element into a Parameter object.  We throw away the
            // const'ness, since we need to modify its next pointer (which still
            // preserves the logical constness of the object).
            //
            Parameter *p = const_cast<Parameter *>(getParameter (tmp));
            //
            // If the list head is null, then set the current node to be the list
            // head; otherwise, the setNext() function adds the node to the tail.
            //
            if (head == NULL)
                head = p;
            else
                current->setNext (p);
            //
            // And set the pointer for the next time around
            //
            current = p;
        }
    }
    return head;
}


/**
 * Convert an XML DOM Element into a DeviceFont object containing a new
 * DeviceFont object.  The given element must be a "DeviceFont" element.
 *
 * @param element  the DeviceFont element
 */
DeviceFont *UPDFFontParser::
getDeviceFont (const DOMElement *element) const
{
#ifdef DEBUG
cout << "UPDFFontParser::getDeviceFont(" << element << ") called" << endl;
#endif
  //
  // Initialize the resulting DeviceFont object
  //
  DeviceFont* deviceFont = new DeviceFont();
  //
  // Get the string parameters and fill them in the DeviceFont
  //
  deviceFont->pszID = getAttribute (element, "ID");
  deviceFont->pszFont_Vendor = getAttribute (element, "Font_Vendor");
  deviceFont->pszEncoding = getAttribute (element, "Encoding");
  deviceFont->pszFontCommandSequenceID = 
    getAttribute (element, "Font_Command_Sequence_ID");
  deviceFont->pszPassive_Font = getAttribute (element, "Passive_Font");
  //
  // Convert the font family string into an enumeration
  //
  char *fontfamily = getAttribute (element, "Font_Family");
  deviceFont->fontFamily = DeviceFont::getFontFamily (fontfamily);
  delete [] fontfamily;

  //
  // Move to Global Metrics, Panose and Name ID Node
  //
  DOMNode* grandChild;
  for (grandChild = element->getFirstChild(); 
       grandChild != 0; 
       grandChild=grandChild->getNextSibling()) {

    const char* nodeName = XMLString::transcode(grandChild->getNodeName());
    // Global metrics
    if(strcmp(nodeName, "GlobalMetrics") == 0) {
      getGlobalMetrics(deviceFont, (DOMElement*)grandChild); 
    }
    // Panose
    else if(strcmp(nodeName, "Panose") == 0) {
      getPanose(deviceFont, (DOMElement*)grandChild);
    }
    // Name IDs
    else if(strcmp(nodeName, "NameIDs") == 0) {
      getNameIDs(deviceFont, (DOMElement*)grandChild);
    }
    delete [] nodeName;
  }
  //
  // And return the result
  //
  return deviceFont;
}


/**
 * Convert an XML DOM Element into a CommandSequence object containing a pair
 * of strings with the name and attribute value.  The given element must be
 * a "CommandSequence" element.
 *
 * @param element  pointer to a "CommandSequence" element
 */
CommandSequence *UPDFFontParser::getCommandSequence (const DOMElement *element) const
{
  //
  // Extract the strings and create the object
  //
  char* id = getAttribute (element, "CommandSequence_ID");
  char* value = getAttribute (element, "CommandSequence");
  return new CommandSequence (id, value);
}


/**
 * Convert an XML DOM Element into an Event object containing the strings.
 * The given element must be an "Event" element.
 *
 * @param element  pointer to an "Event" element
 */
Event *UPDFFontParser::getEvent (const DOMElement *element) const
{
  char * id = getAttribute (element, "ID");
  char * startend = getAttribute (element, "Event_StartEnd");
  char * type = getAttribute (element, "Event_ID");
  char * cmdseq = getAttribute (element, "CommandSequence_ID");
  return new Event (id, startend, type, cmdseq);
}


/**
* Convert an XML DOM Element into an Parameter object containing the strings.
* The given element must be an "Parameter" element.
*
* @param element  pointer to an "Parameter" element
*/
Parameter *UPDFFontParser::getParameter (const DOMElement *element) const
{
    char * id = getAttribute (element, "Parameter_ID");
    char * value = getAttribute (element, "Parameter");
    return new Parameter (id, value);
}


/**
 * Extract the global metrics information and store in the given DeviceFont
 * object.  The indicated DOM Element must be "GlobalMetrics" node.
 *
 * @param deviceFont  the object to modify
 * @param element     the "GlobalMetrics" DOM element for the font
 */
void UPDFFontParser::getGlobalMetrics(DeviceFont* deviceFont, const DOMElement* element) const
{
#ifdef DEBUG
cout << "UPDFFontParser::getGlobalMetrics(" << deviceFont << "," << element << ") called" << endl;
#endif
  deviceFont->iGlobal_Units_Per_Em = getInt (element, "Units_Per_Em");
  deviceFont->iGlobal_Points_Per_Inch = getInt (element, "Points_Per_Inch");
  deviceFont->iGlobal_Spacing = getInt (element, "Spacing");
  deviceFont->iGlobal_Minimum_Height = getInt (element, "Minimum_Height");
  deviceFont->iGlobal_Maximum_Height = getInt (element, "Maximum_Height");
  deviceFont->iGlobal_HeightStep = getInt (element, "HeightStep");
  deviceFont->iGlobal_Slant = getInt (element, "Slant");
  deviceFont->iGlobal_WindowsWeight = getInt (element, "WindowsWeight");
  deviceFont->iGlobal_Windows_Ascender = getInt (element, "Windows_Ascender");
  deviceFont->iGlobal_Windows_Descender = getInt (element, "Windows_Descender");
  deviceFont->iGlobal_Mac_Ascender = getInt (element, "Mac_Ascender");
  deviceFont->iGlobal_Mac_Descender = getInt (element, "Mac_Descender");
  deviceFont->iGlobal_Mac_Line_Gap = getInt (element, "Mac_Line_Gap");
  deviceFont->iGlobal_Typographic_Ascender = getInt (element, "Typographic_Ascender");
  deviceFont->iGlobal_Typographic_Descender = getInt (element, "Typographic_Descender");
  deviceFont->iGlobal_Typographic_Line_Gap = getInt (element, "Typographic_Line_Gap");
  deviceFont->iGlobal_Average_Character_Width = getInt (element, "Average_Character_Width");
  deviceFont->iGlobal_Maximum_Character_Increment = getInt (element, "Maximum_Character_Increment");
  deviceFont->iGlobal_Cap_Height = getInt (element, "Cap_Height");
  deviceFont->iGlobal_x_Height = getInt (element, "x_Height");
  deviceFont->iGlobal_Subscript_X_Size = getInt (element, "Subscript_X_Size");
  deviceFont->iGlobal_Subscript_Y_Size = getInt (element, "Subscript_Y_Size");
  deviceFont->iGlobal_Subscript_X_Offset = getInt (element, "Subscript_X_Offset");
  deviceFont->iGlobal_Subscript_Y_Offset = getInt (element, "Subscript_Y_Offset");
  deviceFont->iGlobal_Superscript_X_Size = getInt (element, "Superscript_X_Size");
  deviceFont->iGlobal_Superscript_Y_Size = getInt (element, "Superscript_Y_Size");
  deviceFont->iGlobal_Superscript_X_Offset = getInt (element, "Superscript_X_Offset");
  deviceFont->iGlobal_Superscript_Y_Offset = getInt (element, "Superscript_Y_Offset");
  deviceFont->iGlobal_Underscore_Size = getInt (element, "Underscore_Size");
  deviceFont->iGlobal_Underscore_Offset = getInt (element, "Underscore_Offset");
  deviceFont->iGlobal_Underscore_Absolute_Values = getInt (element, "Underscore_Absolute_Values");
  deviceFont->iGlobal_Strikeout_Size = getInt (element, "Strikeout_Size");
  deviceFont->iGlobal_Strikeout_Offset = getInt (element, "Strikeout_Offset");
  deviceFont->iGlobal_Strikeout_Absolute_Values = getInt (element, "Strikeout_Absolute_Values");
  deviceFont->iGlobal_Baseline = getInt (element, "Baseline");
  deviceFont->iGlobal_Aspect = getInt (element, "Aspect");
  deviceFont->iGlobal_Caret = getInt (element, "Caret");
  deviceFont->iGlobal_Orientation = getInt (element, "Orientation");
  deviceFont->iGlobal_Degree = getInt (element, "Degree");
  deviceFont->iGlobal_CharacterWidth_A_Value = getInt (element, "CharacterWidth_A_Value");
  deviceFont->iGlobal_CharacterWidth_B_Value = getInt (element, "CharacterWidth_B_Value");
  deviceFont->iGlobal_CharacterWidth_C_Value = getInt (element, "CharacterWidth_C_Value");
}


/**
 * Get the Panose information from the given XML DOM Element and store in
 * the given DeviceFont object.  The Element must be a "Panose" Element.
 *
 * @param deviceFont  the object to modify
 * @param element     the "Panose" DOM element for the font
 */
void UPDFFontParser::getPanose(DeviceFont* deviceFont, const DOMElement* element) const
{
#ifdef DEBUG
cout << "UPDFFontParser::getPanose(" << deviceFont << "," << element << ") called" << endl;
#endif

  XMLCh* str = XMLString::transcode("Panose_Family_Type");
  char* familyType = XMLString::transcode(element->getAttribute(str));
  if (familyType)
  {
    deviceFont->panose.familyType = Panose::getFamilyType (familyType);
    delete [] familyType;
  }
  delete str;

  str = XMLString::transcode("Panose_Serif_Style");
  char* serifStyle = XMLString::transcode(element->getAttribute(str));
  if (serifStyle)
  {
    deviceFont->panose.serifStyle = Panose::getSerifStyle (serifStyle);
    delete [] serifStyle;
  }
  delete str;

  str = XMLString::transcode("Panose_Weight");
  char* weight = XMLString::transcode(element->getAttribute(str));
  if (weight)
  {
    deviceFont->panose.weight = Panose::getWeight(weight);
    delete [] weight;
  }
  delete str;

  str = XMLString::transcode("Panose_Proportion");
  char* proportion = XMLString::transcode(element->getAttribute(str));
  if (proportion)
  {
    deviceFont->panose.proportion = Panose::getProportion(proportion);
    delete [] proportion;
  }
  delete str;

  str = XMLString::transcode("Panose_Contrast");
  char* contrast = XMLString::transcode(element->getAttribute(str));
  if (contrast)
  {
    deviceFont->panose.contrast = Panose::getContrast (contrast);
    delete [] contrast;
  }
  delete str;

  str = XMLString::transcode("Panose_Stroke_Variation");
  char* stroke = XMLString::transcode(element->getAttribute(str));
  if (stroke)
  {
    deviceFont->panose.strokeVariation = Panose::getStrokeVariation(stroke);
    delete [] stroke;
  }
  delete str;

  str = XMLString::transcode("Panose_Arm_Style");
  char* armStyle = XMLString::transcode(element->getAttribute(str));
  if (armStyle)
  {
    deviceFont->panose.armStyle = Panose::getArmStyle(armStyle);
    delete [] armStyle;
  }
  delete str;

  str = XMLString::transcode("Panose_Letterform");
  char* letterform = XMLString::transcode(element->getAttribute(str));
  if (letterform)
  {
    deviceFont->panose.letterform = Panose::getLetterform(letterform);
    delete [] letterform;
  }
  delete str;

  str = XMLString::transcode("Panose_Midline");
  char* midline = XMLString::transcode(element->getAttribute(str));
  if (midline)
  {
    deviceFont->panose.midline = Panose::getMidline(midline);
    delete [] midline;
  }
  delete str;

  str = XMLString::transcode("Panose_X_Height");
  char* xHeight = XMLString::transcode(element->getAttribute(str));
  if (xHeight)
  {
    deviceFont->panose.xHeight = Panose::getXHeight(xHeight);
    delete [] xHeight;
  }
  delete str;
}


/**
 * Extract the NameID information from the given XML DOM Element.  The Element
 * must be a "NameIDs" node.
 *
 * @param deviceFont  the object to modify
 * @param element     the "NameIDs" DOM element for the font
 */
void UPDFFontParser::getNameIDs(DeviceFont* deviceFont,
                                const DOMElement* element) const
{
#ifdef DEBUG
cout << "UPDFFontParser::getNameIDs(" << deviceFont << "," << element << ") called" << endl;
#endif
  DOMNode* idChild;
  XMLCh* str = XMLString::transcode("Name_ID");

  for (idChild = element->getFirstChild(); idChild != 0; idChild=idChild->getNextSibling()) {
    
    const char* nodeName = XMLString::transcode(idChild->getNodeName());

    if(strcmp(nodeName, "FaceName") == 0) {
      char* faceNameId = XMLString::transcode(((DOMElement*)idChild)->getAttribute(str));
      deviceFont->pszFaceName_Name_ID = faceNameId;
    }

    if(strcmp(nodeName, "FamilyName") == 0) {
      char* familyNameId = XMLString::transcode(((DOMElement*)idChild)->getAttribute(str));
      deviceFont->pszFamilyName_Name_ID = familyNameId;
    }

    if(strcmp(nodeName, "UniqueName") == 0) {
      char* uniqueNameId = XMLString::transcode(((DOMElement*)idChild)->getAttribute(str));
      deviceFont->pszUniqueName_Name_ID = uniqueNameId;
    }

    if(strcmp(nodeName, "SlantName") == 0) {
      char* slantNameId = XMLString::transcode(((DOMElement*)idChild)->getAttribute(str));
      deviceFont->pszSlantName_Name_ID = slantNameId;
    }

    if(strcmp(nodeName, "WeightName") == 0) {
      char* weightNameId = XMLString::transcode(((DOMElement*)idChild)->getAttribute(str));
      deviceFont->pszWeightName_Name_ID = weightNameId;
    }
    //
    // And clean up
    //
    delete [] nodeName;
  }
  //
  // And clean up
  //
  delete str;
}


/**
 * A convenience method that reads an integer attribute from the given
 * XML DOM Element and stores that value in the given integer.  Stores
 * a 0 if the attribute doesn't exist or is blank.
 *
 * @param element  XML DOM Element that contains the attribute
 * @param attr     name of the attribute whose value contains an integer
 */
int UPDFFontParser::getInt (const DOMElement *element, const char *attr) const
{
  //
  // Initialize the return value
  //
  int ivalue = 0;
  //
  // Convert the attribute name to an XMLCh
  //
  XMLCh* str = XMLString::transcode(attr);
  //
  // Convert the resulting element value to a normal character array
  //
  char *cval = XMLString::transcode(element->getAttribute(str));
  delete str;
  if (cval == NULL)
    return 0;
  //
  // Check to see if the value is not blank
  //
  if (strlen(cval) > 0)
    {
      //
      // Convert to an integer and store in the indicate memory
      //
      sscanf(cval, "%d", &ivalue);
    }
  //
  // Clean up
  //
  delete [] cval;
  return ivalue;
}


/**
 * Get the attribute of the element as an array of UTF-8 characters.  The
 * caller is responsible for delete[]ing this array.
 *
 * @param e     DOM Element to get the attribute from
 * @param attr  name of the attribute
 */
char * UPDFFontParser::getAttribute (const DOMElement *e, const char *attr) const
{
  XMLCh* str = XMLString::transcode(attr);
  //
  // Get the attribute value as a UTF-16 string
  //
  const XMLCh* value = e->getAttribute(str);

  //
  // Get the length of the UTF-16 string
  //
  int utf16len = XMLString::stringLen(value);
  //
  // In the worst case, there are four bytes per character (plus one
  // byte for the null byte)
  //
  unsigned int maxlen = 1 + 4 * utf16len;
  unsigned char *buffer = new unsigned char[maxlen];
  //char *result = XMLString::transcode(value);
  //
  // Convert to UTF-8
  //
  unsigned int eaten;
  transcoder->transcodeTo (static_cast<const XMLCh *>(value), 1+utf16len,
                           static_cast<XMLByte *>(buffer), maxlen,
                           eaten, XMLTranscoder::UnRep_Throw);
  //
  // For efficiency, we should copy the attribute into a new char *,
  // since it is currently in a 1024-byte buffer.
  //
  delete str;
  return (char *)buffer;
}


/**
 * Return a linked list of DeviceFont objects which contain DeviceFont objects.
 * We have to make a copy of our internal list, and the caller is responsible
 * for deleting this new list.
 */
DeviceFont *UPDFFontParser::getFonts() const
{
  if (myFonts == NULL)
    return NULL;
  //
  // The list head that we will return
  //
  DeviceFont *head = new DeviceFont(*myFonts);
  DeviceFont *current = head;
  //
  // Copy all the elements in the linked list
  //
  const DeviceFont *source = myFonts->getNext();
  while (source != NULL)
    {
      DeviceFont *acopy = new DeviceFont (*source);
      current->setNext(acopy);
      current = acopy;
      source = source->getNext();
    }
  //
  // And return the new list we just created
  //
  return head;
} 


/**
 * Return a linked list of CommandSequence objects which contain strings with
 * the command sequence name and value.  We have to make a copy of our
 * internal list, and the caller is responsible for deleting this new list.
 */
CommandSequence *UPDFFontParser::getCmdSeqs() const
{
  if (myCmdSeqs == NULL)
    return NULL;
  //
  // The list head that we will return
  //
  CommandSequence *head = new CommandSequence(*myCmdSeqs);
  CommandSequence *current = head;
  //
  // Copy all the elements in the linked list
  //
  const CommandSequence *source = myCmdSeqs->getNext();
  while (source != NULL)
    {
      CommandSequence *acopy = new CommandSequence (*source);
      current->setNext(acopy);
      current = acopy;
      source = source->getNext();
    }
  //
  // And return the new list we just created
  //
  return head;
}


/**
 * Return a linked list of Event objects.  We have to make a copy of our
 * internal list, and the caller is responsible for deleting this new list.
 */
Event *UPDFFontParser::getEvents() const
{
  if (myEvents == NULL)
    return NULL;
  //
  // The list head that we will return
  //
  Event *head = new Event(*myEvents);
  Event *current = head;
  //
  // Copy all the elements in the linked list
  //
  const Event *source = myEvents->getNext();
  while (source != NULL)
    {
      Event *acopy = new Event (*source);
      current->setNext(acopy);
      current = acopy;
      source = source->getNext();
    }
  //
  // And return the new list we just created
  //
  return head;
}


/**
* Return a linked list of Parameter objects.  We have to make a copy of our
* internal list, and the caller is responsible for deleting this new list.
*/
Parameter *UPDFFontParser::getParameters() const
{
    if (myParameters == NULL)
        return NULL;
    //
    // The list head that we will return
    //
    Parameter *head = new Parameter(*myParameters);
    Parameter *current = head;
    //
    // Copy all the elements in the linked list
    //
    const Parameter *source = myParameters->getNext();
    while (source != NULL)
    {
        Parameter *acopy = new Parameter (*source);
        current->setNext(acopy);
        current = acopy;
        source = source->getNext();
    }
    //
    // And return the new list we just created
    //
    return head;
}
