/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <UPDFDeviceForm.hpp>

/*
*/
static int     getOmniFormID             (xmlNodePtr      node);

UPDFDeviceForm::
UPDFDeviceForm (Device      *pDevice,
                int          id,
                int          iCapabilities,
                BinaryData  *data,
                HardCopyCap *hcInfo,
                xmlNodePtr   node)
   : DeviceForm (pDevice,
                 id,
                 iCapabilities,
                 data,
                 hcInfo)
{
   node_d = node;
}

UPDFDeviceForm::
~UPDFDeviceForm ()
{
}

DeviceForm * UPDFDeviceForm::
createV (int id)
{
   return UPDFDeviceForm::create (pDevice_d, id);
}

DeviceForm * UPDFDeviceForm::
create (Device *pDevice, PSZCRO pszId)
{
   int id = DeviceForm::nameToID (pszId);

   if (DeviceForm::FORM_UNLISTED != id)
   {
      return create (pDevice, id);
   }
   else
   {
      UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);
      xmlNodePtr  nodeForms   = 0;

      nodeForms = findForms (pUPDFDevice);

      if (nodeForms)
      {
         xmlNodePtr nodeItem  = firstNode (nodeForms->xmlChildrenNode);
         xmlNodePtr nodeFound = 0;

         while (nodeItem != 0)
         {
            PSZCRO pszNodeId = (PSZCRO)xmlGetProp (nodeItem, (xmlChar *)"ID");

            if (pszNodeId)
            {
               if (0 == strcmp (pszNodeId, pszId))
               {
                  nodeFound = nodeItem;
               }

               free ((void *)pszNodeId);
            }

            if (nodeFound)
            {
               id = getOmniFormID (nodeFound);

               if (DeviceForm::FORM_UNLISTED != id)
               {
                  return create (pDevice, id);
               }

               nodeFound = 0;
            }

            nodeItem = nextNode (nodeItem);
         }
      }

      return 0;
   }
}

static int
getOmniFormID (xmlNodePtr node)
{
   int id = DeviceForm::FORM_UNLISTED;

   if (!node)
   {
      return id;
   }

   PSZCRO pszID = (PSZCRO)xmlGetProp (node, (xmlChar *)"Predefined_ID");

   if (pszID)
   {
      id = DeviceForm::nameToID (pszID);

      free ((void *)pszID);
   }

   return id;
}

static xmlNodePtr
skipInvalidForms (xmlNodePtr node)
{
   if (!node)
   {
      return 0;
   }

   int id = getOmniFormID (node);

   while (  node
           && (  id == DeviceForm::FORM_UNLISTED
              || id == DeviceForm::FORM_NONE
              )
           )
   {
      node = nextNode (node);
      id   = getOmniFormID (node);
   }

   return node;
}

static int
converUPDFMarginToOmniMargin (PSZCRO pszMarginInfo)
{
   float flMargin = 0.0;
   int   iReturn  = -1;

   if (1 == sscanf (pszMarginInfo, "%f", &flMargin))
   {
      PSZCRO pszEnd = pszMarginInfo + strlen (pszMarginInfo);

      if (0 == strcmp (pszEnd - 2, "in"))
      {
         iReturn = (int)(flMargin * 25400.0);
      }
      else if (0 == strcmp (pszEnd - 2, "mm"))
      {
         iReturn = (int)(flMargin * 1000.0);
      }
      else
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputUPDFDeviceForm ()) DebugOutput::getErrorStream () << "UPDFDeviceForm::" << __FUNCTION__ << " Unknown margin encoding!" << std::endl;
#endif
      }
   }

   return iReturn;
}

static DeviceForm *
createFromXMLNode (Device      *pDevice,
                   xmlNodePtr   node,
                   int          id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   int                iCapabilities = 0;
   BinaryData        *data          = 0;
   int                iTopMargin    = 0;
   int                iBottomMargin = 0;
   int                iLeftMargin   = 0;
   int                iRightMargin  = 0;
   xmlNodePtr         nodeMargins   = 0;
   DeviceOrientation *pDO           = pUPDFDevice->getCurrentOrientation ();

   switch (pDO->getID ())
   {
   case DeviceOrientation::ORIENTATION_LANDSCAPE:
   case DeviceOrientation::ORIENTATION_REVERSE_LANDSCAPE:
   {
      nodeMargins = FINDUDRENTRY (pUPDFDevice,
                                  node,
                                  "MediaSizeHardwareMarginsLandscape",
                                  UPDFDeviceForm);
      break;
   }
   }

   if (!nodeMargins)
      nodeMargins = FINDUDRENTRY (pUPDFDevice,
                                  node,
                                  "MediaSizeHardwareMargins",
                                  UPDFDeviceForm);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceForm ()) DebugOutput::getErrorStream () << "UPDFDeviceForm::" << __FUNCTION__ << ": nodeMargins = " << nodeMargins << std::endl;
#endif

   if (nodeMargins)
   {
      PSZCRO pszTopMargin    = (PSZCRO)xmlGetProp (nodeMargins, (xmlChar *)"Top");
      PSZCRO pszBottomMargin = (PSZCRO)xmlGetProp (nodeMargins, (xmlChar *)"Bottom");
      PSZCRO pszLeftMargin   = (PSZCRO)xmlGetProp (nodeMargins, (xmlChar *)"Left");
      PSZCRO pszRightMargin  = (PSZCRO)xmlGetProp (nodeMargins, (xmlChar *)"Right");

      iTopMargin    = converUPDFMarginToOmniMargin (pszTopMargin);
      iBottomMargin = converUPDFMarginToOmniMargin (pszBottomMargin);
      iLeftMargin   = converUPDFMarginToOmniMargin (pszLeftMargin);
      iRightMargin  = converUPDFMarginToOmniMargin (pszRightMargin);

      switch (pDO->getID ())
      {
      case DeviceOrientation::ORIENTATION_REVERSE_PORTRAIT:
      case DeviceOrientation::ORIENTATION_REVERSE_LANDSCAPE:
      {
         int iSwap;

         iSwap         = iBottomMargin;
         iBottomMargin = iTopMargin;
         iTopMargin    = iSwap;
         iSwap         = iLeftMargin;
         iLeftMargin   = iRightMargin;
         iRightMargin  = iSwap;
         break;
      }
      }

      if (pszTopMargin)
      {
         free ((void *)pszTopMargin);
      }
      if (pszBottomMargin)
      {
         free ((void *)pszBottomMargin);
      }
      if (pszLeftMargin)
      {
         free ((void *)pszLeftMargin);
      }
      if (pszRightMargin)
      {
         free ((void *)pszRightMargin);
      }
   }

   return new UPDFDeviceForm (pUPDFDevice,
                              id,
                              iCapabilities,
                              data,
                              new HardCopyCap (iLeftMargin,
                                               iTopMargin,
                                               iRightMargin,
                                               iBottomMargin),
                              node);
}

DeviceForm * UPDFDeviceForm::
create (Device *pDevice, int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceForm ()) DebugOutput::getErrorStream () << "UPDFDeviceForm::" << __FUNCTION__ << ": pDevice = " << pDevice << ", id = " << id << std::endl;
#endif

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeForms = 0;

   nodeForms = findForms (pUPDFDevice);

   if (nodeForms)
   {
      xmlNodePtr nodeItem = firstNode (nodeForms->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iFormID = getOmniFormID (nodeItem);

         if (  id == iFormID
            && !(  id == DeviceForm::FORM_UNLISTED
                || id == DeviceForm::FORM_NONE
                )
            )
         {
            return createFromXMLNode (pDevice, nodeItem, id);
         }

         nodeItem = nextNode (nodeItem);
      }

      return 0;
   }
   else
   {
      return 0;
   }
}

PSZCRO UPDFDeviceForm::
getDeviceID ()
{
   return 0;
}

bool UPDFDeviceForm::
isFormSupported (int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeForms = 0;

   nodeForms = findForms (pUPDFDevice);

   if (nodeForms)
   {
      xmlNodePtr nodeItem = firstNode (nodeForms->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iFormID = getOmniFormID (nodeItem);

         if (  id == iFormID
            && !(  id == DeviceForm::FORM_UNLISTED
                || id == DeviceForm::FORM_NONE
                )
            )
         {
            return true;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return false;
}

class UPDFForm_Enumerator : public Enumeration
{
public:
   UPDFForm_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      pDevice_d  = pDevice;
      nodeItem_d = skipInvalidForms (nodeItem);
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      void *pvRet = 0;

      if (nodeItem_d)
      {
         int id = getOmniFormID (nodeItem_d);

         pvRet = createFromXMLNode (pDevice_d,
                                    nodeItem_d,
                                    id);

         nodeItem_d = nextNode (nodeItem_d);
         nodeItem_d = skipInvalidForms (nodeItem_d);
      }

      return pvRet;
   }

private:
   Device     *pDevice_d;
   xmlNodePtr  nodeItem_d;
};

Enumeration * UPDFDeviceForm::
getEnumeration ()
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDFForm_Enumerator (pDevice_d, 0);
   }

   xmlNodePtr nodeForms = 0;

   nodeForms = findForms (pUPDFDevice);

   if (nodeForms)
   {
      xmlNodePtr nodeItem = firstNode (nodeForms->xmlChildrenNode);

      return new UPDFForm_Enumerator (pDevice_d, nodeItem);
   }
   else
   {
      return new UPDFForm_Enumerator (pDevice_d, 0);
   }
}

xmlNodePtr UPDFDeviceForm::
findForms (UPDFDevice *pUPDFDevice)
{
   xmlNodePtr nodeForms = 0;

   if (!pUPDFDevice)
   {
      return nodeForms;
   }

   if (  ((nodeForms = FINDUDRENTRY (pUPDFDevice, nodeForms, "Device_Features",    UPDFDeviceForm)) != 0)
      && ((nodeForms = FINDUDRENTRY (pUPDFDevice, nodeForms, "PrintCapabilities",  UPDFDeviceForm)) != 0)
      && ((nodeForms = FINDUDRENTRY (pUPDFDevice, nodeForms, "Features",           UPDFDeviceForm)) != 0)
      && ((nodeForms = FINDUDRENTRY (pUPDFDevice, nodeForms, "PrintMediaHandling", UPDFDeviceForm)) != 0)
      && ((nodeForms = FINDUDRENTRY (pUPDFDevice, nodeForms, "MediaSize",          UPDFDeviceForm)) != 0)
      )
      return nodeForms;

   return nodeForms;
}

std::string UPDFDeviceForm::
toString (std::ostringstream& oss)
{
   std::ostringstream oss2;

   oss << "{"
       << DeviceForm::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
std::ostream&
operator<< (std::ostream& os, const UPDFDeviceForm& const_self)
{
   UPDFDeviceForm&    self = const_cast<UPDFDeviceForm&>(const_self);
   std::ostringstream oss;

   os << self.toString (oss);

   return os;
}
