/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "UPDFDeviceInstance.hpp"
#include "defines.hpp"

#include <stdio.h>
#include <string.h>

#include <iostream>
#include <sstream>

const static bool fTestNoCompression = true;

UPDFDeviceInstance::
UPDFDeviceInstance (PrintDevice *pDevice)
   : DeviceInstance (pDevice)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " ()" << std::endl;
#endif

   fHaveInitialized_d  = false;
   fHaveSetupPrinter_d = false;
   ptlPrintHead_d.x    = 0;
   ptlPrintHead_d.y    = 0;

   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (pUPDFDevice)
   {
      loadNonDominantDefaults (pUPDFDevice, pUPDFDevice->getRootUDR ());
   }
   loadLocaleDefaults ();
}

UPDFDeviceInstance::
~UPDFDeviceInstance ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::~" << __FUNCTION__ << " ()" << std::endl;
#endif

#ifdef HAVE_SYMBOL_TABLE
   for ( SYMBOLTABLE::iterator next = symbolTable_d.begin () ;
         next != symbolTable_d.end () ;
         next++ )
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::~" << __FUNCTION__ << ": deleting symbol table entry" << std::endl;
#endif
      delete (*next).second;
   }
#endif
}

void UPDFDeviceInstance::
initializeInstance ()
{
   if (fHaveInitialized_d)
      return;

   fHaveInitialized_d = true;

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::initializeInstance ()" << std::endl;
#endif
}

void UPDFDeviceInstance::
loadNonDominantDefaults (UPDFDevice *pUPDFDevice, xmlNodePtr node)
{
   if (!node)
      return;

   node = firstNode (node);

   while (node)
   {
      if (node->xmlChildrenNode)
      {
         loadNonDominantDefaults (pUPDFDevice, node->xmlChildrenNode);
      }

      PSZCRO pszNDR = (PSZCRO )xmlGetProp (node, (xmlChar *)"Non_Dominant_Representative");

      if (pszNDR)
      {
         PSZCRO pszID = (PSZCRO )xmlGetProp (node, (xmlChar *)"ID");

#ifndef RETAIL
         if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << ": pszNDR = " << pszNDR << ", pszID = " << pszID << std::endl;
#endif

         if (pszID)
         {
            xmlNodePtr elm = pUPDFDevice->findEntryKeyValue (node,
                                                             "ID",
                                                             pszNDR,
                                                             DebugOutput::shouldOutputUPDFDeviceInstance ());

            if (elm)
            {
               std::string stringID = pszID;

               xmlObjectMap_d[stringID] = elm;
            }

            free ((void *)pszID);
         }

         free ((void *)pszNDR);
      }

      node = nextNode (node);
   }
}

void UPDFDeviceInstance::
loadLocaleDefaults ()
{
   UPDFDevice *pUPDFDevice       = UPDFDevice::isAUPDFDevice (pDevice_d);
   xmlNodePtr  nodeLocalDefaults = 0;

   if (!pUPDFDevice)
      return;

   if ((nodeLocalDefaults = FINDLOCALEENTRY (pUPDFDevice, nodeLocalDefaults, "LocaleDefaults", UPDFDeviceInstance)) != 0)
   {
      xmlNodePtr nodeItem = firstNode (nodeLocalDefaults->xmlChildrenNode);

      while (nodeItem != 0)
      {
         PSZCRO pszFeature = (PSZCRO )xmlGetProp (nodeItem, (xmlChar *)"Default_Feature");
         PSZCRO pszSetting = (PSZCRO )xmlGetProp (nodeItem, (xmlChar *)"Default_Setting");

         if (  pszFeature
            && pszSetting
            )
         {
            std::string stringFeature = pszFeature;
            xmlNodePtr  elm           = 0;

#ifndef RETAIL
            if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << ": pszFeature = " << pszFeature << ", pszSetting = " << pszSetting << std::endl;
#endif

            elm = FINDUDRENTRYKEYVALUE (pUPDFDevice, "ID", pszSetting, UPDFDeviceInstance);

            if (elm)
            {
               xmlObjectMap_d[stringFeature] = elm;
            }
            else
            {
               std::string stringSetting = pszSetting;

               xmlObjectMap_d[stringFeature] = 0;
               xmlStringMap_d[stringFeature] = stringSetting;

#ifndef RETAIL
               if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << ": Warning: Placing " << pszFeature << "=" << pszSetting << " in string map!" << std::endl;
#endif
            }
         }

         if (pszFeature)
         {
            free ((void *)pszFeature);
         }
         if (pszSetting)
         {
            free ((void *)pszSetting);
         }

         nodeItem = nextNode (nodeItem);
      }
   }

}

#ifdef HAVE_SYMBOL_TABLE
static PSZCRO apszDeviceJobPropertyKeys[] = {
   "RandomAccessMemory.RAM_KB",
   "DeviceResolution.HorizontalResolution",
   "DeviceResolution.VerticalResolution",
   "RasterGraphic.HorizontalResolution",
   "RasterGraphic.VerticalResolution",
   "Positioning.HorizontalResolution",
   "Positioning.VerticalResolution",
   "MediaCopy.Media_Copy_ID",
   "MediaCollation.MediaCollation_Global_ID",
   "MediaCollation.Media_Proprietary_ID",
   "MediaPageOrientation.MediaPageOrientation_Global_ID",
   "MediaSize.MediaSize_Global_ID",
   "MediaSize.Media_Proprietary_ID",
   "MediaType.MediaType_Global_ID",
   "MediaType.Media_Proprietary_ID",
   "MediaSource.MediaSource_Global_ID",
   "MediaSource.Media_Proprietary_ID",
   "MediaDestination.MediaDestination_Global_ID",
   "MediaDestination.Media_Proprietary_ID",
   "MediaCoating.MediaCoating_Global_ID",
   "MediaCoating.Media_Proprietary_ID",
   "MediaColor.MediaColor_Global_ID",
   "MediaColor.Media_Proprietary_ID",
   "MediaWeight.MediaWeight_Global_ID",
   "MediaWeight.Media_Proprietary_ID",
   "MediaStaplingSet.Media_Proprietary_ID",
   "MediaPunchingSet.Media_Proprietary_ID",
   "MediaJogging.MediaJogging_Global_ID",
   "MediaJogging.Media_Proprietary_ID",
   "MediaDuplex.MediaDuplex_Global_ID",
   "MediaDuplex.Media_Proprietary_ID",
   "MediaBooklet.MediaBooklet_Global_ID",
   "MediaBooklet.Media_Proprietary_ID",
   "MediaPageOrder.MediaPageOrder_Global_ID",
   "MediaPageOrder.Media_Proprietary_ID",
   "MediaNUp.MediaNUp_HorizontalPages",
   "MediaNUp.MediaNUp_VerticalPages",
   "MediaNUpBorder.MediaNUpBorder_Global_ID",
   "MediaNUpBorder.Media_Proprietary_ID",
   "MediaNUpPageOrder.MediaNUpPageOrder_Global_ID",
   "MediaNUpPageOrder.Media_Proprietary_ID",
   "MediaTargetSize.MediaTargetSize_Global_ID",
   "MediaTargetSize.Media_Proprietary_ID",
   "MediaScaling.MediaScaling_Global_ID",
   "MediaScaling.Media_Proprietary_ID",
   "MediaPercentage.Media_Percentage_ID"
};

bool UPDFDeviceInstance::
handleDeviceJobProperty (PSZCRO pszKey,
                         PSZCRO pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " ("
                          << pszKey << ", "
                          << pszValue << ")"
                          << std::endl;
#endif

   for (int i = 0; i < (int)dimof (apszDeviceJobPropertyKeys); i++)
   {
      if (0 == strcmp (pszKey, apszDeviceJobPropertyKeys[i]))
      {
         symbolTable_d[std::string (pszKey)] = new Expression (pszValue, true);

         return true;
      }
   }

   return false;
}

class DeviceJobPropertyKeyEnumerator : public Enumeration
{
public:
   DeviceJobPropertyKeyEnumerator (PSZCRO *apszKeys, int iDimOfKeys)
   {
      iWhichKey_d = iDimOfKeys;
      apszKeys_d  = apszKeys;
   }

   virtual bool hasMoreElements ()
   {
      return iWhichKey_d > 0;
   }

   virtual void * nextElement ()
   {
      if (0 < iWhichKey_d)
      {
         iWhichKey_d--;
         return apszKeys_d[iWhichKey_d];
      }

      return 0;
   }

private:
   int    iWhichKey_d;
   PSZCRO *apszKeys_d;
};

Enumeration * UPDFDeviceInstance::
listDeviceJobPropertyKeys ()
{
   return new DeviceJobPropertyKeyEnumerator (apszDeviceJobPropertyKeys, dimof (apszDeviceJobPropertyKeys));
}

std::string * UPDFDeviceInstance::
getJobPropertyType (PSZCRO pszKey)
{
   // @TBD
   return new std::string ("string bob");
}

std::string * UPDFDeviceInstance::
getJobProperty (PSZCRO pszKey)
{
   Expression  *pExpressionValue = 0;
   std::string *pStringReturn    = 0;

   pExpressionValue = symbolTable_d[std::string (pszKey)];

   if (pExpressionValue)
   {
      switch (pExpressionValue->getType ())
      {
      case TYPE_STRING:
      {
         pStringReturn = new std::string (pExpressionValue->getString ());
         break;
      }

      case TYPE_HEX:
      {
         // @TBD
/////////pExpressionValue->getBinaryData ()
         break;
      }
      }
   }

   return pStringReturn;
}

#endif

bool UPDFDeviceInstance::
handleDeviceJobProperty (PSZCRO pszKey,
                         PSZCRO pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " ("
                          << pszKey << ", "
                          << pszValue << ")"
                          << std::endl;
#endif

   if (  !pszKey
      || !*pszKey
      || !pszValue
      || !*pszValue
      )
      return false;

   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
      return false;

   xmlNodePtr  node      = 0;
   std::string stringKey = pszKey;

   node = xmlObjectMap_d[stringKey];

   if (node)
   {
      xmlNodePtr root = pUPDFDevice->findEntryKeyValue (node,
                                                        "ID",
                                                        pszKey,
                                                        DebugOutput::shouldOutputUPDFDeviceInstance ());

      if (root)
      {
         xmlNodePtr newNode = pUPDFDevice->findEntryKeyValue (root,
                                                              "ID",
                                                              pszValue,
                                                              DebugOutput::shouldOutputUPDFDeviceInstance ());

         if (newNode)
         {
            xmlObjectMap_d[stringKey] = newNode;

            return true;
         }
      }
   }

///for (XMLObjectMap::iterator next = xmlObjectMap_d.begin ();
///     next != xmlObjectMap_d.end ();
///     next++)
///{
///}

   return false;
}

class DeviceJobPropertyKeyEnumerator : public Enumeration
{
public:
   DeviceJobPropertyKeyEnumerator (XMLObjectMap::iterator first,
                                   XMLObjectMap::iterator last)
   {
      next_d = first;
      last_d = last;
   }

   virtual ~
   DeviceJobPropertyKeyEnumerator ()
   {
   }

   virtual bool
   hasMoreElements ()
   {
      return next_d != last_d;
   }

   virtual void *
   nextElement ()
   {
      void *pvRet = 0;

      if (next_d != last_d)
      {
         pvRet = (void *)next_d->first.c_str ();

         next_d++;
      }

      return pvRet;
   }

private:
   XMLObjectMap::iterator next_d;
   XMLObjectMap::iterator last_d;
};

Enumeration * UPDFDeviceInstance::
listDeviceJobPropertyKeys ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << std::endl;
#endif

   return new DeviceJobPropertyKeyEnumerator (xmlObjectMap_d.begin (),
                                              xmlObjectMap_d.end ());
}

class DeviceJobPropertyKeyValueEnumerator : public Enumeration
{
public:
   DeviceJobPropertyKeyValueEnumerator (xmlNodePtr elm)
   {
      elm_d     = 0;
      pszNext_d = 0;
      pszLast_d = 0;

      if (  elm
         && elm->xmlChildrenNode
         )
      {
         elm_d = firstNode (elm->xmlChildrenNode);

         pszNext_d = getNextID ();
      }
   }

   virtual ~
   DeviceJobPropertyKeyValueEnumerator ()
   {
      if (pszNext_d)
      {
         free ((void *)pszNext_d);
         pszNext_d = 0;
      }

      if (pszLast_d)
      {
         free ((void *)pszLast_d);
         pszLast_d = 0;
      }
   }

   virtual bool
   hasMoreElements ()
   {
      return pszNext_d != 0;
   }

   virtual void *
   nextElement ()
   {
      if (pszLast_d)
      {
         free ((void *)pszLast_d);
         pszLast_d = 0;
      }

      pszLast_d = pszNext_d;
      pszNext_d = getNextID ();

      return (void *)pszLast_d;
   }

private:
   PSZCRO
   getNextID ()
   {
      while (elm_d != 0)
      {
         PSZCRO pszID = (PSZCRO )xmlGetProp (elm_d, (xmlChar *)"ID");

         elm_d = nextNode (elm_d);

         if (pszID)
         {
            return pszID;
         }
      }

      return 0;
   }

   xmlNodePtr  elm_d;
   PSZRO       pszNext_d;
   PSZRO       pszLast_d;
};

Enumeration * UPDFDeviceInstance::
listKeyValues (PSZCRO pszKey)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " ("
                          << pszKey << ")"
                          << std::endl;
#endif

   Enumeration *pRet        = 0;
   UPDFDevice  *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (  pszKey
      && *pszKey
      && pUPDFDevice
      )
   {
      std::string stringKey = pszKey;

      if (xmlObjectMap_d[stringKey])
      {
         xmlNodePtr root = pUPDFDevice->findEntryKeyValue (pUPDFDevice->getRootUDR (),
                                                           "ID",
                                                           pszKey,
                                                           DebugOutput::shouldOutputUPDFDeviceInstance ());

         pRet = new DeviceJobPropertyKeyValueEnumerator (root);
      }
   }

   if (!pRet)
   {
      pRet = new DeviceJobPropertyKeyValueEnumerator (0);
   }

   return pRet;
}

std::string * UPDFDeviceInstance::
getJobPropertyType (PSZCRO pszKey)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " ("
                          << pszKey << ")"
                          << std::endl;
#endif

   xmlNodePtr   node       = 0;
   std::string  stringKey  = pszKey;
   std::string *pstringRet = 0;

   if (  pszKey
      && *pszKey
      )
   {
      node = xmlObjectMap_d[stringKey];

      if (node)
      {
         PSZCRO pszID = (PSZCRO )xmlGetProp (node, (xmlChar *)"ID");

         if (pszID)
         {
            pstringRet = new std::string ("string");

            *pstringRet += pszID;

            free ((void *)pszID);
         }
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " returns "
                          << (pstringRet ? *pstringRet : "NULL")
                          << std::endl;
#endif

   return pstringRet;
}

std::string * UPDFDeviceInstance::
getJobProperty (PSZCRO pszKey)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " ("
                          << pszKey << ")"
                          << std::endl;
#endif

   xmlNodePtr   node       = 0;
   std::string  stringKey  = pszKey;
   std::string *pstringRet = 0;

   if (  pszKey
      && *pszKey
      )
   {
      node = xmlObjectMap_d[stringKey];

      if (node)
      {
         PSZCRO pszID = (PSZCRO )xmlGetProp (node, (xmlChar *)"ID");

         if (pszID)
         {
            pstringRet = new std::string (pszID);

            free ((void *)pszID);
         }
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " returns "
                          << (pstringRet ? *pstringRet : "NULL")
                          << std::endl;
#endif

   return pstringRet;
}

std::string * UPDFDeviceInstance::
translateKeyValue (PSZCRO pszKey,
                   PSZCRO pszValue)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " ("
                          << pszKey << ", "
                          << pszValue << ")"
                          << std::endl;
#endif

   UPDFDevice  *pUPDFDevice  = UPDFDevice::isAUPDFDevice (pDevice_d);
   xmlNodePtr   node         = 0;
   std::string  stringKey    = pszKey;
   std::string *pstringRet   = 0;

   if (  pszKey
      && *pszKey
      && pUPDFDevice
      )
   {
      node = xmlObjectMap_d[stringKey];

      if (node)
      {
         node = FINDUDRENTRYKEYVALUE (pUPDFDevice,
                                      "ID",
                                      pszValue,
                                      UPDFDeviceInstance);

         PSZCRO pszNameID = (PSZCRO )xmlGetProp (node, (xmlChar *)"Name_ID");

         if (pszNameID)
         {
            xmlNodePtr nodeLocale = FINDLOCALEENTRYKEYVALUE (pUPDFDevice,
                                                             "Name_ID",
                                                             pszNameID,
                                                             UPDFDeviceInstance);

            if (nodeLocale)
            {
               PSZCRO pszLocalizedString = (PSZCRO )xmlGetProp (nodeLocale,
                                                              (xmlChar *)"Localized_String");

               if (pszLocalizedString)
               {
                  pstringRet = new std::string (pszKey);

                  *pstringRet += "=";
                  *pstringRet += pszLocalizedString;

                  free ((void *)pszLocalizedString);
               }
            }

            free ((void *)pszNameID);
         }
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " returns "
                          << (pstringRet ? *pstringRet : "NULL")
                          << std::endl;
#endif

   return pstringRet;
}

void UPDFDeviceInstance::
setupPrinter ()
{
   if (fHaveSetupPrinter_d)
      return;

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::setupPrinter ()" << std::endl;
#endif

   fHaveSetupPrinter_d = true;
}

bool UPDFDeviceInstance::
beginJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " ()" << std::endl;
#endif

   return executeEvent ("Event_PDL_Start", true);
}

bool UPDFDeviceInstance::
beginJob (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " (with props)" << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The job properties are as follows:" << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The orientation is = " << getCurrentOrientation() << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "iDitherID = " << getCurrentDitherID () << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The form is = " << getCurrentForm () << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The tray is = " << getCurrentTray () << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The media is = " << getCurrentMedia () << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The resolution is = " << getCurrentResolution () << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The print mode is = " << getCurrentPrintMode () << std::endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return beginJob ();
}

bool UPDFDeviceInstance::
newFrame ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " ()" << std::endl;
#endif

   ditherNewFrame ();

   ptlPrintHead_d.x = 0;
   ptlPrintHead_d.y = 0;

   return true;
}

bool UPDFDeviceInstance::
newFrame (bool fJobPropertiesChanged)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " (with props)" << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The job properties are as follows:" << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The orientation is = " << getCurrentOrientation() << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "iDitherID = " << getCurrentDitherID () << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The form is = " << getCurrentForm () << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The tray is = " << getCurrentTray () << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The media is = " << getCurrentMedia () << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The resolution is = " << getCurrentResolution () << std::endl;
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "The print mode is = " << getCurrentPrintMode () << std::endl;
#endif

   // @TBD - reinitialize with new job properties

   // Call common code
   return newFrame ();
}

bool UPDFDeviceInstance::
endJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " ()" << std::endl;
#endif

   return true;
}

bool UPDFDeviceInstance::
abortJob ()
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceInstance ()) DebugOutput::getErrorStream () << "UPDFDeviceInstance::" << __FUNCTION__ << " ()" << std::endl;
#endif

   return true;
}

PSZCRO  UPDFDeviceInstance::
getXMLObjectValue (PSZCRO pszObjectName,
                   PSZCRO pszVariableName)
{
   if (  !pszObjectName
      || !pszVariableName
      )
   {
      return 0;
   }

   std::string stringObjectName = pszObjectName;
   xmlNodePtr  elm              = 0;

   elm = xmlObjectMap_d[stringObjectName];

   if (elm)
   {
      return (PSZCRO )xmlGetProp (elm, (xmlChar *)pszVariableName);
   }
   else
   {
      std::string  stringElm;
      PSZ          pszElm    = 0;

      stringElm = xmlStringMap_d[stringObjectName];

      pszElm = (PSZ)malloc (stringElm.length () + 1);
      if (pszElm)
      {
         strcpy (pszElm, stringElm.c_str ());
      }

      return pszElm;
   }

   return 0;
}

bool UPDFDeviceInstance::
executeEvent (PSZCRO pszPrefix, bool fPreEvent)
{
   char        achParameterName[128]; // @TBD
   DeviceData *pData                 = getDeviceData ();
   bool        fExecute              = fPreEvent;

   if (!pData)
      return true;

   int iCount = 0;

   sprintf (achParameterName, "%s_count", pszPrefix);

   if (  !pData->getIntData (achParameterName, &iCount)
      || 0 == iCount
      )
      return true;

   for (int i = 0; i < iCount; i++)
   {
      BinaryData *pbdExpression = 0;
      PSZ         pszEventName  = 0;

      sprintf (achParameterName, "%s_binary_%d", pszPrefix, i);

      if (pData->getBinaryData (achParameterName, &pbdExpression))
      {
      }
      else if (pData->getStringData (achParameterName, &pszEventName))
      {
         fExecute = !fExecute;
      }
      else
      {
      }
   }

   return true;
}

std::string UPDFDeviceInstance::
toString (std::ostringstream& oss)
{
   std::ostringstream oss2;

   oss << "{ "
       << DeviceInstance::toString (oss2) // Add parent's output as well
       << " }";

   return oss.str ();
}

/* Provide a way to print out class data
*/
std::ostream&
operator<< (std::ostream& os, const UPDFDeviceInstance& const_self)
{
   UPDFDeviceInstance& self = const_cast<UPDFDeviceInstance&>(const_self);
   std::ostringstream  oss;

   os << self.toString (oss);

   return os;
}
