/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <UPDFDeviceMedia.hpp>

/*
*/
static int     getOmniMediaID            (xmlNodePtr      node);

UPDFDeviceMedia::
UPDFDeviceMedia (Device      *pDevice,
                 int          id,
                 BinaryData  *data,
                 int          iColorAdjustRequired,
                 int          iAbsorption,
                 xmlNodePtr   node)
   : DeviceMedia (pDevice,
                  id,
                  data,
                  iColorAdjustRequired,
                  iAbsorption)
{
   node_d = node;
}

UPDFDeviceMedia::
~UPDFDeviceMedia ()
{
}

DeviceMedia * UPDFDeviceMedia::
createV (int id)
{
   return UPDFDeviceMedia::create (pDevice_d, id);
}

DeviceMedia * UPDFDeviceMedia::
create (Device *pDevice,
        PSZCRO  pszId)
{
   int id = DeviceMedia::nameToID (pszId);

   if (DeviceMedia::MEDIA_UNLISTED != id)
   {
      return create (pDevice, id);
   }
   else
   {
      UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);
      xmlNodePtr  nodeMedias   = 0;

      nodeMedias = findMedias (pUPDFDevice);

      if (nodeMedias)
      {
         xmlNodePtr nodeItem  = firstNode (nodeMedias->xmlChildrenNode);
         xmlNodePtr nodeFound = 0;

         while (nodeItem != 0)
         {
            PSZCRO pszNodeId = (PSZCRO)xmlGetProp (nodeItem, (xmlChar *)"ID");

            if (pszNodeId)
            {
               if (0 == strcmp (pszNodeId, pszId))
               {
                  nodeFound = nodeItem;
               }

               free ((void *)pszNodeId);
            }

            if (nodeFound)
            {
               id = getOmniMediaID (nodeFound);

               if (DeviceMedia::MEDIA_UNLISTED != id)
               {
                  return create (pDevice, id);
               }

               nodeFound = 0;
            }

            nodeItem = nextNode (nodeItem);
         }
      }

      return 0;
   }
}

static int
getOmniMediaID (xmlNodePtr node)
{
   int id = DeviceMedia::MEDIA_UNLISTED;

   if (!node)
   {
      return id;
   }

   PSZCRO pszID = (PSZCRO)xmlGetProp (node, (xmlChar *)"Predefined_ID");

   if (pszID)
   {
      if (0 == strcasecmp (pszID, "device-setting"))
         id = DeviceMedia::MEDIA_AUTO;
      else if (0 == strcasecmp (pszID, "stationery"))
         id = DeviceMedia::MEDIA_PLAIN;
      else if (0 == strcasecmp (pszID, "stationery-preprinted"))
         id = DeviceMedia::MEDIA_PREPRINTED;
      else if (0 == strcasecmp (pszID, "stationery-letterhead"))
         id = DeviceMedia::MEDIA_LETTERHEAD;
      else if (0 == strcasecmp (pszID, "transparency"))
         id = DeviceMedia::MEDIA_TRANSPARENCY;
      else if (0 == strcasecmp (pszID, "stationery-prepunched"))
         id = DeviceMedia::MEDIA_PREPUNCHED;
      else if (0 == strcasecmp (pszID, "labels"))
         id = DeviceMedia::MEDIA_LABELS;
      else if (0 == strcasecmp (pszID, "cardstock"))
         id = DeviceMedia::MEDIA_CARDSTOCK;

      free ((void *)pszID);
   }

   return id;
}

static xmlNodePtr
skipInvalidMedias (xmlNodePtr node)
{
   if (!node)
   {
      return 0;
   }

   int id = getOmniMediaID (node);

   while (  node
           && (  id == DeviceMedia::MEDIA_UNLISTED
              || id == DeviceMedia::MEDIA_NONE
              )
           )
   {
      node = nextNode (node);
      id   = getOmniMediaID (node);
   }

   return node;
}

static DeviceMedia *
createFromXMLNode (Device      *pDevice,
                   xmlNodePtr   node,
                   int          id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   int         iColorAdjustRequired = 0;
   int         iAbsorption          = DeviceMedia::MEDIA_NO_ABSORPTION;
   BinaryData *data                 = 0;

   return new UPDFDeviceMedia (pUPDFDevice,
                               id,
                               data,
                               iColorAdjustRequired,
                               iAbsorption,
                               node);
}

DeviceMedia * UPDFDeviceMedia::
create (Device *pDevice, int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceMedia ()) DebugOutput::getErrorStream () << "UPDFDeviceMedia::" << __FUNCTION__ << ": pDevice = " << pDevice << ", id = " << id << std::endl;
#endif

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeMedias = 0;

   nodeMedias = findMedias (pUPDFDevice);

   if (nodeMedias)
   {
      xmlNodePtr nodeItem = firstNode (nodeMedias->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iMediaID = getOmniMediaID (nodeItem);

         if (  id == iMediaID
            && !(  id == DeviceMedia::MEDIA_UNLISTED
                || id == DeviceMedia::MEDIA_NONE
                )
            )
         {
            return createFromXMLNode (pDevice, nodeItem, id);
         }

         nodeItem = nextNode (nodeItem);
      }

      return 0;
   }
   else
   {
      return 0;
   }
}

PSZCRO UPDFDeviceMedia::
getDeviceID ()
{
   return 0;
}

bool UPDFDeviceMedia::
isMediaSupported (int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeMedias = 0;

   nodeMedias = findMedias (pUPDFDevice);

   if (nodeMedias)
   {
      xmlNodePtr nodeItem = firstNode (nodeMedias->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iMediaID = getOmniMediaID (nodeItem);

         if (  id == iMediaID
            && !(  id == DeviceMedia::MEDIA_UNLISTED
                || id == DeviceMedia::MEDIA_NONE
                )
            )
         {
            return true;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return false;
}

class UPDFMedia_Enumerator : public Enumeration
{
public:
   UPDFMedia_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      pDevice_d  = pDevice;
      nodeItem_d = skipInvalidMedias (nodeItem);
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      void *pvRet = 0;

      if (nodeItem_d)
      {
         int id = getOmniMediaID (nodeItem_d);

         pvRet = createFromXMLNode (pDevice_d,
                                    nodeItem_d,
                                    id);

         nodeItem_d = nextNode (nodeItem_d);
         nodeItem_d = skipInvalidMedias (nodeItem_d);
      }

      return pvRet;
   }

private:
   Device     *pDevice_d;
   xmlNodePtr  nodeItem_d;
};

Enumeration * UPDFDeviceMedia::
getEnumeration ()
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDFMedia_Enumerator (pDevice_d, 0);
   }

   xmlNodePtr nodeMedias = 0;

   nodeMedias = findMedias (pUPDFDevice);

   if (nodeMedias)
   {
      xmlNodePtr nodeItem = firstNode (nodeMedias->xmlChildrenNode);

      return new UPDFMedia_Enumerator (pDevice_d, nodeItem);
   }
   else
   {
      return new UPDFMedia_Enumerator (pDevice_d, 0);
   }
}

xmlNodePtr UPDFDeviceMedia::
findMedias (UPDFDevice *pUPDFDevice)
{
   xmlNodePtr nodeMedias = 0;

   if (!pUPDFDevice)
   {
      return nodeMedias;
   }

   if (  ((nodeMedias = FINDUDRENTRY (pUPDFDevice, nodeMedias, "Device_Features",    UPDFDeviceMedia)) != 0)
      && ((nodeMedias = FINDUDRENTRY (pUPDFDevice, nodeMedias, "PrintCapabilities",  UPDFDeviceMedia)) != 0)
      && ((nodeMedias = FINDUDRENTRY (pUPDFDevice, nodeMedias, "Features",           UPDFDeviceMedia)) != 0)
      && ((nodeMedias = FINDUDRENTRY (pUPDFDevice, nodeMedias, "PrintMediaHandling", UPDFDeviceMedia)) != 0)
      && ((nodeMedias = FINDUDRENTRY (pUPDFDevice, nodeMedias, "MediaType",          UPDFDeviceMedia)) != 0)
      )
      return nodeMedias;

   return nodeMedias;
}

std::string UPDFDeviceMedia::
toString (std::ostringstream& oss)
{
   std::ostringstream oss2;

   oss << "{"
       << DeviceMedia::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
std::ostream&
operator<< (std::ostream& os, const UPDFDeviceMedia& const_self)
{
   UPDFDeviceMedia&   self = const_cast<UPDFDeviceMedia&>(const_self);
   std::ostringstream oss;

   os << self.toString (oss);

   return os;
}
