/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <UPDFDeviceNUp.hpp>

/*
*/
static int     getOmniNUpID             (xmlNodePtr      node);

UPDFDeviceNUp::
UPDFDeviceNUp (Device      *pDevice,
               int          id,
               BinaryData  *data,
               xmlNodePtr   node,
               bool         fDeviceFeature)
   : DeviceNUp (pDevice,
                id,
                data,
                fDeviceFeature)
{
   node_d = node;
}

UPDFDeviceNUp::
~UPDFDeviceNUp ()
{
}

DeviceNUp * UPDFDeviceNUp::
createV (int id)
{
   return UPDFDeviceNUp::create (pDevice_d, id);
}

DeviceNUp * UPDFDeviceNUp::
create (Device *pDevice,
        PSZCRO  pszId)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);
   xmlNodePtr  nodeNUps = 0;

   nodeNUps = findNUps (pUPDFDevice);

   if (nodeNUps)
   {
      xmlNodePtr nodeItem  = firstNode (nodeNUps->xmlChildrenNode);
      xmlNodePtr nodeFound = 0;

      while (nodeItem != 0)
      {
         PSZCRO pszNodeId = (PSZCRO)xmlGetProp (nodeItem, (xmlChar *)"ID");

         if (pszNodeId)
         {
            if (0 == strcmp (pszNodeId, pszId))
            {
               nodeFound = nodeItem;
            }

            free ((void *)pszNodeId);
         }

         if (nodeFound)
         {
            int id = getOmniNUpID (nodeFound);

            if (DeviceNUp::NUP_UNLISTED != id)
            {
               return create (pDevice, id);
            }

            nodeFound = 0;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return 0;
}

static int
getOmniNUpID (xmlNodePtr node)
{
   int id = DeviceNUp::NUP_UNLISTED;

   if (!node)
   {
      return id;
   }

   PSZCRO pszID = (PSZCRO)xmlGetProp (node, (xmlChar *)"ID");

   if (pszID)
   {
      id = DeviceNUp::nameToID (pszID);

      free ((void *)pszID);
   }

   return id;
}

static xmlNodePtr
skipInvalidNUps (xmlNodePtr node)
{
   if (!node)
   {
      return 0;
   }

   int id = getOmniNUpID (node);

   while (  node
         && (id == DeviceNUp::NUP_UNLISTED)
         )
   {
      node = nextNode (node);
      id   = getOmniNUpID (node);
   }

   return node;
}

static DeviceNUp *
createFromXMLNode (Device      *pDevice,
                   xmlNodePtr   node,
                   int          id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   bool   fDeviceFeature   = true;
   PSZCRO pszDeviceFeature = (PSZCRO)xmlGetProp (node,
                                                 (xmlChar *)"DeviceFeature");

   if (pszDeviceFeature)
   {
      if (0 == strcasecmp (pszDeviceFeature, "false"))
      {
         fDeviceFeature = false;
      }

      free ((void *)pszDeviceFeature);
   }

   BinaryData  *data = 0; // @TBD

   return new UPDFDeviceNUp (pUPDFDevice,
                                id,
                                data,
                                node,
                                !fDeviceFeature);
}

DeviceNUp * UPDFDeviceNUp::
create (Device *pDevice, int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceNUp ()) DebugOutput::getErrorStream () << "UPDFDeviceNUp::" << __FUNCTION__ << ": pDevice = " << pDevice << ", id = " << id << std::endl;
#endif

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeNUps = 0;

   nodeNUps = findNUps (pUPDFDevice);

   if (nodeNUps)
   {
      xmlNodePtr nodeItem = firstNode (nodeNUps->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iNUpID = getOmniNUpID (nodeItem);

         if (  id == iNUpID
            && !(id == DeviceNUp::NUP_UNLISTED)
            )
         {
            return createFromXMLNode (pDevice, nodeItem, id);
         }

         nodeItem = nextNode (nodeItem);
      }

      return 0;
   }
   else
   {
      return 0;
   }
}

PSZCRO UPDFDeviceNUp::
getDeviceID ()
{
   return 0;
}

bool UPDFDeviceNUp::
isNUpSupported (int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeNUps = 0;

   nodeNUps = findNUps (pUPDFDevice);

   if (nodeNUps)
   {
      xmlNodePtr nodeItem = firstNode (nodeNUps->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iNUpID = getOmniNUpID (nodeItem);

         if (  id == iNUpID
            && !(id == DeviceNUp::NUP_UNLISTED)
            )
         {
            return true;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return false;
}

class UPDFNUp_Enumerator : public Enumeration
{
public:
   UPDFNUp_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      pDevice_d  = pDevice;
      nodeItem_d = skipInvalidNUps (nodeItem);
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      void *pvRet = 0;

      if (nodeItem_d)
      {
         int id = getOmniNUpID (nodeItem_d);

         pvRet = createFromXMLNode (pDevice_d,
                                    nodeItem_d,
                                    id);

         nodeItem_d = nextNode (nodeItem_d);
         nodeItem_d = skipInvalidNUps (nodeItem_d);
      }

      return pvRet;
   }

private:
   Device     *pDevice_d;
   xmlNodePtr  nodeItem_d;
};

Enumeration * UPDFDeviceNUp::
getEnumeration ()
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDFNUp_Enumerator (pDevice_d, 0);
   }

   xmlNodePtr nodeNUps = 0;

   nodeNUps = findNUps (pUPDFDevice);

   if (nodeNUps)
   {
      xmlNodePtr nodeItem = firstNode (nodeNUps->xmlChildrenNode);

      return new UPDFNUp_Enumerator (pDevice_d, nodeItem);
   }
   else
   {
      return new UPDFNUp_Enumerator (pDevice_d, 0);
   }
}

xmlNodePtr UPDFDeviceNUp::
findNUps (UPDFDevice *pUPDFDevice)
{
   xmlNodePtr nodeNUps = 0;

   if (!pUPDFDevice)
   {
      return nodeNUps;
   }

   if (  ((nodeNUps = FINDUDRENTRY (pUPDFDevice, nodeNUps, "Device_Features",    UPDFDeviceNUp)) != 0)
      && ((nodeNUps = FINDUDRENTRY (pUPDFDevice, nodeNUps, "PrintCapabilities",  UPDFDeviceNUp)) != 0)
      && ((nodeNUps = FINDUDRENTRY (pUPDFDevice, nodeNUps, "Features",           UPDFDeviceNUp)) != 0)
      && ((nodeNUps = FINDUDRENTRY (pUPDFDevice, nodeNUps, "PrintMediaHandling", UPDFDeviceNUp)) != 0)
      && ((nodeNUps = FINDUDRENTRY (pUPDFDevice, nodeNUps, "MediaNUpList",       UPDFDeviceNUp)) != 0)
      )
      return nodeNUps;

   return nodeNUps;
}

std::string UPDFDeviceNUp::
toString (std::ostringstream& oss)
{
   std::ostringstream oss2;

   oss << "{"
       << DeviceNUp::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
std::ostream&
operator<< (std::ostream& os, const UPDFDeviceNUp& const_self)
{
   UPDFDeviceNUp&     self = const_cast<UPDFDeviceNUp&>(const_self);
   std::ostringstream oss;

   os << self.toString (oss);

   return os;
}
