/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "UPDFDeviceOrientation.hpp"

/*
*/
static int     getOmniOrientationID      (xmlNodePtr nodeOrientation,
                                          xmlNodePtr nodeRotation);

UPDFDeviceOrientation::
UPDFDeviceOrientation (Device     *pDevice,
                       int         id,
                       xmlNodePtr  nodeOrientation,
                       xmlNodePtr  nodeRotation,
                       bool        fDeviceFeature)
   : DeviceOrientation (pDevice,
                        id,
                        fDeviceFeature)
{
   nodeOrientation_d = nodeOrientation;
   nodeRotation_d    = nodeRotation;
}

UPDFDeviceOrientation::
~UPDFDeviceOrientation ()
{
}

DeviceOrientation * UPDFDeviceOrientation::
createV (int id)
{
   return UPDFDeviceOrientation::create (pDevice_d, id);
}

static DeviceOrientation *
createOrientation (Device     *pDevice,
                   xmlNodePtr  nodeOrientation,
                   xmlNodePtr  nodeRotation)
{
   if (  pDevice
      && nodeOrientation
      && nodeRotation
      )
   {
      int id = getOmniOrientationID (nodeOrientation, nodeRotation);

      if (DeviceOrientation::ORIENTATION_UNLISTED == id)
         return 0;

      bool   fDeviceFeature   = true;
      PSZCRO pszDeviceFeature = (PSZCRO)xmlGetProp (nodeOrientation,
                                                    (xmlChar *)"DeviceFeature");

      if (pszDeviceFeature)
      {
         if (0 == strcasecmp (pszDeviceFeature, "false"))
         {
            fDeviceFeature = false;
         }

         free ((void *)pszDeviceFeature);
      }

      return new UPDFDeviceOrientation (pDevice,
                                        id,
                                        nodeOrientation,
                                        nodeRotation,
                                        !fDeviceFeature);
   }

   return 0;
}

DeviceOrientation * UPDFDeviceOrientation::
create (Device *pDevice,
        PSZCRO  pszOrientationID,
        PSZCRO  pszRotationID)
{
   UPDFDevice *pUPDFDevice      = UPDFDevice::isAUPDFDevice (pDevice);
   xmlNodePtr  nodeOrientations = 0;
   xmlNodePtr  nodeRotations    = 0;

   nodeOrientations = findOrientations (pUPDFDevice);
   nodeRotations    = findRotations (pUPDFDevice);

   if (  !nodeOrientations
      || !nodeRotations
      )
      return 0;

   xmlNodePtr nodeItem        = 0;
   xmlNodePtr nodeOrientation = 0;
   xmlNodePtr nodeRotation    = 0;

   nodeItem = firstNode (nodeOrientations->xmlChildrenNode);
   while (nodeItem != 0)
   {
      PSZCRO pszNodeID = (PSZCRO)xmlGetProp (nodeItem, (xmlChar *)"ID");

      if (pszNodeID)
      {
         if (0 == strcmp (pszNodeID, pszOrientationID))
         {
            nodeOrientation = nodeItem;
         }

         free ((void *)pszNodeID);
      }

      if (nodeOrientation)
         break;

      nodeItem = nextNode (nodeItem);
   }

   nodeItem = firstNode (nodeRotations->xmlChildrenNode);
   while (nodeItem != 0)
   {
      PSZCRO pszNodeID = (PSZCRO)xmlGetProp (nodeItem, (xmlChar *)"ID");

      if (pszNodeID)
      {
         if (0 == strcmp (pszNodeID, pszRotationID))
         {
            nodeRotation = nodeItem;
         }

         free ((void *)pszNodeID);
      }

      if (nodeRotation)
         break;

      nodeItem = nextNode (nodeItem);
   }

   return createOrientation (pDevice, nodeOrientation, nodeRotation);
}

static int
getOmniOrientationID (xmlNodePtr nodeOrientation,
                      xmlNodePtr nodeRotation)
{
   int id = DeviceOrientation::ORIENTATION_UNLISTED;

   if (!nodeOrientation)
   {
      return id;
   }

   std::ostringstream oss;
   PSZCRO             pszOrientationID = (PSZCRO)xmlGetProp (nodeOrientation,
                                                            (xmlChar *)"Predefined_ID");
   PSZCRO             pszRotationID    = (PSZCRO)xmlGetProp (nodeRotation,
                                                            (xmlChar *)"Predefined_ID");

   if (  pszOrientationID
      && pszRotationID
      )
   {
      if (0 == strcmp (pszRotationID, "reverse"))
      {
         oss << pszRotationID << '-';
      }
      oss << pszOrientationID << std::ends;

      id = DeviceOrientation::nameToID ((PSZCRO )oss.str ().c_str ());
   }

   if (pszOrientationID)
   {
      free ((void *)pszOrientationID);
   }

   if (pszRotationID)
   {
      free ((void *)pszRotationID);
   }

   return id;
}

DeviceOrientation * UPDFDeviceOrientation::
create (Device *pDevice,
        int     id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeOrientations = 0;
   xmlNodePtr nodeRotations    = 0;

   nodeOrientations = findOrientations (pUPDFDevice);
   nodeRotations    = findRotations (pUPDFDevice);

   if (  !nodeOrientations
      || !nodeRotations
      )
      return 0;

   xmlNodePtr nodeOrientation = 0;
   xmlNodePtr nodeRotation    = 0;

   nodeOrientation = firstNode (nodeOrientations->xmlChildrenNode);
   nodeRotation    = firstNode (nodeRotations->xmlChildrenNode);

   while (  nodeOrientation
         && nodeRotation
         )
   {
      int iOrientationID = getOmniOrientationID (nodeOrientation,
                                                 nodeRotation);

      if (id == iOrientationID)
      {
         break;
      }

      nodeRotation = nextNode (nodeRotation);
      if (!nodeRotation)
      {
         nodeRotation    = firstNode (nodeRotations->xmlChildrenNode);
         nodeOrientation = nextNode (nodeOrientation);
      }
   }

   return createOrientation (pDevice, nodeOrientation, nodeRotation);
}

PSZCRO UPDFDeviceOrientation::
getDeviceID ()
{
   return 0;
}

bool UPDFDeviceOrientation::
isOrientationSupported (int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeOrientations = 0;
   xmlNodePtr nodeRotations    = 0;

   nodeOrientations = findOrientations (pUPDFDevice);
   nodeRotations    = findRotations (pUPDFDevice);

   if (  !nodeOrientations
      || !nodeRotations
      )
      return false;

   xmlNodePtr nodeOrientation = 0;
   xmlNodePtr nodeRotation    = 0;

   nodeOrientation = firstNode (nodeOrientations->xmlChildrenNode);
   nodeRotation    = firstNode (nodeRotations->xmlChildrenNode);

   while (  nodeOrientation
         && nodeRotation
         )
   {
      int iOrientationID = getOmniOrientationID (nodeOrientation,
                                                 nodeRotation);

      if (id == iOrientationID)
      {
         return true;
      }

      nodeRotation = nextNode (nodeRotation);
      if (!nodeRotation)
      {
         nodeRotation    = firstNode (nodeRotations->xmlChildrenNode);
         nodeOrientation = nextNode (nodeOrientation);
      }
   }

   return false;
}

class UPDFOrientation_Enumerator : public Enumeration
{
public:
   UPDFOrientation_Enumerator (Device     *pDevice,
                               xmlNodePtr  nodeOrientations,
                               xmlNodePtr  nodeRotations)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputUPDFDeviceOrientation ())
      {
         DebugOutput::getErrorStream () << "UPDFOrientation_Enumerator::" << __FUNCTION__ << ": nodeOrientations = " << nodeOrientations << std::endl;
         DebugOutput::getErrorStream () << "UPDFOrientation_Enumerator::" << __FUNCTION__ << ": nodeRotations    = " << nodeRotations << std::endl;
      }
#endif

      pDevice_d          = pDevice;
      nodeOrientations_d = nodeOrientations;
      nodeRotations_d    = nodeRotations;
      nodeOrientation_d  = firstNode (nodeOrientations->xmlChildrenNode);
      nodeRotation_d     = firstNode (nodeRotations->xmlChildrenNode);

      skipInvalidOrientations ();
   }

   virtual bool
   hasMoreElements ()
   {
      return (nodeOrientation_d && nodeRotation_d) ? true : false;
   }

   virtual void *
   nextElement ()
   {
      void *pvRet = 0;

      if (  nodeOrientation_d
         && nodeRotation_d
         )
      {
         pvRet = UPDFDeviceOrientation::create (pDevice_d, id_d);

#ifndef RETAIL
         if (DebugOutput::shouldOutputUPDFDeviceOrientation ())
         {
            DebugOutput::getErrorStream () << "UPDFOrientation_Enumerator::" << __FUNCTION__ << ": nextElement: pvRet = " << pvRet << std::endl;
         }
#endif

         moveToNextOrientation ();
         skipInvalidOrientations ();
      }

      return pvRet;
   }

private:

   void
   moveToNextOrientation ()
   {
      nodeRotation_d = nextNode (nodeRotation_d);
      if (!nodeRotation_d)
      {
         nodeRotation_d = firstNode (nodeRotations_d->xmlChildrenNode);

         nodeOrientation_d = nextNode (nodeOrientation_d);
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputUPDFDeviceOrientation ())
      {
         DebugOutput::getErrorStream () << "UPDFOrientation_Enumerator::" << __FUNCTION__ << ": moveToNextOrientation: nodeOrientations_d = " << nodeOrientations_d << std::endl;
         DebugOutput::getErrorStream () << "UPDFOrientation_Enumerator::" << __FUNCTION__ << ": moveToNextOrientation: nodeRotations_d    = " << nodeRotations_d << std::endl;
      }
#endif
   }

   void
   skipInvalidOrientations ()
   {
      if (  !nodeOrientations_d
         || !nodeRotations_d
         || !nodeOrientation_d
         || !nodeRotation_d
         )
      {
         return;
      }

      id_d = getOmniOrientationID (nodeOrientation_d, nodeRotation_d);

#ifndef RETAIL
      if (DebugOutput::shouldOutputUPDFDeviceOrientation ())
      {
         DebugOutput::getErrorStream () << "UPDFOrientation_Enumerator::" << __FUNCTION__ << ": skipInvalidOrientations: id_d = " << id_d << std::endl;
      }
#endif

      while (  (  nodeOrientation_d
               && nodeRotation_d
               )
            && (  id_d == DeviceOrientation::ORIENTATION_UNLISTED
               || id_d == DeviceOrientation::ORIENTATION_NONE
               )
            )
      {
         moveToNextOrientation ();

         id_d = getOmniOrientationID (nodeOrientation_d, nodeRotation_d);

#ifndef RETAIL
         if (DebugOutput::shouldOutputUPDFDeviceOrientation ())
         {
            DebugOutput::getErrorStream () << "UPDFOrientation_Enumerator::" << __FUNCTION__ << ": skipInvalidOrientations: id_d = " << id_d << std::endl;
         }
#endif
      }
   }

   Device     *pDevice_d;
   xmlNodePtr  nodeOrientations_d;
   xmlNodePtr  nodeRotations_d;
   xmlNodePtr  nodeOrientation_d;
   xmlNodePtr  nodeRotation_d;
   int         id_d;
};

Enumeration * UPDFDeviceOrientation::
getEnumeration ()
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDFOrientation_Enumerator (pDevice_d, 0, 0);
   }

   xmlNodePtr nodeOrientations = 0;
   xmlNodePtr nodeRotations    = 0;

   nodeOrientations = findOrientations (pUPDFDevice);
   nodeRotations    = findRotations (pUPDFDevice);

   if (  nodeOrientations
      && nodeRotations
      )
   {
      return new UPDFOrientation_Enumerator (pDevice_d,
                                             nodeOrientations,
                                             nodeRotations);
   }
   else
   {
      return new UPDFOrientation_Enumerator (pDevice_d, 0, 0);
   }
}

xmlNodePtr UPDFDeviceOrientation::
findOrientations (UPDFDevice *pUPDFDevice)
{
   xmlNodePtr nodeOrientations = 0;

   if (!pUPDFDevice)
   {
      return nodeOrientations;
   }

   if (  ((nodeOrientations = FINDUDRENTRY (pUPDFDevice, nodeOrientations, "Device_Features",      UPDFDeviceOrientation)) != 0)
      && ((nodeOrientations = FINDUDRENTRY (pUPDFDevice, nodeOrientations, "PrintCapabilities",    UPDFDeviceOrientation)) != 0)
      && ((nodeOrientations = FINDUDRENTRY (pUPDFDevice, nodeOrientations, "Features",             UPDFDeviceOrientation)) != 0)
      && ((nodeOrientations = FINDUDRENTRY (pUPDFDevice, nodeOrientations, "PrintMediaHandling",   UPDFDeviceOrientation)) != 0)
      && ((nodeOrientations = FINDUDRENTRY (pUPDFDevice, nodeOrientations, "MediaPageOrientation", UPDFDeviceOrientation)) != 0)
      )
      return nodeOrientations;

   return nodeOrientations;
}

xmlNodePtr UPDFDeviceOrientation::
findRotations (UPDFDevice *pUPDFDevice)
{
   xmlNodePtr nodeRotations = 0;

   if (!pUPDFDevice)
   {
      return nodeRotations;
   }

   if (  ((nodeRotations = FINDUDRENTRY (pUPDFDevice, nodeRotations, "Device_Features",    UPDFDeviceOrientation)) != 0)
      && ((nodeRotations = FINDUDRENTRY (pUPDFDevice, nodeRotations, "PrintCapabilities",  UPDFDeviceOrientation)) != 0)
      && ((nodeRotations = FINDUDRENTRY (pUPDFDevice, nodeRotations, "Features",           UPDFDeviceOrientation)) != 0)
      && ((nodeRotations = FINDUDRENTRY (pUPDFDevice, nodeRotations, "PrintMediaHandling", UPDFDeviceOrientation)) != 0)
      && ((nodeRotations = FINDUDRENTRY (pUPDFDevice, nodeRotations, "MediaPageRotation",  UPDFDeviceOrientation)) != 0)
      )
      return nodeRotations;

   return nodeRotations;
}

std::string UPDFDeviceOrientation::
toString (std::ostringstream& oss)
{
   std::ostringstream oss2;

   oss << "{";

   // @TBD
   printNode (nodeOrientation_d, DebugOutput::shouldOutputUPDFDeviceOrientation ());
   printNode (nodeRotation_d,    DebugOutput::shouldOutputUPDFDeviceOrientation ());

   oss << DeviceOrientation::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
std::ostream&
operator<< (std::ostream& os, const UPDFDeviceOrientation& const_self)
{
   UPDFDeviceOrientation& self = const_cast<UPDFDeviceOrientation&>(const_self);
   std::ostringstream     oss;

   os << self.toString (oss);

   return os;
}
