/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <UPDFDevicePrintMode.hpp>

UPDFDevicePrintMode::
UPDFDevicePrintMode (Device      *pDevice,
                     int          id,
                     int          iPhysicalCount,
                     int          iLogicalCount,
                     int          iPlanes,
                     xmlNodePtr   node)
   : DevicePrintMode (pDevice,
                      id,
                      iPhysicalCount,
                      iLogicalCount,
                      iPlanes)
{
   node_d = node;
}

UPDFDevicePrintMode::
~UPDFDevicePrintMode ()
{
}

DevicePrintMode * UPDFDevicePrintMode::
createV (int id)
{
   return UPDFDevicePrintMode::create (pDevice_d, id);
}

DevicePrintMode * UPDFDevicePrintMode::
create (Device *pDevice,
        PSZCRO  pszId)
{
   return create (pDevice, DevicePrintMode::nameToID (pszId));
}

static int
getOmniPrintModeID (xmlNodePtr node)
{
   int id = DevicePrintMode::PRINT_MODE_NONE;

   if (!node)
   {
      return id;
   }

   PSZCRO pszID = (PSZCRO)xmlGetProp (node, (xmlChar *)"Predefined_ID");

   if (pszID)
   {
      if (0 == strcasecmp (pszID, "monochrome"))
         id = DevicePrintMode::PRINT_MODE_1_ANY;

      free ((void *)pszID);
   }

   return id;
}

static DevicePrintMode *
createFromXMLNode (Device      *pDevice,
                   xmlNodePtr   node,
                   int          id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   int iPhysicalCount = 0;
   int iLogicalCount  = 0;
   int iPlanes        = 0;

   switch (id)
   {
   case DevicePrintMode::PRINT_MODE_1_ANY:
   {
      iPhysicalCount = 1;
      iLogicalCount  = 1;
      iPlanes        = 1;
      break;
   }

   default:
      // @TBD
      break;
   }

   return new UPDFDevicePrintMode (pUPDFDevice,
                                   DevicePrintMode::PRINT_MODE_1_ANY,
                                   iPhysicalCount,
                                   iLogicalCount,
                                   iPlanes,
                                   node);
}

DevicePrintMode * UPDFDevicePrintMode::
create (Device *pDevice, int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDevicePrintMode ()) DebugOutput::getErrorStream () << "UPDFDevicePrintMode::" << __FUNCTION__ << ": pDevice = " << pDevice << ", id = " << id << std::endl;
#endif

   if (!pUPDFDevice)
   {
      return 0;
   }

   return createFromXMLNode (pDevice, 0, id);
}

PSZCRO UPDFDevicePrintMode::
getDeviceID ()
{
   return 0;
}

bool UPDFDevicePrintMode::
isPrintModeSupported (int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodePrintModes = 0;

   nodePrintModes = findPrintModes (pUPDFDevice);

   if (nodePrintModes)
   {
      xmlNodePtr nodeItem = firstNode (nodePrintModes->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iPrintModeID = getOmniPrintModeID (nodeItem);

         if (  id == iPrintModeID
            && !(  id == DevicePrintMode::PRINT_MODE_UNLISTED
                || id == DevicePrintMode::PRINT_MODE_NONE
                )
            )
         {
            return true;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return false;
}

class UPDFPrintMode_Enumerator : public Enumeration
{
public:
   UPDFPrintMode_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      pDevice_d             = pDevice;
      fHasColor_d           = false; // @TBD
      iPrintModesReturned_d = 0;
   }

   virtual bool hasMoreElements ()
   {
      if (fHasColor_d)
      {
         return iPrintModesReturned_d < 2 ? true : false;
      }
      else
      {
         return iPrintModesReturned_d < 1 ? true : false;
      }
   }

   virtual void *nextElement ()
   {
      void *pvRet = 0;

      if (hasMoreElements ())
      {
         iPrintModesReturned_d++;

         pvRet = createFromXMLNode (pDevice_d,
                                    0,
                                    0);
      }

      return pvRet;
   }

private:
   Device     *pDevice_d;
   bool        fHasColor_d;
   int         iPrintModesReturned_d;
};

Enumeration * UPDFDevicePrintMode::
getEnumeration ()
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDFPrintMode_Enumerator (pDevice_d, 0);
   }

   xmlNodePtr nodePrintModes = 0;

   nodePrintModes = findPrintModes (pUPDFDevice);

   if (nodePrintModes)
   {
      xmlNodePtr nodeItem = firstNode (nodePrintModes->xmlChildrenNode);

      return new UPDFPrintMode_Enumerator (pDevice_d, nodeItem);
   }
   else
   {
      return new UPDFPrintMode_Enumerator (pDevice_d, 0);
   }
}

xmlNodePtr UPDFDevicePrintMode::
findPrintModes (UPDFDevice *pUPDFDevice)
{
   xmlNodePtr nodePrintModes = 0;

   if (!pUPDFDevice)
   {
      return nodePrintModes;
   }

   if (  ((nodePrintModes = FINDUDRENTRY (pUPDFDevice, nodePrintModes, "Device_Features",   UPDFDevicePrintMode)) != 0)
      && ((nodePrintModes = FINDUDRENTRY (pUPDFDevice, nodePrintModes, "PrintCapabilities", UPDFDevicePrintMode)) != 0)
      && ((nodePrintModes = FINDUDRENTRY (pUPDFDevice, nodePrintModes, "Features",          UPDFDevicePrintMode)) != 0)
      && ((nodePrintModes = FINDUDRENTRY (pUPDFDevice, nodePrintModes, "ColorHandling",     UPDFDevicePrintMode)) != 0)
      )
      return nodePrintModes;

   return nodePrintModes;
}

std::string UPDFDevicePrintMode::
toString (std::ostringstream& oss)
{
   std::ostringstream oss2;

   oss << "{"
       << DevicePrintMode::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
std::ostream&
operator<< (std::ostream& os, const UPDFDevicePrintMode& const_self)
{
   UPDFDevicePrintMode& self = const_cast<UPDFDevicePrintMode&>(const_self);
   std::ostringstream   oss;

   os << self.toString (oss);

   return os;
}
