/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <UPDFDeviceResolution.hpp>

/*
*/
static int     getOmniResolutionID       (xmlNodePtr   node,
                                          UPDFDevice  *pUPDFDevice,
                                          int         *piXRes,
                                          int         *piYRes);

UPDFDeviceResolution::
UPDFDeviceResolution (Device      *pDevice,
                      int          id,
                      int          iXRes,
                      int          iYRes,
                      int          iXInternalRes,
                      int          iYInternalRes,
                      BinaryData  *pbdData,
                      int          iCapabilities,
                      int          iDestinationBitsPerPel,
                      int          iScanlineMultiple,
                      xmlNodePtr   node)
   : DeviceResolution (pDevice,
                       id,
                       iXRes,
                       iYRes,
                       iXInternalRes,
                       iYInternalRes,
                       pbdData,
                       iCapabilities,
                       iDestinationBitsPerPel,
                       iScanlineMultiple)
{
   node_d = node;
}

UPDFDeviceResolution::
~UPDFDeviceResolution ()
{
}

DeviceResolution * UPDFDeviceResolution::
createV (int id)
{
   return UPDFDeviceResolution::create (pDevice_d, id);
}

DeviceResolution * UPDFDeviceResolution::
create (Device *pDevice,
        PSZCRO  pszId)
{
   int id = DeviceResolution::nameToID (pszId);

   if (DeviceResolution::RESOLUTION_UNLISTED != id)
   {
      return create (pDevice, id);
   }
   else
   {
      UPDFDevice *pUPDFDevice     = UPDFDevice::isAUPDFDevice (pDevice);
      xmlNodePtr  nodeResolutions = 0;
      int         iXRes           = 0;
      int         iYRes           = 0;

      nodeResolutions = findResolutions (pUPDFDevice);

      if (nodeResolutions)
      {
         xmlNodePtr nodeItem  = firstNode (nodeResolutions->xmlChildrenNode);
         xmlNodePtr nodeFound = 0;

         while (nodeItem != 0)
         {
            PSZCRO pszNodeId = (PSZCRO)xmlGetProp (nodeItem, (xmlChar *)"ID");

            if (pszNodeId)
            {
               if (0 == strcmp (pszNodeId, pszId))
               {
                  nodeFound = nodeItem;
               }

               free ((void *)pszNodeId);
            }

            if (nodeFound)
            {
               id = getOmniResolutionID (nodeFound,
                                         pUPDFDevice,
                                         &iXRes,
                                         &iYRes);

               if (DeviceResolution::RESOLUTION_UNLISTED != id)
               {
                  return create (pDevice, id);
               }

               nodeFound = 0;
            }

            nodeItem = nextNode (nodeItem);
         }
      }

      return 0;
   }
}

static int
getOmniResolutionID (xmlNodePtr   node,
                     UPDFDevice  *pUPDFDevice,
                     int         *piXRes,
                     int         *piYRes)
{
   int id            = DeviceResolution::RESOLUTION_UNLISTED;
   int iVirtualUnits = pUPDFDevice->getYVirtualUnits ();

   if (iVirtualUnits == 0)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputUPDFDeviceResolution ()) DebugOutput::getErrorStream () << "UPDFDeviceResolution::" << __FUNCTION__ << ": Error: Could not find virtual units @ " << __LINE__ << std::endl;
#endif

      return id;
   }

   if (!node)
   {
      return id;
   }

   PSZRO pszProprietaryID = (PSZRO)xmlGetProp (node, (xmlChar *)"Proprietary_ID");

   if (!pszProprietaryID)
   {
      pszProprietaryID = (PSZRO)xmlGetProp (node, (xmlChar *)"Predefined_ID");
   }

   if (pszProprietaryID)
   {
      int iXRes = 0,
          iYRes = 0,
          iRet  = 0;

      iRet = sscanf (pszProprietaryID, "deviceResolution_%dx%d", &iXRes, &iYRes);

#ifndef RETAIL
      if (DebugOutput::shouldOutputUPDFDeviceResolution ()) DebugOutput::getErrorStream () << "UPDFDeviceResolution::" << __FUNCTION__ << ": iXRes = " << iXRes << std::endl;
      if (DebugOutput::shouldOutputUPDFDeviceResolution ()) DebugOutput::getErrorStream () << "UPDFDeviceResolution::" << __FUNCTION__ << ": iYRes = " << iYRes << std::endl;
      if (DebugOutput::shouldOutputUPDFDeviceResolution ()) DebugOutput::getErrorStream () << "UPDFDeviceResolution::" << __FUNCTION__ << ": iRet  = " << iRet << std::endl;
#endif

      if (  iXRes
         && iYRes
         )
      {
         static char achName[26]; // @TBD

#if 0 // @HACK
         iXRes = iVirtualUnits / iXRes;
         iYRes = iVirtualUnits / iYRes;
#endif

         sprintf (achName, "%dx%d", iXRes, iYRes);

#ifndef RETAIL
         if (DebugOutput::shouldOutputUPDFDeviceResolution ()) DebugOutput::getErrorStream () << "UPDFDeviceResolution::" << __FUNCTION__ << ": achName = " << achName << std::endl;
#endif

         id = DeviceResolution::nameToID (achName);

         *piXRes = iXRes;
         *piYRes = iYRes;
      }
   }

   if (pszProprietaryID)
   {
      free ((void *)pszProprietaryID);
   }

   return id;
}

static xmlNodePtr
skipInvalidResolutions (xmlNodePtr  node,
                        UPDFDevice *pUPDFDevice,
                        int        *piXRes,
                        int        *piYRes)
{
   if (!node)
   {
      return 0;
   }

   int id = getOmniResolutionID (node, pUPDFDevice, piXRes, piYRes);

   while (  node
           && (  id == DeviceResolution::RESOLUTION_UNLISTED
              || id == DeviceResolution::RESOLUTION_NONE
              )
           )
   {
      node = nextNode (node);
      id   = getOmniResolutionID (node, pUPDFDevice, piXRes, piYRes);
   }

   return node;
}

static DeviceResolution *
createFromXMLNode (Device      *pDevice,
                   xmlNodePtr   node,
                   int          id,
                   int          iXRes,
                   int          iYRes)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   int          iXInternalRes          = 0;
   int          iYInternalRes          = 0;
   BinaryData  *pbdData                = 0;
   int          iCapabilities          = 0;
   int          iDestinationBitsPerPel = 0;
   int          iScanlineMultiple      = 1;

   return new UPDFDeviceResolution (pUPDFDevice,
                                    id,
                                    iXRes,
                                    iYRes,
                                    iXInternalRes,
                                    iYInternalRes,
                                    pbdData,
                                    iCapabilities,
                                    iDestinationBitsPerPel,
                                    iScanlineMultiple,
                                    node);
}

DeviceResolution * UPDFDeviceResolution::
create (Device *pDevice, int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceResolution ()) DebugOutput::getErrorStream () << "UPDFDeviceResolution::" << __FUNCTION__ << ": pDevice = " << pDevice << ", id = " << id << std::endl;
#endif

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeResolutions = findResolutions (pUPDFDevice);

   if (nodeResolutions)
   {
      xmlNodePtr nodeItem = firstNode (nodeResolutions->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iXRes         = 0;
         int iYRes         = 0;
         int iResolutionID = getOmniResolutionID (nodeItem,
                                                  pUPDFDevice,
                                                  &iXRes,
                                                  &iYRes);

         if (  id == iResolutionID
            && !(  id == DeviceResolution::RESOLUTION_UNLISTED
                || id == DeviceResolution::RESOLUTION_NONE
                )
            )
         {
            return createFromXMLNode (pDevice,
                                      nodeItem,
                                      id,
                                      iXRes,
                                      iYRes);
         }

         nodeItem = nextNode (nodeItem);
      }

      return 0;
   }
   else
   {
      return 0;
   }
}

PSZCRO UPDFDeviceResolution::
getDeviceID ()
{
   return 0;
}

bool UPDFDeviceResolution::
isResolutionSupported (int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeResolutions = 0;

   nodeResolutions = findResolutions (pUPDFDevice);

   if (nodeResolutions)
   {
      xmlNodePtr nodeItem = firstNode (nodeResolutions->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iXRes         = 0;
         int iYRes         = 0;
         int iResolutionID = getOmniResolutionID (nodeItem,
                                                  pUPDFDevice,
                                                  &iXRes,
                                                  &iYRes);

         if (  id == iResolutionID
            && !(  id == DeviceResolution::RESOLUTION_UNLISTED
                || id == DeviceResolution::RESOLUTION_NONE
                )
            )
         {
            return true;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return false;
}

class UPDFResolution_Enumerator : public Enumeration
{
public:
   UPDFResolution_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      int iXRes = 0;
      int iYRes = 0;

      pDevice_d     = pDevice;
      pUPDFDevice_d = UPDFDevice::isAUPDFDevice (pDevice);
      nodeItem_d    = skipInvalidResolutions (nodeItem,
                                              pUPDFDevice_d,
                                              &iXRes,
                                              &iYRes);
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      void *pvRet = 0;

      if (nodeItem_d)
      {
         int iXRes         = 0;
         int iYRes         = 0;
         int iResolutionID = getOmniResolutionID (nodeItem_d,
                                                  pUPDFDevice_d,
                                                  &iXRes,
                                                  &iYRes);

         pvRet = createFromXMLNode (pDevice_d,
                                    nodeItem_d,
                                    iResolutionID,
                                    iXRes,
                                    iYRes);

         nodeItem_d = nextNode (nodeItem_d);
         nodeItem_d = skipInvalidResolutions (nodeItem_d,
                                              pUPDFDevice_d,
                                              &iXRes,
                                              &iYRes);
      }

      return pvRet;
   }

private:
   Device      *pDevice_d;
   UPDFDevice  *pUPDFDevice_d;
   xmlNodePtr   nodeItem_d;
};

Enumeration * UPDFDeviceResolution::
getEnumeration ()
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDFResolution_Enumerator (pDevice_d, 0);
   }

   xmlNodePtr nodeResolutions = 0;

   nodeResolutions = findResolutions (pUPDFDevice);

   if (nodeResolutions)
   {
      xmlNodePtr nodeItem = firstNode (nodeResolutions->xmlChildrenNode);

      return new UPDFResolution_Enumerator (pDevice_d, nodeItem);
   }
   else
   {
      return new UPDFResolution_Enumerator (pDevice_d, 0);
   }
}

xmlNodePtr UPDFDeviceResolution::
findResolutions (UPDFDevice *pUPDFDevice)
{
   xmlNodePtr nodeResolutions = 0;

   if (!pUPDFDevice)
   {
      return nodeResolutions;
   }

   if (  ((nodeResolutions = FINDUDRENTRY (pUPDFDevice, nodeResolutions, "Device_Features",   UPDFDeviceResolution)) != 0)
      && ((nodeResolutions = FINDUDRENTRY (pUPDFDevice, nodeResolutions, "PrintCapabilities", UPDFDeviceResolution)) != 0)
      && ((nodeResolutions = FINDUDRENTRY (pUPDFDevice, nodeResolutions, "Features",          UPDFDeviceResolution)) != 0)
      && ((nodeResolutions = FINDUDRENTRY (pUPDFDevice, nodeResolutions, "DeviceResolution",  UPDFDeviceResolution)) != 0)
      )
      return nodeResolutions;

   return nodeResolutions;
}

std::string UPDFDeviceResolution::
toString (std::ostringstream& oss)
{
   std::ostringstream oss2;

   oss << "{"
       << DeviceResolution::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
std::ostream&
operator<< (std::ostream& os, const UPDFDeviceResolution& const_self)
{
   UPDFDeviceResolution& self = const_cast<UPDFDeviceResolution&>(const_self);
   std::ostringstream    oss;

   os << self.toString (oss);

   return os;
}
