/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <UPDFDeviceSide.hpp>

/*
*/
static int     getOmniSidesID             (xmlNodePtr      node);

UPDFDeviceSide::
UPDFDeviceSide (Device      *pDevice,
                PSZRO        pszJobProperties,
                BinaryData  *data,
                xmlNodePtr   node,
                bool         fSimulationRequired)
   : DeviceSide (pDevice,
                 pszJobProperties,
                 data,
                 fSimulationRequired)
{
   node_d = node;
}

UPDFDeviceSide::
~UPDFDeviceSide ()
{
}

DeviceSide * UPDFDeviceSide::
create (Device *pDevice, PSZCRO pszId)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);
   xmlNodePtr  nodeSidess = 0;

   nodeSidess = findSides (pUPDFDevice);

   if (nodeSidess)
   {
      xmlNodePtr nodeItem  = firstNode (nodeSidess->xmlChildrenNode);
      xmlNodePtr nodeFound = 0;

      while (nodeItem != 0)
      {
         PSZCRO pszNodeId = (PSZCRO)xmlGetProp (nodeItem, (xmlChar *)"ID");

         if (pszNodeId)
         {
            if (0 == strcmp (pszNodeId, pszId))
            {
               nodeFound = nodeItem;
            }

            free ((void *)pszNodeId);
         }

         if (nodeFound)
         {
            int id = getOmniSidesID (nodeFound);

            if (DeviceSide::SIDES_UNLISTED != id)
            {
               return create (pDevice, id);
            }

            nodeFound = 0;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return 0;
}

static int
getOmniSidesID (xmlNodePtr node)
{
   int id = DeviceSide::SIDES_UNLISTED;

   if (!node)
   {
      return id;
   }

   PSZCRO pszID = (PSZCRO)xmlGetProp (node, (xmlChar *)"Predefined_ID");

   if (pszID)
   {
      id = DeviceSide::nameToID (pszID);

      free ((void *)pszID);
   }

   return id;
}

static xmlNodePtr
skipInvalidSidess (xmlNodePtr node)
{
   if (!node)
   {
      return 0;
   }

   int id = getOmniSidesID (node);

   while (  node
         && (id == DeviceSide::SIDES_UNLISTED)
         )
   {
      node = nextNode (node);
      id   = getOmniSidesID (node);
   }

   return node;
}

static DeviceSide *
createFromXMLNode (Device      *pDevice,
                   xmlNodePtr   node,
                   int          id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

   if (!pUPDFDevice)
   {
      return 0;
   }

   bool   fDeviceFeature   = true;
   PSZCRO pszDeviceFeature = (PSZCRO)xmlGetProp (node,
                                                 (xmlChar *)"DeviceFeature");

   if (pszDeviceFeature)
   {
      if (0 == strcasecmp (pszDeviceFeature, "false"))
      {
         fDeviceFeature = false;
      }

      free ((void *)pszDeviceFeature);
   }

   BinaryData  *data = 0; // @TBD

   return new UPDFDeviceSide (pUPDFDevice,
                                id,
                                data,
                                node,
                                !fDeviceFeature);
}

DeviceSide * UPDFDeviceSide::
create (Device *pDevice, int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice);

#ifndef RETAIL
   if (DebugOutput::shouldOutputUPDFDeviceSides ()) DebugOutput::getErrorStream () << "UPDFDeviceSides::" << __FUNCTION__ << ": pDevice = " << pDevice << ", id = " << id << std::endl;
#endif

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeSidess = 0;

   nodeSidess = findSides (pUPDFDevice);

   if (nodeSidess)
   {
      xmlNodePtr nodeItem = firstNode (nodeSidess->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iSidesID = getOmniSidesID (nodeItem);

         if (  id == iSidesID
            && !(id == DeviceSide::SIDES_UNLISTED)
            )
         {
            return createFromXMLNode (pDevice, nodeItem, id);
         }

         nodeItem = nextNode (nodeItem);
      }

      return 0;
   }
   else
   {
      return 0;
   }
}

PSZCRO UPDFDeviceSide::
getDeviceID ()
{
   return 0;
}

bool UPDFDeviceSide::
isSidesSupported (int id)
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return 0;
   }

   xmlNodePtr nodeSidess = 0;

   nodeSidess = findSides (pUPDFDevice);

   if (nodeSidess)
   {
      xmlNodePtr nodeItem = firstNode (nodeSidess->xmlChildrenNode);

      while (nodeItem != 0)
      {
         int iSidesID = getOmniSidesID (nodeItem);

         if (  id == iSidesID
            && !(id == DeviceSide::SIDES_UNLISTED)
            )
         {
            return true;
         }

         nodeItem = nextNode (nodeItem);
      }
   }

   return false;
}

class UPDFSides_Enumerator : public Enumeration
{
public:
   UPDFSides_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      pDevice_d  = pDevice;
      nodeItem_d = skipInvalidSidess (nodeItem);
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      void *pvRet = 0;

      if (nodeItem_d)
      {
         int id = getOmniSidesID (nodeItem_d);

         pvRet = createFromXMLNode (pDevice_d,
                                    nodeItem_d,
                                    id);

         nodeItem_d = nextNode (nodeItem_d);
         nodeItem_d = skipInvalidSidess (nodeItem_d);
      }

      return pvRet;
   }

private:
   Device     *pDevice_d;
   xmlNodePtr  nodeItem_d;
};

Enumeration * UPDFDeviceSide::
getEnumeration ()
{
   UPDFDevice *pUPDFDevice = UPDFDevice::isAUPDFDevice (pDevice_d);

   if (!pUPDFDevice)
   {
      return new UPDFSides_Enumerator (pDevice_d, 0);
   }

   xmlNodePtr nodeSidess = 0;

   nodeSidess = findSides (pUPDFDevice);

   if (nodeSidess)
   {
      xmlNodePtr nodeItem = firstNode (nodeSidess->xmlChildrenNode);

      return new UPDFSides_Enumerator (pDevice_d, nodeItem);
   }
   else
   {
      return new UPDFSides_Enumerator (pDevice_d, 0);
   }
}

xmlNodePtr UPDFDeviceSide::
findSides (UPDFDevice *pUPDFDevice)
{
   xmlNodePtr nodeSidess = 0;

   if (!pUPDFDevice)
   {
      return nodeSidess;
   }

   if (  ((nodeSidess = FINDUDRENTRY (pUPDFDevice, nodeSidess, "Device_Features",    UPDFDeviceSides)) != 0)
      && ((nodeSidess = FINDUDRENTRY (pUPDFDevice, nodeSidess, "PrintCapabilities",  UPDFDeviceSides)) != 0)
      && ((nodeSidess = FINDUDRENTRY (pUPDFDevice, nodeSidess, "Features",           UPDFDeviceSides)) != 0)
      && ((nodeSidess = FINDUDRENTRY (pUPDFDevice, nodeSidess, "PrintMediaHandling", UPDFDeviceSides)) != 0)
      && ((nodeSidess = FINDUDRENTRY (pUPDFDevice, nodeSidess, "MediaSidesList",    UPDFDeviceSides)) != 0)
      )
      return nodeSidess;

   return nodeSidess;
}

std::string UPDFDeviceSide::
toString (std::ostringstream& oss)
{
   std::ostringstream oss2;

   oss << "{"
       << DeviceSide::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
std::ostream&
operator<< (std::ostream& os, const UPDFDeviceSide& const_self)
{
   UPDFDeviceSide&    self = const_cast<UPDFDeviceSide&>(const_self);
   std::ostringstream oss;

   os << self.toString (oss);

   return os;
}
