/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <stdarg.h>

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/valid.h>
#include <libxml/xmlschemas.h>

#include <glob.h>

#include "defines.hpp"

const static bool vfDebug        = false;
const static bool vfDumpFile     = false;
const static bool vfDumpElements = true;

void
printXMLNode (xmlNodePtr elm)
{
   xmlElemDump (stdout, elm->doc, elm);
   fprintf (stdout, "\n");
}

xmlNodePtr
firstNode (xmlNodePtr elm)
{
   while (  0 != elm
         && (  XML_COMMENT_NODE == elm->type
            || XML_TEXT_NODE    == elm->type
            )
         )
   {
      // Move to the next node
      elm = elm->next;
   }

   return elm;
}

xmlNodePtr
nextNode (xmlNodePtr elm)
{
   do
   {
      // Move to the next node
      elm = elm->next;

   } while (  0 != elm
           && (  XML_COMMENT_NODE == elm->type
              || XML_TEXT_NODE    == elm->type
              )
           );

   return elm;
}

char *
bundleStringData (xmlNodePtr nodeItem)
{
   return (char *)xmlNodeListGetString (nodeItem->doc, nodeItem->children, 1);
}

void
MyValidityErrorFunc (void *ctx, const char *msg, ...)
{
   va_list list;

   va_start (list, msg);

   printf ("Error: ");
   vprintf (msg, list);

   va_end (list);

   if (ctx)
   {
#if 0
      OmniDomParser *myCtx = (OmniDomParser *)ctx;

      myCtx->setErrorCondition ();
#endif
   }
}

void
MyValidityWarningFunc (void *ctx, const char *msg, ...)
{
   va_list list;

   va_start (list, msg);

   printf ("Warning: ");
   vprintf (msg, list);

   va_end (list);
}

const xmlChar *
mapFormName (const xmlChar *pszFrom)
{
   typedef struct _Mapping {
      PSZCRO pszFrom;
      PSZCRO pszTo;
   } MAPPING;

   static MAPPING apszNames[] = {
      { "FORM_10_X_11",                 "na_10x11_10.00x11.00in"                     },
      { "FORM_11_X_12",                 "na_11x12_11.00x12.00in"                     },
      { "FORM_11_X_15",                 "na_11x15_11.00x15.00in"                     },
      { "FORM_12_X_19",                 "na_12x19_12.00x19.00in"                     },
      { "FORM_15_X_11",                 "na_15x11_15.00x11.00in"                     },
      { "FORM_170_X_210",               "na_170x210_1700.00x2100.00mm"               },
      { "FORM_182_X_210",               "na_180x210_1820.00x2100.00mm"               },
      { "FORM_2A0",                     "iso_2a0_11890.00x16820.00mm"                },
      { "FORM_3_X_5_CARD",              "na_3x5-card_3.00x5.00in"                    },
      { "FORM_4A0",                     "iso_4a0_16820.00x23780.00mm"                },
      { "FORM_4_X_6_CARD",              "na_4x6-card_4.00x6.00in"                    },
      { "FORM_5_X_7",                   "na_5x7_5.00x7.00in"                         },
      { "FORM_5_X_8_CARD",              "na_5x8-card_5.00x8.00in"                    },
      { "FORM_8_25_X_13",               "na_8.25x13_8.25x13.00in"                    },
      { "FORM_8_5_X_12_4",              "na_8.5x12.4_8.50x12.40in"                   },
      { "FORM_8_X_10_5",                "na_8x10.5_8.00x10.50in"                     },
      { "FORM_8_X_10_CARD",             "na_8x10-card_8.00x10.00in"                  },
      { "FORM_A0",                      "iso_a0_8410.00x11890.00mm"                  },
      { "FORM_A1",                      "iso_a1_5940.00x8410.00mm"                   },
      { "FORM_A10",                     "iso_a10_260.00x370.00mm"                    },
      { "FORM_A2",                      "iso_a2_4200.00x5940.00mm"                   },
      { "FORM_A2_ENVELOPE",             "na_a2_4.38x5.75in"                          },
      { "FORM_A3",                      "iso_a3_2970.00x4200.00mm"                   },
      { "FORM_A3_EXTRA",                "iso_a3-extra_3220.00x4450.00mm"             },
      { "FORM_A3_WIDE",                 "iso_a3-wide_3302.00x4826.00mm"              },
      { "FORM_A4",                      "iso_a4_2100.00x2970.00mm"                   },
      { "FORM_A4_EXTRA",                "iso_a4-extra_2355.00x3223.00mm"             },
      { "FORM_A4_TAB",                  "iso_a4-tab_2250.00x2970.00mm"               },
      { "FORM_A4_WIDE",                 "iso_a4-wide_2235.00x3556.00mm"              },
      { "FORM_A5",                      "iso_a5_1480.00x2100.00mm"                   },
      { "FORM_A5_EXTRA",                "iso_a5-extra_1740.00x2350.00mm"             },
      { "FORM_A6",                      "iso_a6_1050.00x1480.00mm"                   },
      { "FORM_A6_CARD",                 "na_a6-card_4.13x5.83in"                     },
      { "FORM_A7",                      "iso_a7_740.00x1050.00mm"                    },
      { "FORM_A8",                      "iso_a8_520.00x740.00mm"                     },
      { "FORM_A9",                      "iso_a9_370.00x520.00mm"                     },
      { "FORM_B0",                      "iso_b0_10000.00x14140.00mm"                 },
      { "FORM_B1",                      "iso_b1_7070.00x10000.00mm"                  },
      { "FORM_B10",                     "iso_b10_310.00x440.00mm"                    },
      { "FORM_B2",                      "iso_b2_5000.00x7070.00mm"                   },
      { "FORM_B3",                      "iso_b3_3530.00x5000.00mm"                   },
      { "FORM_B4",                      "iso_b4_2500.00x3530.00mm"                   },
      { "FORM_B5",                      "iso_b5_1760.00x2500.00mm"                   },
      { "FORM_B5_EXTRA",                "iso_b5-extra_2010.00x2760.00mm"             },
      { "FORM_B6",                      "iso_b6_1250.00x1760.00mm"                   },
      { "FORM_B6_C4",                   "iso_b6c4_1250.00x3240.00mm"                 },
      { "FORM_B7",                      "iso_b7_880.00x1250.00mm"                    },
      { "FORM_B8",                      "iso_b8_620.00x880.00mm"                     },
      { "FORM_B9",                      "iso_b9_440.00x620.00mm"                     },
      { "FORM_B_PLUS",                  "na_b-plus_12.00x19.17in"                    },
      { "FORM_C0",                      "iso_c0_9170.00x12970.00mm"                  },
      { "FORM_C1",                      "iso_c1_6480.00x9170.00mm"                   },
      { "FORM_C10",                     "iso_c10_280.00x400.00mm"                    },
      { "FORM_C10_ENVELOPE",            "na_c10_4.12x9.50in"                         },
      { "FORM_C2",                      "iso_c2_4580.00x6480.00mm"                   },
      { "FORM_C3",                      "iso_c3_3240.00x4580.00mm"                   },
      { "FORM_C4",                      "iso_c4_2290.00x3240.00mm"                   },
      { "FORM_C5",                      "iso_c5_1620.00x2290.00mm"                   },
      { "FORM_C5_ENVELOPE",             "na_c5_6.38x9.02in"                          },
      { "FORM_C6",                      "iso_c6_1140.00x1620.00mm"                   },
      { "FORM_C6_C5",                   "iso_c6c5_1140.00x2290.00mm"                 },
      { "FORM_C6_ENVELOPE",             "na_c6_4.49x6.38in"                          },
      { "FORM_C7",                      "iso_c7_810.00x1140.00mm"                    },
      { "FORM_C7_C6",                   "iso_c7c6_810.00x1620.00mm"                  },
      { "FORM_C7_ENVELOPE",             "na_c7_3.87x7.50in"                          },
      { "FORM_C8",                      "iso_c8_570.00x810.00mm"                     },
      { "FORM_C9",                      "iso_c9_400.00x570.00mm"                     },
      { "FORM_C9_ENVELOPE",             "na_c9_3.87x8.87in"                          },
      { "FORM_CARD_148",                "na_card-148_1480.00x1050.00mm"              },
      { "FORM_CHOU2_ENVELOPE",          "jpn_chou2_1111.00x1460.00mm"                },
      { "FORM_CHOU3_ENVELOPE",          "jpn_chou3_1200.00x2350.00mm"                },
      { "FORM_CHOU4_ENVELOPE",          "jpn_chou4_900.00x2050.00mm"                 },
      { "FORM_D5_ENVELOPE",             "na_d5_6.93x9.84in"                          },
      { "FORM_DAI_PA_KAI",              "om_dai-pa-kai_2750.00x3950.00mm"            },
      { "FORM_DISK_LABELS",             "na_disk-labels_2.13x2.76in"                 },
      { "FORM_DL_ENVELOPE",             "na_dl_4.33x8.66in"                          },
      { "FORM_E1",                      "iso_e1_7112.00x10160.00mm"                  },
      { "FORM_EDP",                     "na_edp_11.00x14.00in"                       },
      { "FORM_ENVELOPE_10_X_13",        "na_envelope-10x13_10.00x13.00in"            },
      { "FORM_ENVELOPE_10_X_14",        "na_envelope-10x14_10.00x14.00in"            },
      { "FORM_ENVELOPE_10_X_15",        "na_envelope-10x15_10.00x15.00in"            },
      { "FORM_ENVELOPE_132_X_220",      "na_envelope-132x220_5.20x8.66in"            },
      { "FORM_ENVELOPE_6_1_2",          "na_envelope-6.5_6.50x3.62in"                },
      { "FORM_ENVELOPE_6_X_9",          "na_envelope-6x9_6.00x9.00in"                },
      { "FORM_ENVELOPE_7_X_9",          "na_envelope-7x9_7.00x9.00in"                },
      { "FORM_ENVELOPE_9_X_11",         "na_envelope-9x11_9.00x11.00in"              },
      { "FORM_EUROPEAN_EDP",            "na_eur-edp_11.57x14.00in"                   },
      { "FORM_EUROPEAN_FANFOLD",        "na_fanfold-eur_8.50x12.00in"                },
      { "FORM_EURO_LABELS",             "na_euro-labels_1.42x3.50in"                 },
      { "FORM_EXECUTIVE",               "na_executive_7.25x10.50in"                  },
      { "FORM_FANFOLD_1",               "na_fanfold-1_14.50x11.00in"                 },
      { "FORM_FANFOLD_2",               "na_fanfold-2_12.00x8.50in"                  },
      { "FORM_FANFOLD_3",               "na_fanfold-3_9.50x11.00in"                  },
      { "FORM_FANFOLD_4",               "na_fanfold-4_8.50x12.00in"                  },
      { "FORM_FANFOLD_5",               "na_fanfold-5_8.50x11.00in"                  },
      { "FORM_FOLIO",                   "om_folio_2159.00x3300.00mm"                 },
      { "FORM_FOLIO_SP",                "om_folio-sp_2150.00x3150.00mm"              },
      { "FORM_FOOLSCAP",                "na_foolscap_8.00x13.00in"                   },
      { "FORM_FOOLSCAP_WIDE",           "na_foolscap-wide_8.50x13.00in"              },
      { "FORM_GERMAN_12_X_250_FANFOLD", "na_german-12x250-fanfold_3048.00x2400.00mm" },
      { "FORM_GERMAN_LEGAL_FANFOLD",    "na_german-legal-fanfold_8.50x13.00in"       },
      { "FORM_GOVERNMENT_LEGAL",        "na_govt-letter_8.00x13.00in"                },
      { "FORM_GOVERNMENT_LETTER",       "na_govt-legal_8.00x10.00in"                 },
      { "FORM_HAGAKI_CARD",             "jpn_hagaki_1000.00x1480.00mm"               },
      { "FORM_HALF_LETTER",             "na_half-letter_5.50x8.50in"                 },
      { "FORM_INDEX_4_X_6_EXT",         "na_index-4x6-ext_6.00x8.00in"               },
      { "FORM_INVITE_ENVELOPE",         "om_invite_2200.00x2200.00mm"                },
      { "FORM_ITALIAN_ENVELOPE",        "om_italian_1000.00x2300.00mm"               },
      { "FORM_JIS_B0",                  "jis_b0_10300.00x14560.00mm"                 },
      { "FORM_JIS_B1",                  "jis_b1_7280.00x10300.00mm"                  },
      { "FORM_JIS_B10",                 "jis_b10_320.00x450.00mm"                    },
      { "FORM_JIS_B2",                  "jis_b2_5150.00x7280.00mm"                   },
      { "FORM_JIS_B3",                  "jis_b3_3640.00x5150.00mm"                   },
      { "FORM_JIS_B4",                  "jis_b4_2570.00x3640.00mm"                   },
      { "FORM_JIS_B5",                  "jis_b5_1820.00x2570.00mm"                   },
      { "FORM_JIS_B6",                  "jis_b6_1280.00x1820.00mm"                   },
      { "FORM_JIS_B7",                  "jis_b7_910.00x1280.00mm"                    },
      { "FORM_JIS_B8",                  "jis_b8_640.00x910.00mm"                     },
      { "FORM_JIS_B9",                  "jis_b9_450.00x640.00mm"                     },
      { "FORM_JIS_EXEC",                "jis_exec_2160.00x3300.00mm"                 },
      { "FORM_JUURO_KU_KAI",            "om_juuro-ku-kai_1980.00x2750.00mm"          },
      { "FORM_KAHU_ENVELOPE",           "jpn_kahu_2400.00x3221.00mm"                 },
      { "FORM_KAKU2_ENVELOPE",          "jpn_kaku2_2400.00x3320.00mm"                },
      { "FORM_LEDGER",                  "na_ledger_17.00x11.00in"                    },
      { "FORM_LEGAL",                   "na_legal_8.50x14.00in"                      },
      { "FORM_LEGAL_EXTRA",             "na_legal-extra_9.50x15.00in"                },
      { "FORM_LETTER",                  "na_letter_8.50x11.00in"                     },
      { "FORM_LETTER_EXTRA",            "na_letter-extra_9.50x12.00in"               },
      { "FORM_LETTER_PLUS",             "na_letter-plus_8.50x12.69in"                },
      { "FORM_LETTER_WIDE",             "na_letter-wide_8.99x13.30in"                },
      { "FORM_MONARCH_ENVELOPE",        "na_monarch_3.88x7.50in"                     },
      { "FORM_NUMBER_10_ENVELOPE",      "na_number-10_4.12x9.50in"                   },
      { "FORM_NUMBER_11_ENVELOPE",      "na_number-11_4.50x10.38in"                  },
      { "FORM_NUMBER_12_ENVELOPE",      "na_number-12_4.75x11.00in"                  },
      { "FORM_NUMBER_14_ENVELOPE",      "na_number-14_5.00x11.50in"                  },
      { "FORM_NUMBER_9_ENVELOPE",       "na_number-9_3.88x8.88in"                    },
      { "FORM_OUFUKU_CARD",             "jpn_oufuku_1480.00x2000.00mm"               },
      { "FORM_PANORAMIC",               "na_panoramic_2100.00x5940.00mm"             },
      { "FORM_PA_KAI",                  "om_pa-kai_2670.00x3890.00mm"                },
      { "FORM_PERSONAL_ENVELOPE",       "na_personal_3.62x6.50in"                    },
      { "FORM_PHOTO_100_150",           "na_photo-100x150_1000.00x1500.00mm"         },
      { "FORM_PHOTO_200_300",           "na_photo-200x300_2000.00x3000.00mm"         },
      { "FORM_PHOTO_4_6",               "na_photo-4x6_4.00x6.00in"                   },
      { "FORM_PLOTTER_SIZE_A",          "na_plotter-size-a_8.50x11.00in"             },
      { "FORM_PLOTTER_SIZE_B",          "na_plotter-size-b_11.00x17.00in"            },
      { "FORM_PLOTTER_SIZE_C",          "na_plotter-size-c_17.00x22.00in"            },
      { "FORM_PLOTTER_SIZE_D",          "na_plotter-size-d_22.00x34.00in"            },
      { "FORM_PLOTTER_SIZE_E",          "na_plotter-size-e_34.00x44.00in"            },
      { "FORM_PLOTTER_SIZE_F",          "na_plotter-size-f_44.00x68.00in"            },
      { "FORM_POSTCARD",                "na_postcard_3.93x5.79in"                    },
      { "FORM_POSTFIX_ENVELOPE",        "om_postfix_1140.00x2290.00mm"               },
      { "FORM_PRC10_ENVELOPE",          "prc_10_3240.00x4580.00mm"                   },
      { "FORM_PRC1_ENVELOPE",           "prc_1_1020.00x1650.00mm"                    },
      { "FORM_PRC2_ENVELOPE",           "prc_2_1020.00x1760.00mm"                    },
      { "FORM_PRC3_ENVELOPE",           "prc_3_1250.00x1760.00mm"                    },
      { "FORM_PRC4_ENVELOPE",           "prc_4_1100.00x2080.00mm"                    },
      { "FORM_PRC5_ENVELOPE",           "prc_5_1100.00x2200.00mm"                    },
      { "FORM_PRC6_ENVELOPE",           "prc_6_1200.00x3200.00mm"                    },
      { "FORM_PRC7_ENVELOPE",           "prc_7_1600.00x2300.00mm"                    },
      { "FORM_PRC8_ENVELOPE",           "prc_8_1200.00x3090.00mm"                    },
      { "FORM_PRC9_ENVELOPE",           "prc_9_2290.00x3240.00mm"                    },
      { "FORM_PRC_16K",                 "prc_16k_1460.00x2150.00mm"                  },
      { "FORM_PRC_32K",                 "prc_32k_970.00x1510.00mm"                   },
      { "FORM_QUARTO",                  "na_quarto_8.50x10.83in"                     },
      { "FORM_RA0",                     "iso_ra0_8600.00x12200.00mm"                 },
      { "FORM_RA1",                     "iso_ra1_6100.00x8600.00mm"                  },
      { "FORM_RA2",                     "iso_ra2_4300.00x6100.00mm"                  },
      { "FORM_ROC_16K",                 "roc_16k_1968.50x2730.50mm"                  },
      { "FORM_ROC_8K",                  "roc_8k_2730.50x3937.00mm"                   },
      { "FORM_SHIPPING_LABELS",         "na_shipping-labels_2.13x3.98in"             },
      { "FORM_SRA0",                    "iso_sra0_9000.00x12800.00mm"                },
      { "FORM_SRA1",                    "iso_sra1_6400.00x9000.00mm"                 },
      { "FORM_SRA2",                    "iso_sra2_4500.00x6400.00mm"                 },
      { "FORM_STANDARD_LABELS_CLEAR",   "na_standard-labels-clear_1.10x3.50in"       },
      { "FORM_STANDARD_LABELS_WHITE",   "na_standard-labels-white_1.10x3.50in"       },
      { "FORM_STATEMENT",               "na_statement_5.50x8.50in"                   },
      { "FORM_SUPER_A",                 "na_super-a_2270.76x3556.00mm"               },
      { "FORM_SUPER_A3_B",              "na_super-a3-b_3289.20x4830.20mm"            },
      { "FORM_SUPER_B",                 "na_super-b_3302.00x4826.00mm"               },
      { "FORM_TABLOID",                 "na_tabloid_11.00x17.00in"                   },
      { "FORM_UNIVERSAL",               "na_universal_11.69x17.00in"                 },
      { "FORM_US_A_ARCHITECTURAL",      "na_arch-a_9.00x12.00in"                     },
      { "FORM_US_B_ARCHITECTURAL",      "na_arch-b_12.00x18.00in"                    },
      { "FORM_US_C_ARCHITECTURAL",      "na_arch-c_18.00x24.00in"                    },
      { "FORM_US_D_ARCHITECTURAL",      "na_arch-d_24.00x36.00in"                    },
      { "FORM_US_E_ARCHITECTURAL",      "na_arch-e_36.00x48.00in"                    },
      { "FORM_US_FANFOLD",              "na_fanfold-us_14.88x11.00in"                },
      { "FORM_WIDE",                    "na_wide-format_13.58x11.00in"               },
      { "FORM_YOU4_ENVELOPE",           "jpn_you4_1050.00x2350.00mm"                 },
   };

   if (0 == strncmp ((char *)pszFrom, "FORM_ROLL_69_5MM", 16))
   {
      static char  achHack[64];
      const char  *pszComma     = 0;
      float        flY          = 0.0;
      bool         fSuccess     = false;

      pszComma = strchr ((char *)pszFrom, ',');

      if (pszComma)
      {
         if (1 == sscanf (pszComma + 1, "%f", &flY))
         {
            fSuccess = true;
         }
      }
      else
      {
         flY = 25400.0;

         fSuccess = true;
      }

      if (fSuccess)
      {
         flY /= 1000.0;

         sprintf (achHack, "na_roll-69.5_69.50x%1.2fmm", flY);

         return (const xmlChar *)achHack;
      }
   }
   else if (0 == strncmp ((char *)pszFrom, "FORM_ROLL_76_2MM", 16))
   {
      static char  achHack[64];
      const char  *pszComma     = 0;
      float        flY          = 0.0;
      bool         fSuccess     = false;

      pszComma = strchr ((char *)pszFrom, ',');

      if (pszComma)
      {
         if (1 == sscanf (pszComma + 1, "%f", &flY))
         {
            fSuccess = true;
         }
      }
      else
      {
         flY = 25400.0;

         fSuccess = true;
      }

      if (fSuccess)
      {
         flY /= 1000.0;

         sprintf (achHack, "na_roll-76.2_76.20x%1.2fmm", flY);

         return (const xmlChar *)achHack;
      }
   }

   int iLow  = 0;
   int iMid  = (int)dimof (apszNames) / 2;
   int iHigh = (int)dimof (apszNames) - 1;
   int iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp ((char *)pszFrom, apszNames[iMid].pszFrom);

      if (0 == iResult)
      {
         return (const xmlChar *)apszNames[iMid].pszTo;
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return 0;
}

const xmlChar *
mapTrayName (const xmlChar *pszFrom)
{
   if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_NONE"))
   {
      return (const xmlChar *)"None";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_AUTO"))
   {
      return (const xmlChar *)"AutoSelect";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_USE_PRINTER_SETTING"))
   {
      return (const xmlChar *)"PanelSelect";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_DEPEND_ON_PANEL"))
   {
      return (const xmlChar *)"PanelSelect";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_AUTO_FEEDER"))
   {
      return (const xmlChar *)"Continuous";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_AUTOSWITCH"))
   {
      return (const xmlChar *)"AutoSelect";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_TRAY"))
   {
      return (const xmlChar *)"Tray";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_UPPER_TRAY"))
   {
      return (const xmlChar *)"Top";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_LOWER_TRAY"))
   {
      return (const xmlChar *)"Bottom";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_MULTI_TRAY"))
   {
      return (const xmlChar *)"AutoSelect";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_CASSETTE"))
   {
      return (const xmlChar *)"Tray";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_UPPER_CASSETTE"))
   {
      return (const xmlChar *)"Top";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_LOWER_CASSETTE"))
   {
      return (const xmlChar *)"Bottom";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_MULTI_CASSETTE"))
   {
      return (const xmlChar *)"AutoSelect";
   }
/*
"IBM/IBM 5577-H02 Connections.xml" 100 23:      <connectionTray>TRAY_OPTION_CASSETTE</connectionTray>
"IBM/IBM 5577-H02 Trays.xml" 41 13:      <name>TRAY_OPTION_CASSETTE</name>
"IBM/IBM 5584-H02 Connections.xml" 100 23:      <connectionTray>TRAY_OPTION_CASSETTE</connectionTray>
"IBM/IBM 5584-H02 Trays.xml" 41 13:      <name>TRAY_OPTION_CASSETTE</name>
"HP LaserJet/HP LaserJet 4V_4MV Trays.xml" 46 13:      <name>TRAY_OPTION_CASSETTE</name>
"HP LaserJet/HP LaserJet 4_4M Trays.xml" 46 13:      <name>TRAY_OPTION_CASSETTE</name>
"HP LaserJet/HP LaserJet 5Si_5Si Mx_5Si Mopier Trays.xml" 46 13:      <name>TRAY_OPTION_CASSETTE</name>
"HP LaserJet/HP LaserJet 5_5M Trays.xml" 46 13:      <name>TRAY_OPTION_CASSETTE</name>
*/
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_OPTION_CASSETTE"))
   {
      return (const xmlChar *)"BypassTray";
   }
/*
Too many to list!
*/
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_MANUAL_FEEDER"))
   {
      return (const xmlChar *)"AutoSelect";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_MANUAL_ENVELOPE"))
   {
      return (const xmlChar *)"Envelope";
   }
/*
None!
*/
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_PAPER_FEEDER"))
   {
      return (const xmlChar *)"AutoSelect";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_ENVELOPE_FEEDER"))
   {
      return (const xmlChar *)"Envelope";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_CSF"))
   {
      return (const xmlChar *)"Continuous";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_FRONT_CONTINUOUS"))
   {
      return (const xmlChar *)"Front";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_REAR_CONTINUOUS"))
   {
      return (const xmlChar *)"Rear";
   }
/*
Too many to list!
*/
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_SINGLE_SHEET"))
   {
      return (const xmlChar *)"AutoSelect";
   }
/*
"HP DeskJet/HP DeskJet Portable Connections.xml" 52 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet Portable Connections.xml" 58 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet Portable Connections.xml" 64 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet Portable Connections.xml" 70 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet 310 Connections.xml" 52 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet 310 Connections.xml" 58 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet 310 Connections.xml" 64 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet 310 Connections.xml" 70 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet 310.xml" 47 13:      <tray>TRAY_SHEET_FEEDER</tray>
"HP DeskJet/HP DeskJet 320.xml" 47 13:      <tray>TRAY_SHEET_FEEDER</tray>
"HP DeskJet/HP DeskJet 340 Connections.xml" 76 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet 340 Connections.xml" 82 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet 340 Connections.xml" 88 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet 340 Connections.xml" 94 23:      <connectionTray>TRAY_SHEET_FEEDER</connectionTray>
"HP DeskJet/HP DeskJet 340 Trays.xml" 36 13:      <name>TRAY_SHEET_FEEDER</name>
"HP DeskJet/HP DeskJet Portable Trays.xml" 31 13:      <name>TRAY_SHEET_FEEDER</name>
"HP DeskJet/HP DeskJet Portable.xml" 45 13:      <tray>TRAY_SHEET_FEEDER</tray>
*/
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_SHEET_FEEDER"))
   {
      return (const xmlChar *)"AutoSelect";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_BIN_1"))
   {
      return (const xmlChar *)"Tray-1";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_BIN_2"))
   {
      return (const xmlChar *)"Tray-2";
   }
/*
"HP DeskJet/HP DeskJet 340 Connections.xml" 28 23:      <connectionTray>TRAY_PORTABLE_SHEET</connectionTray>
"HP DeskJet/HP DeskJet 340 Connections.xml" 34 23:      <connectionTray>TRAY_PORTABLE_SHEET</connectionTray>
"HP DeskJet/HP DeskJet 340 Connections.xml" 40 23:      <connectionTray>TRAY_PORTABLE_SHEET</connectionTray>
"HP DeskJet/HP DeskJet 340 Connections.xml" 46 23:      <connectionTray>TRAY_PORTABLE_SHEET</connectionTray>
"HP DeskJet/HP DeskJet 340 Trays.xml" 26 13:      <name>TRAY_PORTABLE_SHEET</name>
"HP DeskJet/HP DeskJet 340.xml" 47 13:      <tray>TRAY_PORTABLE_SHEET</tray>
*/
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_PORTABLE_SHEET"))
   {
      return (const xmlChar *)"AutoSelect";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_TRAY_1"))
   {
      return (const xmlChar *)"Tray-1";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_TRAY_2"))
   {
      return (const xmlChar *)"Tray-2";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_TRAY_3"))
   {
      return (const xmlChar *)"Tray-3";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_TRAY_4"))
   {
      return (const xmlChar *)"Tray-4";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_TRAY_5"))
   {
      return (const xmlChar *)"Tray-5";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_TRAY_6"))
   {
      return (const xmlChar *)"Tray-6";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_CASSETTE1"))
   {
      return (const xmlChar *)"Tray-1";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_CASSETTE2"))
   {
      return (const xmlChar *)"Tray-2";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_CASSETTE3"))
   {
      return (const xmlChar *)"Tray-3";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_CASSETTE4"))
   {
      return (const xmlChar *)"Tray-4";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_CASSETTE5"))
   {
      return (const xmlChar *)"Tray-5";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_FRONT_TRAY"))
   {
      return (const xmlChar *)"Front";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_OPTION_MULTI_SF"))
   {
      return (const xmlChar *)"AutoSelect";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_MULTI_FEEDER"))
   {
      return (const xmlChar *)"AutoSelect";
   }
/*
None!
*/
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_PAPER_DECK"))
   {
      return (const xmlChar *)"Tray";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_TRACTOR_UNIT"))
   {
      return (const xmlChar *)"Continuous";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_ROLL_1"))
   {
      return (const xmlChar *)"Roll-1";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_ROLL_2"))
   {
      return (const xmlChar *)"Roll-2";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_ROLL_3"))
   {
      return (const xmlChar *)"Roll-3";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_ROLL_4"))
   {
      return (const xmlChar *)"Roll-4";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_ROLL_5"))
   {
      return (const xmlChar *)"Roll-5";
   }
/*
"IBM/IBM Infoprint 20 Connections.xml" 34 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 20 Connections.xml" 82 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 20 Trays.xml" 31 13:      <name>TRAY_AUXILIARY_TRAY</name>
"IBM/IBM Infoprint 21 Connections.xml" 118 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 124 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 130 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 136 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 142 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 148 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 154 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 160 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 166 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 172 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 178 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 184 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 190 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 196 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 202 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 838 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 844 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 850 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 856 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 862 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 868 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 874 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 880 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 886 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 892 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 898 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 904 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 910 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 916 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Connections.xml" 922 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 32 Connections.xml" 34 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 32 Connections.xml" 76 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 21 Trays.xml" 31 13:      <name>TRAY_AUXILIARY_TRAY</name>
"IBM/IBM Infoprint 40 Connections.xml" 34 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 40 Connections.xml" 70 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 32 Trays.xml" 31 13:      <name>TRAY_AUXILIARY_TRAY</name>
"IBM/IBM Infoprint 70 Connections.xml" 136 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 142 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 148 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 154 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 160 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 166 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 172 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 178 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 184 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 190 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 196 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 202 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 208 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 214 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 220 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 226 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 232 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 238 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 892 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 898 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 904 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 910 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 916 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 922 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 928 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 934 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 940 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 946 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 952 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 958 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 964 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 970 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 976 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 982 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 988 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 70 Connections.xml" 994 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Infoprint 40 Trays.xml" 31 13:      <name>TRAY_AUXILIARY_TRAY</name>
"IBM/IBM Infoprint 70 Trays.xml" 31 13:      <name>TRAY_AUXILIARY_TRAY</name>
"IBM/IBM Network Printer 12 Connections.xml" 34 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Network Printer 12 Connections.xml" 70 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Network Printer 12 Trays.xml" 31 13:      <name>TRAY_AUXILIARY_TRAY</name>
"IBM/IBM Network Printer 17 Connections.xml" 34 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Network Printer 17 Connections.xml" 76 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Network Printer 17 Trays.xml" 31 13:      <name>TRAY_AUXILIARY_TRAY</name>
"IBM/IBM Network Printer 24 Connections.xml" 34 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
"IBM/IBM Network Printer 24 Connections.xml" 76 23:      <connectionTray>TRAY_AUXILIARY_TRAY</connectionTray>
*/
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_AUXILIARY_TRAY"))
   {
      return (const xmlChar *)"BypassTray";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_LARGE_CAPACITY_TRAY"))
   {
      return (const xmlChar *)"LargeCapacity";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_HIGH_CAPACITY_FEEDER"))
   {
      return (const xmlChar *)"LargeCapacity";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_DEPEND_ON_PANEL"))
   {
      return (const xmlChar *)"PanelSelect";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"TRAY_ZERO_MARGINS"))
   {
      return (const xmlChar *)"Tray-1";
   }

   return 0;
}

const xmlChar *
mapResolutionName (const xmlChar *pszFrom)
{
   if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_60_X_60"))
   {
      return (const xmlChar *)"60x60";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_60_X_72"))
   {
      return (const xmlChar *)"60x72";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_60_X_180"))
   {
      return (const xmlChar *)"60x180";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_60_X_360"))
   {
      return (const xmlChar *)"60x360";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_72_X_72"))
   {
      return (const xmlChar *)"72x72";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_75_X_75"))
   {
      return (const xmlChar *)"75x75";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_80_X_60"))
   {
      return (const xmlChar *)"80x60";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_80_X_72"))
   {
      return (const xmlChar *)"80x72";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_90_X_180"))
   {
      return (const xmlChar *)"90x180";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_90_X_360"))
   {
      return (const xmlChar *)"90x360";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_90_X_60"))
   {
      return (const xmlChar *)"90x60";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_90_X_72"))
   {
      return (const xmlChar *)"90x72";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_90_X_90"))
   {
      return (const xmlChar *)"90x90";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_100_X_100"))
   {
      return (const xmlChar *)"100x100";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_120_X_60"))
   {
      return (const xmlChar *)"120x60";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_120_X_72"))
   {
      return (const xmlChar *)"120x72";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_120_X_120"))
   {
      return (const xmlChar *)"120x120";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_120_X_144"))
   {
      return (const xmlChar *)"120x144";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_120_X_180"))
   {
      return (const xmlChar *)"120x180";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_120_X_360"))
   {
      return (const xmlChar *)"120x360";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_144_X_72"))
   {
      return (const xmlChar *)"144x72";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_150_X_150"))
   {
      return (const xmlChar *)"150x150";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_180_X_180"))
   {
      return (const xmlChar *)"180x180";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_180_X_360"))
   {
      return (const xmlChar *)"180x360";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_200_X_200"))
   {
      return (const xmlChar *)"200x200";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_240_X_60"))
   {
      return (const xmlChar *)"240x60";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_240_X_72"))
   {
      return (const xmlChar *)"240x72";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_240_X_144"))
   {
      return (const xmlChar *)"240x144";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_240_X_240"))
   {
      return (const xmlChar *)"240x240";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_300_X_300"))
   {
      return (const xmlChar *)"300x300";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_360_X_180"))
   {
      return (const xmlChar *)"360x180";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_360_X_360"))
   {
      return (const xmlChar *)"360x360";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_360_X_720"))
   {
      return (const xmlChar *)"360x720";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_400_X_400"))
   {
      return (const xmlChar *)"400x400";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_600_X_300"))
   {
      return (const xmlChar *)"600x300";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_600_X_600"))
   {
      return (const xmlChar *)"600x600";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_720_X_360"))
   {
      return (const xmlChar *)"720x360";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_720_X_720"))
   {
      return (const xmlChar *)"720x720";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_1200_X_600"))
   {
      return (const xmlChar *)"1200x600";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_1200_X_1200"))
   {
      return (const xmlChar *)"1200x1200";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_1440_X_720"))
   {
      return (const xmlChar *)"1440x720";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_2880_X_720"))
   {
      return (const xmlChar *)"2880x720";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_DRAFT"))
   {
      return (const xmlChar *)0;
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_FINE"))
   {
      return (const xmlChar *)0;
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_GROUP_3"))
   {
      return (const xmlChar *)0;
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_GROUP_4"))
   {
      return (const xmlChar *)0;
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_HIGH"))
   {
      return (const xmlChar *)0;
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_LOW"))
   {
      return (const xmlChar *)0;
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_MEDIUM"))
   {
      return (const xmlChar *)0;
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_NORMAL"))
   {
      return (const xmlChar *)0;
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_PHOTO_QUALITY"))
   {
      return (const xmlChar *)0;
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"RESOLUTION_PRESENTATION"))
   {
      return (const xmlChar *)0;
   }

   return 0;
}

const xmlChar *
mapOrientationName (const xmlChar *pszFrom)
{
   if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"ORIENTATION_PORTRAIT"))
   {
      return (const xmlChar *)"Portrait";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"ORIENTATION_LANDSCAPE"))
   {
      return (const xmlChar *)"Landscape";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"ORIENTATION_REVERSE_PORTRAIT"))
   {
      return (const xmlChar *)"ReversePortrait";
   }
   else if (0 == xmlStrcmp (pszFrom, (const xmlChar *)"ORIENTATION_REVERSE_LANDSCAPE"))
   {
      return (const xmlChar *)"ReverseLandscape";
   }

   return 0;
}

bool
processForms (xmlNodePtr nodeForms,
              bool       fUpdate)
{
   xmlNodePtr nodeForm    = 0;
   xmlNodePtr nodeElement = 0;
   bool       fRet        = true;

   while (nodeForms)
   {
      nodeForm = firstNode (nodeForms->children);

      while (nodeForm)
      {
         nodeElement = firstNode (nodeForm->children);

         while (nodeElement)
         {
            if (vfDebug) std::cerr << nodeElement->name << std::endl;

            if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"name"))
            {
               xmlNodePtr     nodeNew        = 0;
               const xmlChar *pszNameText    = 0;
               const xmlChar *pszNewNameText = 0;

               // Get current name contents
               pszNameText = xmlNodeListGetString (nodeElement->doc,
                                                   nodeElement->children,
                                                   1);

               pszNewNameText = mapFormName (pszNameText);
               if (!pszNewNameText)
               {
                  std::cerr << "Error:  Cannot map \"" << pszNameText << "\" to anything!" << std::endl;
               }

               if (  fUpdate
                  && pszNewNameText
                  )
               {
                  if (vfDumpElements) std::cout << pszNameText << "->" << pszNewNameText << std::endl;

                  // Create new XML tag that has the old name's contents
                  nodeNew = xmlNewTextChild (nodeElement,
                                             0,
                                             (const xmlChar *)"omniName",
                                             pszNameText);

                  // Add it right after this one and not at the end.
                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Add whitespace to match current indentation
                  nodeNew = xmlNewText ((const xmlChar *)"\n      ");

                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Change the name to the newly mapped name
                  xmlNodeSetContent (nodeElement, pszNewNameText);
               }
               else
               {
                  fRet = false;
               }

               // Clean up
               if (pszNameText)
               {
                  xmlFree ((void *)pszNameText);
               }
            }

            nodeElement = nextNode (nodeElement);
         }

         nodeForm = nextNode (nodeForm);
      }

      nodeForms = nextNode (nodeForms);
   }

   return fRet;
}

bool
processTrays (xmlNodePtr nodeTrays,
              bool       fUpdate)
{
   xmlNodePtr nodeTray    = 0;
   xmlNodePtr nodeElement = 0;
   bool       fRet        = true;

   while (nodeTrays)
   {
      nodeTray = firstNode (nodeTrays->children);

      while (nodeTray)
      {
         nodeElement = firstNode (nodeTray->children);

         while (nodeElement)
         {
            if (vfDebug) std::cerr << nodeElement->name << std::endl;

            if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"name"))
            {
               xmlNodePtr     nodeNew        = 0;
               const xmlChar *pszNameText    = 0;
               const xmlChar *pszNewNameText = 0;

               // Get current name contents
               pszNameText = xmlNodeListGetString (nodeElement->doc,
                                                   nodeElement->children,
                                                   1);

               pszNewNameText = mapTrayName (pszNameText);
               if (!pszNewNameText)
               {
                  std::cerr << "Error:  Cannot map \"" << pszNameText << "\" to anything!" << std::endl;
               }

               if (  fUpdate
                  && pszNewNameText
                  )
               {
                  if (vfDumpElements) std::cout << pszNameText << "->" << pszNewNameText << std::endl;

                  // Create new XML tag that has the old name's contents
                  nodeNew = xmlNewTextChild (nodeElement,
                                             0,
                                             (const xmlChar *)"omniName",
                                             pszNameText);

                  // Add it right after this one and not at the end.
                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Add whitespace to match current indentation
                  nodeNew = xmlNewText ((const xmlChar *)"\n      ");

                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Change the name to the newly mapped name
                  xmlNodeSetContent (nodeElement, pszNewNameText);
               }
               else
               {
                  fRet = false;
               }

               // Clean up
               if (pszNameText)
               {
                  xmlFree ((void *)pszNameText);
               }
            }

            nodeElement = nextNode (nodeElement);
         }

         nodeTray = nextNode (nodeTray);
      }

      nodeTrays = nextNode (nodeTrays);
   }

   return fRet;
}

bool
processResolutions (xmlNodePtr nodeResolutions,
                    bool       fUpdate)
{
   xmlNodePtr nodeResolution = 0;
   xmlNodePtr nodeElement    = 0;
   bool       fRet           = true;

   while (nodeResolutions)
   {
      nodeResolution = firstNode (nodeResolutions->children);

      while (nodeResolution)
      {
         nodeElement = firstNode (nodeResolution->children);

         while (nodeElement)
         {
            if (vfDebug) std::cerr << nodeElement->name << std::endl;

            if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"name"))
            {
               char           achResolution[64];
               xmlNodePtr     nodeNew        = 0;
               const xmlChar *pszNameText    = 0;
               const xmlChar *pszNewNameText = 0;

               // Get current name contents
               pszNameText = xmlNodeListGetString (nodeElement->doc,
                                                   nodeElement->children,
                                                   1);

               pszNewNameText = mapResolutionName (pszNameText);

               if (  !pszNewNameText
                  || !*pszNewNameText
                  )
               {
                  xmlNodePtr nodeXRes = 0;
                  xmlNodePtr nodeYRes = 0;

                  achResolution[0] = '\0';

                  nodeXRes = nextNode (nodeElement);
                  if (nodeXRes)
                  {
                     nodeYRes = nextNode (nodeXRes);
                  }

                  if (  nodeXRes
                     && 0 == xmlStrcmp (nodeXRes->name, (const xmlChar *)"xRes")
                     && nodeYRes
                     && 0 == xmlStrcmp (nodeYRes->name, (const xmlChar *)"yRes")
                     )
                  {
                     const xmlChar *pszXRes = 0;
                     const xmlChar *pszYRes = 0;

                     pszXRes = xmlNodeListGetString (nodeXRes->doc,
                                                     nodeXRes->children,
                                                     1);

                     pszYRes = xmlNodeListGetString (nodeYRes->doc,
                                                     nodeYRes->children,
                                                     1);

/////////////////////if (vfDebug) std::cerr << "pszXRes       = " << (pszXRes ? (char *)pszXRes : "") << std::dec << std::endl;
/////////////////////if (vfDebug) std::cerr << "pszYRes       = " << (pszYRes ? (char *)pszYRes : "") << std::dec << std::endl;

                     if (  pszXRes
                        && pszYRes
                        )
                     {
                        sprintf (achResolution, "%sx%s", pszXRes, pszYRes);
                     }

                     if (pszXRes)
                     {
                        xmlFree ((void *)pszXRes);
                     }
                     if (pszYRes)
                     {
                        xmlFree ((void *)pszYRes);
                     }
                  }

                  if (!achResolution[0])
                  {
                     std::cerr << "Error:  Cannot map \"" << pszNameText << "\" to anything!" << std::endl;
                  }
                  else
                  {
                     pszNewNameText = (const xmlChar *)achResolution;
                  }

//////////////////std::cerr << "nodeXRes      = " << std::hex << (int)nodeXRes << std::dec << std::endl;
//////////////////std::cerr << "nodeYRes      = " << std::hex << (int)nodeYRes << std::dec << std::endl;
//////////////////std::cerr << "achResolution = " << achResolution << std::dec << std::endl;
               }

               if (  fUpdate
                  && pszNewNameText
                  )
               {
                  if (vfDumpElements) std::cout << pszNameText << "->" << pszNewNameText << std::endl;

                  // Create new XML tag that has the old name's contents
                  nodeNew = xmlNewTextChild (nodeElement,
                                             0,
                                             (const xmlChar *)"omniName",
                                             pszNameText);

                  // Add it right after this one and not at the end.
                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Add whitespace to match current indentation
                  nodeNew = xmlNewText ((const xmlChar *)"\n      ");

                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Change the name to the newly mapped name
                  xmlNodeSetContent (nodeElement, pszNewNameText);
               }
               else
               {
                  fRet = false;
               }

               // Clean up
               if (pszNameText)
               {
                  xmlFree ((void *)pszNameText);
               }
            }

            nodeElement = nextNode (nodeElement);
         }

         nodeResolution = nextNode (nodeResolution);
      }

      nodeResolutions = nextNode (nodeResolutions);
   }

   return fRet;
}

bool
processOrientations (xmlNodePtr nodeOrientations,
                     bool       fUpdate)
{
   xmlNodePtr nodeOrientation = 0;
   xmlNodePtr nodeElement     = 0;
   bool       fRet            = true;

   while (nodeOrientations)
   {
      nodeOrientation = firstNode (nodeOrientations->children);

      while (nodeOrientation)
      {
         nodeElement = firstNode (nodeOrientation->children);

         while (nodeElement)
         {
            if (vfDebug) std::cerr << nodeElement->name << std::endl;

            if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"name"))
            {
               xmlNodePtr     nodeNew        = 0;
               const xmlChar *pszNameText    = 0;
               const xmlChar *pszNewNameText = 0;

               // Get current name contents
               pszNameText = xmlNodeListGetString (nodeElement->doc,
                                                   nodeElement->children,
                                                   1);

               pszNewNameText = mapOrientationName (pszNameText);
               if (!pszNewNameText)
               {
                  std::cerr << "Error:  Cannot map \"" << pszNameText << "\" to anything!" << std::endl;
               }

               if (  fUpdate
                  && pszNewNameText
                  )
               {
                  if (vfDumpElements) std::cout << pszNameText << "->" << pszNewNameText << std::endl;

                  // Create new XML tag that has the old name's contents
                  nodeNew = xmlNewTextChild (nodeElement,
                                             0,
                                             (const xmlChar *)"omniName",
                                             pszNameText);

                  // Add it right after this one and not at the end.
                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Add whitespace to match current indentation
                  nodeNew = xmlNewText ((const xmlChar *)"\n      ");

                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Change the name to the newly mapped name
                  xmlNodeSetContent (nodeElement, pszNewNameText);
               }
               else
               {
                  fRet = false;
               }

               // Clean up
               if (pszNameText)
               {
                  xmlFree ((void *)pszNameText);
               }
            }

            nodeElement = nextNode (nodeElement);
         }

         nodeOrientation = nextNode (nodeOrientation);
      }

      nodeOrientations = nextNode (nodeOrientations);
   }

   return fRet;
}

const xmlChar *
convertResolutionFromResolution (xmlDocPtr      docResolutions,
                                 const xmlChar *pszOldResolutionName)
{
   xmlNodePtr     nodeRoot             = 0;
   const xmlChar *pszNewResolutionName = 0;

   if (!docResolutions)
   {
      return 0;
   }

   nodeRoot = firstNode (xmlDocGetRootElement (docResolutions));

   xmlNodePtr nodeResolutions = nodeRoot;
   xmlNodePtr nodeResolution  = 0;
   xmlNodePtr nodeElement     = 0;

   while (nodeResolutions)
   {
      nodeResolution = firstNode (nodeResolutions->children);

      while (nodeResolution)
      {
         nodeElement = firstNode (nodeResolution->children);

         while (nodeElement)
         {
            if (vfDebug) std::cerr << nodeElement->name << std::endl;

            if (  0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"name")
               || 0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"omniName")
               )
            {
               char           achResolution[64];
               const xmlChar *pszNameText       = 0;

               // Get current name contents
               pszNameText = xmlNodeListGetString (nodeElement->doc,
                                                   nodeElement->children,
                                                   1);

               if (0 == xmlStrcmp (pszNameText, pszOldResolutionName))
               {
                  xmlNodePtr nodeXRes = 0;
                  xmlNodePtr nodeYRes = 0;

                  achResolution[0] = '\0';

                  nodeXRes = nextNode (nodeElement);
                  if (nodeXRes)
                  {
                     nodeYRes = nextNode (nodeXRes);
                  }

                  if (  nodeXRes
                     && 0 == xmlStrcmp (nodeXRes->name, (const xmlChar *)"xRes")
                     && nodeYRes
                     && 0 == xmlStrcmp (nodeYRes->name, (const xmlChar *)"yRes")
                     )
                  {
                     const xmlChar *pszXRes = 0;
                     const xmlChar *pszYRes = 0;

                     pszXRes = xmlNodeListGetString (nodeXRes->doc,
                                                     nodeXRes->children,
                                                     1);

                     pszYRes = xmlNodeListGetString (nodeYRes->doc,
                                                     nodeYRes->children,
                                                     1);

                     if (vfDebug) std::cerr << "pszXRes       = " << (pszXRes ? (char *)pszXRes : "") << std::dec << std::endl;
                     if (vfDebug) std::cerr << "pszYRes       = " << (pszYRes ? (char *)pszYRes : "") << std::dec << std::endl;

                     if (  pszXRes
                        && pszYRes
                        )
                     {
                        char *pszRet = 0;

                        sprintf (achResolution, "%sx%s", pszXRes, pszYRes);

                        if (vfDumpElements) std::cout << pszNameText << "->" << achResolution << std::endl;

                        pszRet = (char *)malloc (strlen (achResolution) + 1);
                        if (pszRet)
                        {
                           strcpy (pszRet, achResolution);

                           pszNewResolutionName = (const xmlChar *)pszRet;
                        }
                     }

                     if (pszXRes)
                     {
                        xmlFree ((void *)pszXRes);
                     }
                     if (pszYRes)
                     {
                        xmlFree ((void *)pszYRes);
                     }
                  }

//////////////////std::cerr << "nodeXRes      = " << std::hex << (int)nodeXRes << std::dec << std::endl;
//////////////////std::cerr << "nodeYRes      = " << std::hex << (int)nodeYRes << std::dec << std::endl;
//////////////////std::cerr << "achResolution = " << achResolution << std::dec << std::endl;
               }

               // Clean up
               if (pszNameText)
               {
                  xmlFree ((void *)pszNameText);
               }
            }

            nodeElement = nextNode (nodeElement);
         }

         nodeResolution = nextNode (nodeResolution);
      }

      nodeResolutions = nextNode (nodeResolutions);
   }

   return pszNewResolutionName;
}

const xmlChar *
convertResolutionFromDevice (xmlNodePtr     nodeDevice,
                             xmlDocPtr     *pdocResolutions,
                             const xmlChar *pszOldResolutionName)
{
   if (  pdocResolutions
      && *pdocResolutions
      )
   {
      return convertResolutionFromResolution (*pdocResolutions,
                                              pszOldResolutionName);
   }

   xmlNodePtr     nodeAttributes       = 0;
   const xmlChar *pszNewResolutionName = 0;

   nodeAttributes = firstNode (nodeDevice->children);

   while (nodeAttributes)
   {
      if (  0 == xmlStrcmp (nodeAttributes->name, (const xmlChar *)"Has")
         || 0 == xmlStrcmp (nodeAttributes->name, (const xmlChar *)"Uses")
         )
      {
         xmlDocPtr      doc              = 0;
         xmlNodePtr     nodeRoot         = 0;
         const xmlChar *pszResolutionXML = 0;

         pszResolutionXML = xmlNodeListGetString (nodeAttributes->doc,
                                                  nodeAttributes->children,
                                                  1);

         if (vfDebug) std::cerr << "pszResolutionXML = " << (pszResolutionXML ? (char *)pszResolutionXML : "") << std::endl;

         if (!pszResolutionXML)
         {
            break;
         }

         doc = xmlParseFile ((char *)pszResolutionXML);

         if (!doc)
         {
            break;
         }

         nodeRoot = firstNode (xmlDocGetRootElement (doc));
         if (  nodeRoot
            && 0 == xmlStrcmp ((const xmlChar *)"deviceResolutions", nodeRoot->name)
            )
         {
            if (pdocResolutions)
            {
               *pdocResolutions = doc;
            }

            pszNewResolutionName = convertResolutionFromResolution (doc,
                                                                    pszOldResolutionName);
         }
         else
         {
            if (doc)
            {
               xmlFreeDoc (doc);
            }
         }

         if (pszResolutionXML)
         {
            xmlFree ((void *)pszResolutionXML);
         }
      }

      nodeAttributes = nextNode (nodeAttributes);

      if (vfDebug) std::cerr << "nodeAttributes = " << std::hex<< (int)nodeAttributes << std::dec << std::endl;
   }

   return pszNewResolutionName;
}

const xmlChar *
convertResolutionFromGamma (const char     *pszGammaFileName,
                            xmlDocPtr      *pdocResolutions,
                            const xmlChar  *pszOldResolutionName)
{
   const xmlChar *pszNewResolution = 0;
   xmlDocPtr      doc              = 0;
   xmlNodePtr     nodeRoot         = 0;

   if (  pdocResolutions
      && *pdocResolutions
      )
   {
      return convertResolutionFromResolution (*pdocResolutions,
                                              pszOldResolutionName);
   }
   else
   {
      std::ostringstream oss;
      glob_t             globbuf;
      int                rc;

      // Call glob
      memset (&globbuf, 0, sizeof (globbuf));

      oss << "*.xml"
          << std::ends;

      rc = glob (oss.str ().c_str (), 0, NULL, &globbuf);

      if (  rc
         || 0 == globbuf.gl_pathc
         )
      {
         return 0;
      }

      // Call succeded
      for (int i = 0; i < (int)globbuf.gl_pathc && !pszNewResolution; i++)
      {
         doc = xmlParseFile (globbuf.gl_pathv[i]);
         if (!doc)
         {
            continue;
         }

         nodeRoot = firstNode (xmlDocGetRootElement (doc));
         if (!nodeRoot)
         {
            continue;
         }

         if (0 == xmlStrcmp ((const xmlChar *)"Device", nodeRoot->name))
         {
            xmlNodePtr nodeAttributes = 0;

            nodeAttributes = firstNode (nodeRoot->children);

            while (  nodeAttributes
                  && !pszNewResolution
                  )
            {
               if (  0 == xmlStrcmp (nodeAttributes->name, (const xmlChar *)"Has")
                  || 0 == xmlStrcmp (nodeAttributes->name, (const xmlChar *)"Uses")
                  )
               {
                  const xmlChar *pszFileXML = 0;

                  pszFileXML = xmlNodeListGetString (nodeAttributes->doc,
                                                     nodeAttributes->children,
                                                     1);

                  if (pszFileXML)
                  {
                     if (vfDebug) std::cerr << "! " << pszFileXML << " = " << pszGammaFileName << std::endl;

                     if (0 == xmlStrcmp ((const xmlChar *)pszGammaFileName, pszFileXML))
                     {
                        pszNewResolution = convertResolutionFromDevice (nodeRoot,
                                                                        pdocResolutions,
                                                                        pszOldResolutionName);
                     }
                     else
                     {
                        xmlFree ((void *)pszFileXML);
                     }
                  }
               }

               nodeAttributes = nextNode (nodeAttributes);
            }
         }

         xmlFreeDoc (doc);
      }

      globfree (&globbuf);
   }

   return pszNewResolution;
}

bool
processGammas (xmlNodePtr  nodeGammas,
               bool        fUpdate,
               const char *pszGammaFileName)
{
   xmlNodePtr     nodeGamma      = 0;
   xmlNodePtr     nodeElement    = 0;
   xmlDocPtr      docResolutions = 0;
   bool           fRet           = true;

   while (nodeGammas)
   {
      nodeGamma = firstNode (nodeGammas->children);

      while (nodeGamma)
      {
         nodeElement = firstNode (nodeGamma->children);

         while (nodeElement)
         {
            if (vfDebug) std::cerr << nodeElement->name << std::endl;

            if (  0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"gammaTableResolution")
//             || 0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"gammaTableMedia")
               )
            {
               xmlNodePtr     nodeNew          = 0;
               const xmlChar *pszNameText      = 0;
               const xmlChar *pszNewNameText   = 0;
               const char    *pszOldName       = 0;
               bool           fFreeNewNameText = false;

               // Get current name contents
               pszNameText = xmlNodeListGetString (nodeElement->doc,
                                                   nodeElement->children,
                                                   1);

               if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"gammaTableResolution"))
               {
                  pszNewNameText = mapResolutionName (pszNameText);
                  pszOldName     = "omniResolution";

                  if (  !pszNewNameText
                     || !*pszNewNameText
                     )
                  {
                     pszNewNameText   = convertResolutionFromGamma (pszGammaFileName,
                                                                    &docResolutions,
                                                                    pszNameText);
                     fFreeNewNameText = true;
                  }
               }
//             else if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"gammaTableMedia"))
//             {
//                pszNewNameText = mapMediaName (pszNameText);
//                pszOldName     = "omniMedia";
//             }

               if (  fUpdate
                  && pszNewNameText
                  )
               {
                  if (vfDumpElements) std::cout << pszNameText << "->" << pszNewNameText << std::endl;

                  // Create new XML tag that has the old name's contents
                  nodeNew = xmlNewTextChild (nodeElement,
                                             0,
                                             (const xmlChar *)pszOldName,
                                             pszNameText);

                  // Add it right after this one and not at the end.
                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Add whitespace to match current indentation
                  nodeNew = xmlNewText ((const xmlChar *)"\n      ");

                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Change the name to the newly mapped name
                  xmlNodeSetContent (nodeElement, pszNewNameText);
               }
               else
               {
                  fRet = false;
               }

               // Clean up
               if (pszNameText)
               {
                  xmlFree ((void *)pszNameText);
               }
               if (  fFreeNewNameText
                  && pszNewNameText
                  )
               {
                  xmlFree ((void *)pszNewNameText);
               }
            }

            nodeElement = nextNode (nodeElement);
         }

         nodeGamma = nextNode (nodeGamma);
      }

      nodeGammas = nextNode (nodeGammas);
   }

   if (docResolutions)
   {
      xmlFreeDoc (docResolutions);
   }

   return fRet;
}

bool
processConnections (xmlNodePtr  nodeConnections,
                    bool        fUpdate,
                    const char *pszFileName)
{
   xmlNodePtr nodeConnection = 0;
   xmlNodePtr nodeElement    = 0;
   bool       fRet           = true;

   while (nodeConnections)
   {
      nodeConnection = firstNode (nodeConnections->children);

      while (nodeConnection)
      {
         nodeElement = firstNode (nodeConnection->children);

         while (nodeElement)
         {
            if (  0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"connectionForm")
               || 0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"connectionTray")
//             || 0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"connectionMedia")
               )
            {
               xmlNodePtr     nodeNew          = 0;
               const xmlChar *pszNameText      = 0;
               const xmlChar *pszNewNameText   = 0;
               const char    *pszOldName       = 0;

               // Get current name contents
               pszNameText = xmlNodeListGetString (nodeElement->doc,
                                                   nodeElement->children,
                                                   1);

               if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"connectionForm"))
               {
                  pszNewNameText = mapFormName (pszNameText);
                  pszOldName     = "omniForm";
               }
               else if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"connectionTray"))
               {
                  pszNewNameText = mapTrayName (pszNameText);
                  pszOldName     = "omniTray";

                  if (!pszNewNameText)
                  {
                     std::cerr << "Error:  Cannot map \"" << pszNameText << "\" to anything!" << std::endl;
                  }
               }
///////////////else if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"connectionMedia"))
///////////////{
///////////////   pszNewNameText = mapMediaName (pszNameText);
///////////////   pszOldName     = "omniMedia";
///////////////}

               if (  fUpdate
                  && pszNewNameText
                  )
               {
                  if (vfDumpElements) std::cout << pszNameText << "->" << pszNewNameText << std::endl;

                  // Create new XML tag that has the old name's contents
                  nodeNew = xmlNewTextChild (nodeElement,
                                             0,
                                             (const xmlChar *)pszOldName,
                                             pszNameText);

                  // Add it right after this one and not at the end.
                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Add whitespace to match current indentation
                  nodeNew = xmlNewText ((const xmlChar *)"\n      ");

                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Change the name to the newly mapped name
                  xmlNodeSetContent (nodeElement, pszNewNameText);
               }
               else
               {
                  fRet = false;
               }

               // Clean up
               if (pszNameText)
               {
                  xmlFree ((void *)pszNameText);
               }
            }

            nodeElement = nextNode (nodeElement);
         }

         nodeConnection = nextNode (nodeConnection);
      }

      nodeConnections = nextNode (nodeConnections);
   }

   return fRet;
}

bool
processDevice (xmlNodePtr nodeDevice,
               bool       fUpdate)
{
   xmlNodePtr nodeAttributes = 0;
   xmlNodePtr nodeElement    = 0;
   bool       fRet           = true;

   nodeAttributes = firstNode (nodeDevice->children);

   while (nodeAttributes)
   {
      if (0 == xmlStrcmp (nodeAttributes->name, (const xmlChar *)"DefaultJobProperties"))
      {
         nodeElement = firstNode (nodeAttributes->children);

         while (nodeElement)
         {
            if (  0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"orientation")
               || 0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"form")
               || 0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"tray")
//             || 0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"media")
               || 0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"resolution")
               )
            {
               xmlNodePtr          nodeNew          = 0;
               const xmlChar      *pszNameText      = 0;
               const xmlChar      *pszNewNameText   = 0;
               const char         *pszOldName       = 0;
               bool                fFreeNewNameText = false;

               // Get current name contents
               pszNameText = xmlNodeListGetString (nodeElement->doc,
                                                   nodeElement->children,
                                                   1);

               if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"orientation"))
               {
                  pszNewNameText = mapOrientationName (pszNameText);
                  pszOldName     = "omniOrientation";
               }
               if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"form"))
               {
                  pszNewNameText = mapFormName (pszNameText);
                  pszOldName     = "omniForm";
               }
               if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"tray"))
               {
                  pszNewNameText = mapTrayName (pszNameText);
                  pszOldName     = "omniTray";
               }
//             if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"media"))
//             {
//                pszNewNameText = mapMediaName (pszNameText);
//                pszOldName     = "omniMedia";
//             }
               if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"resolution"))
               {
                  pszNewNameText = mapResolutionName (pszNameText);
                  pszOldName     = "omniResolution";

                  if (  !pszNewNameText
                     || !*pszNewNameText
                     )
                  {
                     pszNewNameText   = convertResolutionFromDevice (nodeDevice, 0, pszNameText);
                     fFreeNewNameText = true;
                  }
               }

///////////////if (vfDebug) std::cerr << nodeElement->name << std::endl;
///////////////if (vfDebug) std::cerr << "pszNameText    = " << (pszNameText ? (const char *)pszNameText : "NULL") << std::endl;
///////////////if (vfDebug) std::cerr << "pszNewNameText = " << (pszNewNameText ? (const char *)pszNewNameText : "NULL") << std::endl;
///////////////if (vfDebug) std::cerr << "pszOldName     = " << (pszOldName ? pszOldName : "NULL") << std::endl;

               if (  fUpdate
                  && pszNewNameText
                  )
               {
                  if (vfDumpElements) std::cout << pszNameText << "->" << pszNewNameText << std::endl;

                  // Create new XML tag that has the old name's contents
                  nodeNew = xmlNewTextChild (nodeElement,
                                             0,
                                             (const xmlChar *)pszOldName,
                                             pszNameText);

                  // Add it right after this one and not at the end.
                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Add whitespace to match current indentation
                  nodeNew = xmlNewText ((const xmlChar *)"\n      ");

                  xmlAddNextSibling (nodeElement, nodeNew);

                  // Change the name to the newly mapped name
                  xmlNodeSetContent (nodeElement, pszNewNameText);
               }
               else
               {
                  fRet = false;
               }

               // Clean up
               if (pszNameText)
               {
                  xmlFree ((void *)pszNameText);
               }
               if (  fFreeNewNameText
                  && pszNewNameText
                  )
               {
                  free ((void *)pszNewNameText);
               }
            }

            nodeElement = nextNode (nodeElement);
         }
      }

      nodeAttributes = nextNode (nodeAttributes);
   }

///xmlDocDump (stdout, nodeDevice->doc);

   // Query the old job properties
   const char *pszOrientation     = 0;
   const char *pszForm            = 0;
   const char *pszTray            = 0;
   const char *pszMedia           = 0;
   const char *pszResolution      = 0;
   const char *pszDither          = 0;
   const char *pszPrintmode       = 0;
   const char *pszOmniOrientation = 0;
   const char *pszOmniForm        = 0;
   const char *pszOmniTray        = 0;
   const char *pszOmniMedia       = 0;
   const char *pszOmniResolution  = 0;
   const char *pszOmniDither      = 0;
   const char *pszOmniPrintmode   = 0;
   xmlNodePtr  nodeNewValue       = 0;
   xmlNodePtr  nodeNewWS          = 0;

   nodeAttributes = firstNode (nodeDevice->children);

   while (nodeAttributes)
   {
      if (0 == xmlStrcmp (nodeAttributes->name, (const xmlChar *)"DefaultJobProperties"))
      {
         // Get the first non-text child in the list
         nodeElement = firstNode (nodeAttributes->children);

         if (nodeElement == NULL)
         {
            return false;
         }

         pszOrientation = bundleStringData (nodeElement);
         // Move to the next node
         nodeElement = nextNode (nodeElement);
         if (nodeElement == NULL)
         {
            return false;
         }
         if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"omniOrientation"))
         {
            pszOmniOrientation = bundleStringData (nodeElement);
            // Move to the next node
            nodeElement = nextNode (nodeElement);
            if (nodeElement == NULL)
            {
               return false;
            }
         }

         pszForm = bundleStringData (nodeElement);
         // Move to the next node
         nodeElement = nextNode (nodeElement);
         if (nodeElement == NULL)
         {
            return false;
         }
         if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"omniForm"))
         {
            pszOmniForm = bundleStringData (nodeElement);
            // Move to the next node
            nodeElement = nextNode (nodeElement);
            if (nodeElement == NULL)
            {
               return false;
            }
         }

         pszTray = bundleStringData (nodeElement);
         // Move to the next node
         nodeElement = nextNode (nodeElement);
         if (nodeElement == NULL)
         {
            return false;
         }
         if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"omniTray"))
         {
            pszOmniTray = bundleStringData (nodeElement);
            // Move to the next node
            nodeElement = nextNode (nodeElement);
            if (nodeElement == NULL)
            {
               return false;
            }
         }

         pszMedia = bundleStringData (nodeElement);
         // Move to the next node
         nodeElement = nextNode (nodeElement);
         if (nodeElement == NULL)
         {
            return false;
         }
         if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"omniMedia"))
         {
            pszOmniMedia = bundleStringData (nodeElement);
            // Move to the next node
            nodeElement = nextNode (nodeElement);
            if (nodeElement == NULL)
            {
               return false;
            }
         }

         pszResolution = bundleStringData (nodeElement);
         // Move to the next node
         nodeElement = nextNode (nodeElement);
         if (nodeElement == NULL)
         {
            return false;
         }
         if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"omniResolution"))
         {
            pszOmniResolution = bundleStringData (nodeElement);
            // Move to the next node
            nodeElement = nextNode (nodeElement);
            if (nodeElement == NULL)
            {
               return false;
            }
         }

         pszDither = bundleStringData (nodeElement);
         // Move to the next node
         nodeElement = nextNode (nodeElement);
         if (nodeElement == NULL)
         {
            return false;
         }
         if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"omniDither"))
         {
            pszOmniDither = bundleStringData (nodeElement);
            // Move to the next node
            nodeElement = nextNode (nodeElement);
            if (nodeElement == NULL)
               return false;
         }

         pszPrintmode = bundleStringData (nodeElement);
         // Move to the next node
         nodeElement = nextNode (nodeElement);
         if (nodeElement != NULL)
         {
            if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"omniPrintmode"))
            {
               pszOmniPrintmode = bundleStringData (nodeElement);
               // Move to the next node
               nodeElement = nextNode (nodeElement);
            }
         }

         xmlNodePtr nodePrevAttributes = nodeAttributes->prev;

         // Delete the <DefaultJobProperties>...</DefaultJobProperties> tree
         xmlUnlinkNode (nodeAttributes);
         xmlFreeNodeList (nodeAttributes);

         nodeAttributes = xmlNewNode (0, (const xmlChar *)"DefaultJobProperties");
         if (nodeAttributes)
         {
            xmlAddNextSibling (nodePrevAttributes, nodeAttributes);
         }
      }

      nodeAttributes = nextNode (nodeAttributes);
   }

   nodeAttributes = firstNode (nodeDevice->children);

   while (nodeAttributes)
   {
      if (0 == xmlStrcmp (nodeAttributes->name, (const xmlChar *)"DefaultJobProperties"))
      {
         nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
         if (nodeNewWS)
         {
            xmlAddChild (nodeAttributes, nodeNewWS);
         }

         nodeElement = nodeAttributes->children;
/////////if (vfDebug) std::cerr << "nodeElement = " << std::hex << (int)nodeElement << std::dec << std::endl;

         xmlNodePtr nodeCopies = xmlNewNode (0, (const xmlChar *)"Copies");
         if (nodeCopies)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodeCopies);

            nodeNewValue = xmlNewText ((const xmlChar *)"1");
            if (nodeNewValue)
            {
               xmlAddChild (nodeCopies, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeCopies, nodeNewWS);
            }
         }

         xmlNodePtr nodeDither = xmlNewNode (0, (const xmlChar *)"dither");
         if (nodeDither)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodeDither);

            nodeNewValue = xmlNewText ((const xmlChar *)pszDither);
            if (nodeNewValue)
            {
               xmlAddChild (nodeDither, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeDither, nodeNewWS);
            }
         }

         xmlNodePtr nodeForm = xmlNewNode (0, (const xmlChar *)"Form");
         if (nodeForm)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodeForm);

            nodeNewValue = xmlNewText ((const xmlChar *)pszForm);
            if (nodeNewValue)
            {
               xmlAddChild (nodeForm, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeForm, nodeNewWS);
            }

            if (pszOmniForm)
            {
               xmlNodePtr nodeOmniForm = xmlNewNode (0, (const xmlChar *)"omniForm");
               if (nodeOmniForm)
               {
                  nodeElement = xmlAddNextSibling (nodeElement, nodeOmniForm);

                  nodeNewValue = xmlNewText ((const xmlChar *)pszOmniForm);
                  if (nodeNewValue)
                  {
                     xmlAddChild (nodeOmniForm, nodeNewValue);
                  }

                  nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
                  if (nodeNewWS)
                  {
                     nodeElement = xmlAddNextSibling (nodeOmniForm, nodeNewWS);
                  }
               }
            }
         }

         xmlNodePtr nodeMedia = xmlNewNode (0, (const xmlChar *)"media");
         if (nodeMedia)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodeMedia);

            nodeNewValue = xmlNewText ((const xmlChar *)pszMedia);
            if (nodeNewValue)
            {
               xmlAddChild (nodeMedia, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeMedia, nodeNewWS);
            }

            if (pszOmniMedia)
            {
               xmlNodePtr nodeOmniMedia = xmlNewNode (0, (const xmlChar *)"omniMedia");
               if (nodeOmniMedia)
               {
                  nodeElement = xmlAddNextSibling (nodeElement, nodeOmniMedia);

                  nodeNewValue = xmlNewText ((const xmlChar *)pszOmniMedia);
                  if (nodeNewValue)
                  {
                     xmlAddChild (nodeOmniMedia, nodeNewValue);
                  }

                  nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
                  if (nodeNewWS)
                  {
                     nodeElement = xmlAddNextSibling (nodeOmniMedia, nodeNewWS);
                  }
               }
            }
         }

         xmlNodePtr nodeDNUp = xmlNewNode (0, (const xmlChar *)"deviceNumberUp");
         if (nodeDNUp)
         {
            xmlNodePtr nodeNextDNUpChild = 0;

            nodeElement = xmlAddNextSibling (nodeElement, nodeDNUp);

            nodeNewWS = xmlNewText ((const xmlChar *)"\n         ");
            if (nodeNewWS)
            {
               nodeNextDNUpChild = xmlAddChild (nodeDNUp, nodeNewWS);
            }

            xmlNodePtr nodeNUp = xmlNewNode (0, (const xmlChar *)"NumberUp");
            if (nodeNUp)
            {
               xmlNodePtr nodeNextNUpChild = 0;

               nodeNextDNUpChild = xmlAddNextSibling (nodeNextDNUpChild, nodeNUp);

               xmlNewProp (nodeNUp, (const xmlChar *)"FORMAT", (const xmlChar *)"XbyY");

               xmlNodePtr nodeX = xmlNewNode (0, (const xmlChar *)"x");
               if (nodeX)
               {
                  nodeNewWS = xmlNewText ((const xmlChar *)"\n            ");
                  if (nodeNewWS)
                  {
                     nodeNextNUpChild = xmlAddChild (nodeNUp, nodeNewWS);
                  }

                  nodeNextNUpChild = xmlAddNextSibling (nodeNextNUpChild, nodeX);

                  nodeNewValue = xmlNewText ((const xmlChar *)"1");
                  if (nodeNewValue)
                  {
                     xmlAddChild (nodeX, nodeNewValue);
                  }
               }

               xmlNodePtr nodeY = xmlNewNode (0, (const xmlChar *)"y");
               if (nodeY)
               {
                  nodeNewWS = xmlNewText ((const xmlChar *)"\n            ");
                  if (nodeNewWS)
                  {
                     nodeNextNUpChild = xmlAddChild (nodeNUp, nodeNewWS);
                  }

                  nodeNextNUpChild = xmlAddNextSibling (nodeNextNUpChild, nodeY);

                  nodeNewValue = xmlNewText ((const xmlChar *)"1");
                  if (nodeNewValue)
                  {
                     xmlAddChild (nodeY, nodeNewValue);
                  }

                  nodeNewWS = xmlNewText ((const xmlChar *)"\n         ");
                  if (nodeNewWS)
                  {
                     nodeNextNUpChild = xmlAddNextSibling (nodeNextNUpChild, nodeNewWS);
                  }
               }

               nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
               if (nodeNewWS)
               {
                  xmlAddNextSibling (nodeNUp, nodeNewWS);
               }
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n         ");
            if (nodeNewWS)
            {
               nodeNextDNUpChild = xmlAddNextSibling (nodeNextDNUpChild, nodeNewWS);
            }

            xmlNodePtr nodeNUPD = xmlNewNode (0, (const xmlChar *)"NumberUpPresentationDirection");
            if (nodeNUPD)
            {
               nodeNextDNUpChild = xmlAddNextSibling (nodeNextDNUpChild, nodeNUPD);

               nodeNewValue = xmlNewText ((const xmlChar *)"TorightTobottom");
               if (nodeNewValue)
               {
                  xmlAddChild (nodeNUPD, nodeNewValue);
               }
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeDNUp, nodeNewWS);
            }
         }

         xmlNodePtr nodeRotation = xmlNewNode (0, (const xmlChar *)"Rotation");
         if (nodeRotation)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodeRotation);

            nodeNewValue = xmlNewText ((const xmlChar *)pszOrientation);
            if (nodeNewValue)
            {
               xmlAddChild (nodeRotation, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeRotation, nodeNewWS);
            }

            if (pszOmniOrientation)
            {
               xmlNodePtr nodeOmniRotation = xmlNewNode (0, (const xmlChar *)"omniOrientation");
               if (nodeOmniRotation)
               {
                  nodeElement = xmlAddNextSibling (nodeElement, nodeOmniRotation);

                  nodeNewValue = xmlNewText ((const xmlChar *)pszOmniOrientation);
                  if (nodeNewValue)
                  {
                     xmlAddChild (nodeOmniRotation, nodeNewValue);
                  }

                  nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
                  if (nodeNewWS)
                  {
                     nodeElement = xmlAddNextSibling (nodeOmniRotation, nodeNewWS);
                  }
               }
            }
         }

         xmlNodePtr nodeOutputBin = xmlNewNode (0, (const xmlChar *)"OutputBin");
         if (nodeOutputBin)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodeOutputBin);

            nodeNewValue = xmlNewText ((const xmlChar *)"Top");
            if (nodeNewValue)
            {
               xmlAddChild (nodeOutputBin, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeOutputBin, nodeNewWS);
            }
         }

         xmlNodePtr nodePrintmode = xmlNewNode (0, (const xmlChar *)"printmode");
         if (nodePrintmode)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodePrintmode);

            nodeNewValue = xmlNewText ((const xmlChar *)pszPrintmode);
            if (nodeNewValue)
            {
               xmlAddChild (nodePrintmode, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodePrintmode, nodeNewWS);
            }

            if (pszOmniPrintmode)
            {
               xmlNodePtr nodeOmniPrintmode = xmlNewNode (0, (const xmlChar *)"omniPrintmode");
               if (nodeOmniPrintmode)
               {
                  nodeElement = xmlAddNextSibling (nodeElement, nodeOmniPrintmode);

                  nodeNewValue = xmlNewText ((const xmlChar *)pszOmniPrintmode);
                  if (nodeNewValue)
                  {
                     xmlAddChild (nodeOmniPrintmode, nodeNewValue);
                  }

                  nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
                  if (nodeNewWS)
                  {
                     nodeElement = xmlAddNextSibling (nodeOmniPrintmode, nodeNewWS);
                  }
               }
            }
         }

         xmlNodePtr nodeResolution = xmlNewNode (0, (const xmlChar *)"Resolution");
         if (nodeResolution)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodeResolution);

            nodeNewValue = xmlNewText ((const xmlChar *)pszResolution);
            if (nodeNewValue)
            {
               xmlAddChild (nodeResolution, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeResolution, nodeNewWS);
            }

            if (pszOmniResolution)
            {
               xmlNodePtr nodeOmniResolution = xmlNewNode (0, (const xmlChar *)"omniResolution");
               if (nodeOmniResolution)
               {
                  nodeElement = xmlAddNextSibling (nodeElement, nodeOmniResolution);

                  nodeNewValue = xmlNewText ((const xmlChar *)pszOmniResolution);
                  if (nodeNewValue)
                  {
                     xmlAddChild (nodeOmniResolution, nodeNewValue);
                  }

                  nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
                  if (nodeNewWS)
                  {
                     nodeElement = xmlAddNextSibling (nodeOmniResolution, nodeNewWS);
                  }
               }
            }
         }

         xmlNodePtr nodeDeviceScaling = xmlNewNode (0, (const xmlChar *)"deviceScaling");
         if (nodeDeviceScaling)
         {
            xmlNodePtr nodeNextDSChild = 0;

            nodeElement = xmlAddNextSibling (nodeElement, nodeDeviceScaling);

            nodeNewWS = xmlNewText ((const xmlChar *)"\n         ");
            if (nodeNewWS)
            {
               nodeNextDSChild = xmlAddChild (nodeDeviceScaling, nodeNewWS);
            }

            xmlNodePtr nodeScalingPercentage = xmlNewNode (0, (const xmlChar *)"ScalingPercentage");
            if (nodeScalingPercentage)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeScalingPercentage);

               nodeNewValue = xmlNewText ((const xmlChar *)"0");
               if (nodeNewValue)
               {
                  xmlAddChild (nodeScalingPercentage, nodeNewValue);
               }
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n         ");
            if (nodeNewWS)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeNewWS);
            }

            xmlNodePtr nodeScalingType = xmlNewNode (0, (const xmlChar *)"ScalingType");
            if (nodeScalingType)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeScalingType);

               nodeNewValue = xmlNewText ((const xmlChar *)"None");
               if (nodeNewValue)
               {
                  xmlAddChild (nodeScalingType, nodeNewValue);
               }
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeNewWS);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeDeviceScaling, nodeNewWS);
            }
         }

         xmlNodePtr nodeSheetCollate = xmlNewNode (0, (const xmlChar *)"SheetCollate");
         if (nodeSheetCollate)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodeSheetCollate);

            nodeNewValue = xmlNewText ((const xmlChar *)"SheetUncollated");
            if (nodeNewValue)
            {
               xmlAddChild (nodeSheetCollate, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeSheetCollate, nodeNewWS);
            }
         }

         xmlNodePtr nodeSides = xmlNewNode (0, (const xmlChar *)"Sides");
         if (nodeSides)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodeSides);

            nodeNewValue = xmlNewText ((const xmlChar *)"OneSidedFront");
            if (nodeNewValue)
            {
               xmlAddChild (nodeSides, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeSides, nodeNewWS);
            }
         }

         xmlNodePtr nodeDeviceStitching = xmlNewNode (0, (const xmlChar *)"deviceStitching");
         if (nodeDeviceStitching)
         {
            xmlNodePtr nodeNextDSChild = 0;

            nodeElement = xmlAddNextSibling (nodeElement, nodeDeviceStitching);

            nodeNewWS = xmlNewText ((const xmlChar *)"\n         ");
            if (nodeNewWS)
            {
               nodeNextDSChild = xmlAddChild (nodeDeviceStitching, nodeNewWS);
            }

            xmlNodePtr nodeStitchingPosition = xmlNewNode (0, (const xmlChar *)"StitchingPosition");
            if (nodeStitchingPosition)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeStitchingPosition);

               nodeNewValue = xmlNewText ((const xmlChar *)"0");
               if (nodeNewValue)
               {
                  xmlAddChild (nodeStitchingPosition, nodeNewValue);
               }
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n         ");
            if (nodeNewWS)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeNewWS);
            }

            xmlNodePtr nodeStitchingReferenceEdge = xmlNewNode (0, (const xmlChar *)"StitchingReferenceEdge");
            if (nodeStitchingReferenceEdge)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeStitchingReferenceEdge);

               nodeNewValue = xmlNewText ((const xmlChar *)"Top");
               if (nodeNewValue)
               {
                  xmlAddChild (nodeStitchingReferenceEdge, nodeNewValue);
               }
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n         ");
            if (nodeNewWS)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeNewWS);
            }

            xmlNodePtr nodeStitchingType = xmlNewNode (0, (const xmlChar *)"StitchingType");
            if (nodeStitchingType)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeStitchingType);

               nodeNewValue = xmlNewText ((const xmlChar *)"None");
               if (nodeNewValue)
               {
                  xmlAddChild (nodeStitchingType, nodeNewValue);
               }
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n         ");
            if (nodeNewWS)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeNewWS);
            }

            xmlNodePtr nodeStitchingCount = xmlNewNode (0, (const xmlChar *)"StitchingCount");
            if (nodeStitchingCount)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeStitchingCount);

               nodeNewValue = xmlNewText ((const xmlChar *)"0");
               if (nodeNewValue)
               {
                  xmlAddChild (nodeStitchingCount, nodeNewValue);
               }
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n         ");
            if (nodeNewWS)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeNewWS);
            }

            xmlNodePtr nodeStitchingAngle = xmlNewNode (0, (const xmlChar *)"StitchingAngle");
            if (nodeStitchingAngle)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeStitchingAngle);

               nodeNewValue = xmlNewText ((const xmlChar *)"0");
               if (nodeNewValue)
               {
                  xmlAddChild (nodeStitchingAngle, nodeNewValue);
               }
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeNextDSChild = xmlAddNextSibling (nodeNextDSChild, nodeNewWS);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeDeviceStitching, nodeNewWS);
            }
         }

         xmlNodePtr nodeTray = xmlNewNode (0, (const xmlChar *)"InputTray");
         if (nodeTray)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodeTray);

            nodeNewValue = xmlNewText ((const xmlChar *)pszTray);
            if (nodeNewValue)
            {
               xmlAddChild (nodeTray, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeTray, nodeNewWS);
            }

            if (pszOmniTray)
            {
               xmlNodePtr nodeOmniTray = xmlNewNode (0, (const xmlChar *)"omniTray");
               if (nodeOmniTray)
               {
                  nodeElement = xmlAddNextSibling (nodeElement, nodeOmniTray);

                  nodeNewValue = xmlNewText ((const xmlChar *)pszOmniTray);
                  if (nodeNewValue)
                  {
                     xmlAddChild (nodeOmniTray, nodeNewValue);
                  }

                  nodeNewWS = xmlNewText ((const xmlChar *)"\n      ");
                  if (nodeNewWS)
                  {
                     nodeElement = xmlAddNextSibling (nodeOmniTray, nodeNewWS);
                  }
               }
            }
         }

         xmlNodePtr nodeTrimming = xmlNewNode (0, (const xmlChar *)"Trimming");
         if (nodeTrimming)
         {
            nodeElement = xmlAddNextSibling (nodeElement, nodeTrimming);

            nodeNewValue = xmlNewText ((const xmlChar *)"None");
            if (nodeNewValue)
            {
               xmlAddChild (nodeTrimming, nodeNewValue);
            }

            nodeNewWS = xmlNewText ((const xmlChar *)"\n   ");
            if (nodeNewWS)
            {
               nodeElement = xmlAddNextSibling (nodeTrimming, nodeNewWS);
            }
         }
      }

      nodeAttributes = nextNode (nodeAttributes);
   }

   if (pszOrientation)
   {
      xmlFree ((void *)pszOrientation);
   }
   if (pszForm)
   {
      xmlFree ((void *)pszForm);
   }
   if (pszTray)
   {
      xmlFree ((void *)pszTray);
   }
   if (pszMedia)
   {
      xmlFree ((void *)pszMedia);
   }
   if (pszResolution)
   {
      xmlFree ((void *)pszResolution);
   }
   if (pszDither)
   {
      xmlFree ((void *)pszDither);
   }
   if (pszPrintmode)
   {
      xmlFree ((void *)pszPrintmode);
   }
   if (pszOmniOrientation)
   {
      xmlFree ((void *)pszOmniOrientation);
   }
   if (pszOmniForm)
   {
      xmlFree ((void *)pszOmniForm);
   }
   if (pszOmniTray)
   {
      xmlFree ((void *)pszOmniTray);
   }
   if (pszOmniMedia)
   {
      xmlFree ((void *)pszOmniMedia);
   }
   if (pszOmniResolution)
   {
      xmlFree ((void *)pszOmniResolution);
   }
   if (pszOmniDither)
   {
      xmlFree ((void *)pszOmniDither);
   }
   if (pszOmniPrintmode)
   {
      xmlFree ((void *)pszOmniPrintmode);
   }

   return true; //fRet;
}

bool
processFile (const char *pszFileName,
             bool        fUpdate)
{
#ifdef DTD
   xmlValidCtxt        validatorDTD    = {
      0,
      MyValidityErrorFunc,
      MyValidityWarningFunc
   };
#endif
   xmlDocPtr           doc             = 0;
   xmlNodePtr          nodeRoot        = 0;
   bool                fRC             = false;

   doc = xmlParseFile (pszFileName);

   if (vfDebug) std::cerr << "doc = " << std::hex << (int)doc << std::dec << std::endl;

   if (!doc)
   {
      return false;
   }

#ifdef DTD
   int iRC = xmlValidateDocument (&validatorDTD, doc);

   if (vfDebug) std::cerr << "xmlValidateDocument = " << iRC << std::endl;
#endif

   nodeRoot = firstNode (xmlDocGetRootElement (doc));
   if (!nodeRoot)
   {
      return false;
   }
   else if (0 == xmlStrcmp ((const xmlChar *)"deviceForms", nodeRoot->name))
   {
      fRC = processForms (nodeRoot, fUpdate);
   }
   else if (0 == xmlStrcmp ((const xmlChar *)"deviceTrays", nodeRoot->name))
   {
      fRC = processTrays (nodeRoot, fUpdate);
   }
   else if (0 == xmlStrcmp ((const xmlChar *)"deviceResolutions", nodeRoot->name))
   {
      fRC = processResolutions (nodeRoot, fUpdate);
   }
   else if (0 == xmlStrcmp ((const xmlChar *)"deviceOrientations", nodeRoot->name))
   {
      fRC = processOrientations (nodeRoot, fUpdate);
   }
   else if (0 == xmlStrcmp ((const xmlChar *)"deviceGammaTables", nodeRoot->name))
   {
      fRC = processGammas (nodeRoot, fUpdate, pszFileName);
   }
   else if (0 == xmlStrcmp ((const xmlChar *)"deviceConnections", nodeRoot->name))
   {
      fRC = processConnections (nodeRoot, fUpdate, pszFileName);
   }
   else if (0 == xmlStrcmp ((const xmlChar *)"Device", nodeRoot->name))
   {
      fRC = processDevice (nodeRoot, fUpdate);
   }

   if (  fRC
      && vfDumpFile
      )
   {
      xmlDocDump (stdout, doc);
   }

   if (  fRC
      && fUpdate
      )
   {
      xmlSaveFile (pszFileName, doc);
   }

   xmlFreeDoc (doc);

   return true;
}

int
main (int argc, char *argv[])
{
   std::ostringstream oss;
   glob_t             globbuf;
   int                rc;

   // Call glob
   memset (&globbuf, 0, sizeof (globbuf));

   oss << "*.xml"
       << std::ends;

   rc = glob (oss.str ().c_str (), 0, NULL, &globbuf);

   if (rc)
   {
      std::cerr << "Error: glob (" << oss.str () << ") failed. rc = " << rc << std::endl;

      return __LINE__;
   }

   if (0 == globbuf.gl_pathc)
   {
      std::cerr << "Error: glob (" << oss.str () << ") returned 0 files!" << std::endl;

      return __LINE__;
   }

   // Call succeded
   for (int i = 0; i < (int)globbuf.gl_pathc; i++)
   {
      std::cout << globbuf.gl_pathv[i] << std::endl;

      processFile (globbuf.gl_pathv[i], true);
   }

   globfree (&globbuf);

   return 0;
}
