/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <stdarg.h>

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/valid.h>
#include <libxml/xmlschemas.h>

#include <glob.h>

#include "defines.hpp"

const static bool vfDebug        = false;
const static bool vfDumpFile     = false;
const static bool vfDumpElements = true;

void
printXMLNode (xmlNodePtr elm)
{
   xmlElemDump (stdout, elm->doc, elm);
   fprintf (stdout, "\n");
}

xmlNodePtr
firstNode (xmlNodePtr elm)
{
   while (  0 != elm
         && (  XML_COMMENT_NODE == elm->type
            || XML_TEXT_NODE    == elm->type
            )
         )
   {
      // Move to the next node
      elm = elm->next;
   }

   return elm;
}

xmlNodePtr
nextNode (xmlNodePtr elm)
{
   do
   {
      // Move to the next node
      elm = elm->next;

   } while (  0 != elm
           && (  XML_COMMENT_NODE == elm->type
              || XML_TEXT_NODE    == elm->type
              )
           );

   return elm;
}

char *
bundleStringData (xmlNodePtr nodeItem)
{
   return (char *)xmlNodeListGetString (nodeItem->doc, nodeItem->children, 1);
}

void
MyValidityErrorFunc (void *ctx, const char *msg, ...)
{
   va_list list;

   va_start (list, msg);

   printf ("Error: ");
   vprintf (msg, list);

   va_end (list);

   if (ctx)
   {
#if 0
      OmniDomParser *myCtx = (OmniDomParser *)ctx;

      myCtx->setErrorCondition ();
#endif
   }
}

void
MyValidityWarningFunc (void *ctx, const char *msg, ...)
{
   va_list list;

   va_start (list, msg);

   printf ("Warning: ");
   vprintf (msg, list);

   va_end (list);
}

bool
processGammas (xmlNodePtr  nodeGammas,
               bool        fUpdate,
               const char *pszGammaFileName)
{
   xmlNodePtr     nodeGamma      = 0;
   xmlNodePtr     nodeElement    = 0;
   xmlDocPtr      docResolutions = 0;
   bool           fRet           = true;

   while (nodeGammas)
   {
      nodeGamma = firstNode (nodeGammas->children);

      while (nodeGamma)
      {
         nodeElement = firstNode (nodeGamma->children);

         while (nodeElement)
         {
            if (vfDebug) std::cerr << nodeElement->name << std::endl;

            if (  0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"gammaTableResolution")
               || 0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"gammaTableMedia")
               || 0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"gammaTablePrintMode")
               )
            {
               xmlNodePtr     nodeNew          = 0;
               const xmlChar *pszNameText      = 0;
               const xmlChar *pszNewNameText   = 0;

               // Get current name contents
               pszNameText = xmlNodeListGetString (nodeElement->doc,
                                                   nodeElement->children,
                                                   1);

               if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"gammaTableResolution"))
               {
                  pszNewNameText = (const xmlChar *)"Resolution";
               }
               else if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"gammaTableMedia"))
               {
                  pszNewNameText = (const xmlChar *)"media";
               }
               else if (0 == xmlStrcmp (nodeElement->name, (const xmlChar *)"gammaTablePrintMode"))
               {
                  pszNewNameText = (const xmlChar *)"printmode";
               }

               if (  fUpdate
                  && pszNewNameText
                  )
               {
                  if (vfDumpElements)
                     std::cout << "<"
                               << nodeElement->name
                               << ">"
                               << pszNameText
                               << "</"
                               << nodeElement->name
                               << "> -> <"
                               << pszNewNameText
                               << ">"
                               << pszNameText
                               << "</"
                               << pszNewNameText
                               << ">"
                               << std::endl;

                  // Change the name to the newly mapped name
                  xmlNodeSetName (nodeElement, pszNewNameText);

                  fRet = true;
               }

               // Clean up
               if (pszNameText)
               {
                  xmlFree ((void *)pszNameText);
               }
            }

            nodeElement = nextNode (nodeElement);
         }

         nodeGamma = nextNode (nodeGamma);
      }

      nodeGammas = nextNode (nodeGammas);
   }

   if (docResolutions)
   {
      xmlFreeDoc (docResolutions);
   }

   return fRet;
}

bool
processFile (const char *pszFileName,
             bool        fUpdate)
{
#ifdef DTD
   xmlValidCtxt        validatorDTD    = {
      0,
      MyValidityErrorFunc,
      MyValidityWarningFunc
   };
#endif
   xmlDocPtr           doc             = 0;
   xmlNodePtr          nodeRoot        = 0;
   bool                fRC             = false;

   doc = xmlParseFile (pszFileName);

   if (vfDebug) std::cerr << "doc = " << std::hex << (int)doc << std::dec << std::endl;

   if (!doc)
   {
      return false;
   }

#ifdef DTD
   int iRC = xmlValidateDocument (&validatorDTD, doc);

   if (vfDebug) std::cerr << "xmlValidateDocument = " << iRC << std::endl;
#endif

   nodeRoot = firstNode (xmlDocGetRootElement (doc));
   if (!nodeRoot)
   {
      return false;
   }
   else if (0 == xmlStrcmp ((const xmlChar *)"deviceGammaTables", nodeRoot->name))
   {
      fRC = processGammas (nodeRoot, fUpdate, pszFileName);
   }

   if (  fRC
      && vfDumpFile
      )
   {
      xmlDocDump (stdout, doc);
   }

   if (  fRC
      && fUpdate
      )
   {
      xmlSaveFile (pszFileName, doc);
   }

   xmlFreeDoc (doc);

   return true;
}

int
main (int argc, char *argv[])
{
   std::ostringstream oss;
   glob_t             globbuf;
   int                rc;

   // Call glob
   memset (&globbuf, 0, sizeof (globbuf));

   oss << "*.xml"
       << std::ends;

   rc = glob (oss.str ().c_str (), 0, NULL, &globbuf);

   if (rc)
   {
      std::cerr << "Error: glob (" << oss.str () << ") failed. rc = " << rc << std::endl;

      return __LINE__;
   }

   if (0 == globbuf.gl_pathc)
   {
      std::cerr << "Error: glob (" << oss.str () << ") returned 0 files!" << std::endl;

      return __LINE__;
   }

   // Call succeded
   for (int i = 0; i < (int)globbuf.gl_pathc; i++)
   {
      std::cout << globbuf.gl_pathv[i] << std::endl;

      processFile (globbuf.gl_pathv[i], true);
   }

   globfree (&globbuf);

   return 0;
}
