/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <stdarg.h>

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/valid.h>
#include <libxml/xmlschemas.h>

#include <glob.h>

#include "Device.hpp"

const static bool vfDebug        = false;
const static bool vfDumpFile     = false;
const static bool vfDumpElements = true;

void
printXMLNode (xmlNodePtr elm)
{
   xmlElemDump (stdout, elm->doc, elm);
   fprintf (stdout, "\n");
}

xmlNodePtr
firstNode (xmlNodePtr elm)
{
   while (  0 != elm
         && (  XML_COMMENT_NODE == elm->type
            || XML_TEXT_NODE    == elm->type
            )
         )
   {
      // Move to the next node
      elm = elm->next;
   }

   return elm;
}

xmlNodePtr
nextNode (xmlNodePtr elm)
{
   do
   {
      // Move to the next node
      elm = elm->next;

   } while (  0 != elm
           && (  XML_COMMENT_NODE == elm->type
              || XML_TEXT_NODE    == elm->type
              )
           );

   return elm;
}

char *
bundleStringData (xmlNodePtr nodeItem)
{
   return (char *)xmlNodeListGetString (nodeItem->doc, nodeItem->children, 1);
}

void
MyValidityErrorFunc (void *ctx, const char *msg, ...)
{
   va_list list;

   va_start (list, msg);

   printf ("Error: ");
   vprintf (msg, list);

   va_end (list);

   if (ctx)
   {
#if 0
      OmniDomParser *myCtx = (OmniDomParser *)ctx;

      myCtx->setErrorCondition ();
#endif
   }
}

void
MyValidityWarningFunc (void *ctx, const char *msg, ...)
{
   va_list list;

   va_start (list, msg);

   printf ("Warning: ");
   vprintf (msg, list);

   va_end (list);
}

bool
processTree (xmlNodePtr nodeRoot,
             bool       fUpdate)
{
   bool fRC = false;

   while (nodeRoot)
   {
      if (  0 == xmlStrcmp ((const xmlChar *)"name", nodeRoot->name)
         || 0 == xmlStrcmp ((const xmlChar *)"connectionForm", nodeRoot->name)
         )
      {
         const xmlChar *pszNameText    = 0;

         // Get current name contents
         pszNameText = xmlNodeListGetString (nodeRoot->doc,
                                             nodeRoot->children,
                                             1);

         if (pszNameText)
         {
            PSZRO pszShortName = 0;
            PSZRO pszLongName  = 0;

            pszShortName = DeviceForm::getShortFormName ((PSZCRO)pszNameText);

            if (pszShortName)
            {
               pszLongName = DeviceForm::getLongFormName ((PSZCRO)pszNameText);
            }

            if (  pszLongName
               && 0 != strcmp ((PSZRO)pszNameText, pszLongName)
               && fUpdate
               )
            {
               std::cout << SAFE_PRINT_PSZ ((PSZRO)pszNameText)
                         << " -> "
                         << SAFE_PRINT_PSZ (pszLongName)
                         << std::endl;

               // Change the name to the newly mapped name
               xmlNodeSetContent (nodeRoot, (const xmlChar*)pszLongName);

               fRC = true;
            }

            if (pszShortName)
            {
               free ((void *)pszShortName);
            }

            if (pszLongName)
            {
               free ((void *)pszLongName);
            }
         }

         // Clean up
         if (pszNameText)
         {
            xmlFree ((void *)pszNameText);
         }
      }
      else if (nodeRoot->children)
      {
         fRC = processTree (nodeRoot->children, fUpdate) || fRC;
      }

      nodeRoot = nextNode (nodeRoot);
   }

   return fRC;
}

bool
processFile (const char *pszFileName,
             bool        fUpdate)
{
#ifdef DTD
   xmlValidCtxt        validatorDTD    = {
      0,
      MyValidityErrorFunc,
      MyValidityWarningFunc
   };
#endif
   xmlDocPtr           doc             = 0;
   xmlNodePtr          nodeRoot        = 0;
   bool                fRC             = false;

   doc = xmlParseFile (pszFileName);

   if (vfDebug) std::cerr << "doc = " << std::hex << (int)doc << std::dec << std::endl;

   if (!doc)
   {
      return false;
   }

#ifdef DTD
   int iRC = xmlValidateDocument (&validatorDTD, doc);

   if (vfDebug) std::cerr << "xmlValidateDocument = " << iRC << std::endl;
#endif

   nodeRoot = firstNode (xmlDocGetRootElement (doc));
   if (!nodeRoot)
   {
      return false;
   }

   fRC = processTree (nodeRoot, fUpdate);

   if (  fRC
      && vfDumpFile
      )
   {
      xmlDocDump (stdout, doc);
   }

   if (  fRC
      && fUpdate
      )
   {
      xmlSaveFile (pszFileName, doc);
   }

   xmlFreeDoc (doc);

   return true;
}

int
main (int argc, char *argv[])
{
   std::ostringstream oss;
   glob_t             globbuf;
   int                rc;

   // Call glob
   memset (&globbuf, 0, sizeof (globbuf));

   oss << "*.xml"
       << std::ends;

   rc = glob (oss.str ().c_str (), 0, NULL, &globbuf);

   if (rc)
   {
      std::cerr << "Error: glob (" << oss.str () << ") failed. rc = " << rc << std::endl;

      return __LINE__;
   }

   if (0 == globbuf.gl_pathc)
   {
      std::cerr << "Error: glob (" << oss.str () << ") returned 0 files!" << std::endl;

      return __LINE__;
   }

   // Call succeded
   for (int i = 0; i < (int)globbuf.gl_pathc; i++)
   {
      std::cout << globbuf.gl_pathv[i] << std::endl;

      processFile (globbuf.gl_pathv[i], true);
   }

   globfree (&globbuf);

   return 0;
}
